var async    = require('async');
var moment = require('moment');
var canvass, compareScreenshot;

function screenshot (client) {
    var fileName = moment().format('YYYYMMDDHHmmss');
    var screenShotPath = './automated-tests/screenshots/';
    var fileFormat = '.png';
    var filePath = screenShotPath+fileName+fileFormat;
    client.saveScreenshot(filePath);
}

module.exports = {
    '@tags': ['cf_delete'],
    before: function (client) {
        client
            .url(client.launchUrl)
            .waitForElementVisible('body', 1000);
    },

    after: function (client) {

        client.end();
    },
    /**
     *  Test: 9-001
     *  Expects:
     *      - Hark Designer will be loaded in the browser.
     *      - Nodes and links should be drawn in the sheet.
     */
    "9-001": function (client) {
        async.series([
                //load HARK Designer
                function(callback){
                    canvass = client.page.canvass();
                    canvass.navigate();

                    //check if body is now visible
                    canvass.expect.element('body').to.be.present;

                    //should have Main sheet
                    canvass.expect.element('@mainSheet').to.be.visible;

                    callback();
                },
                //setup fixture
                function (callback) {
                    canvass.loadFixture(client, function () {
                        callback();
                    });
                },
                //check if nodeConstant1 is present in MAIN Sheet
                function(callback) {
                    //select MAIN sheet
                    canvass.selectSheet('@mainSheetTab', function () {
                        canvass.expect.element('@nodeConstant1').to.be.visible;
                        callback();
                    });
                }
            ],
            function(){
                //finished
            });
    },
    /**
     *  Test: 9-002
     *  Expects:
     *      - Context menu will appear
     */
    "9-002": function (client) {
        async.series([
                //select MAIN sheet
                function (callback) {
                    canvass.selectSheet('@sheet1Tab', function () {
                        callback();
                    });
                },
                //mouse over nodeConstant1
                function (callback) {
                    client.moveToElement('#node_ChannelSelector_1_delim_INPUT > div', 3, 3, function () {
                        callback();
                    });
                },
                //right click
                function (callback) {
                    client.mouseButtonClick('right', function(){
                        callback();
                    });
                },
                //context menu should be displayed
                function (callback) {
                    canvass.waitForElementPresent('@terminalMenu', 1000, function () {
                        canvass.expect.element('@terminalMenu').to.be.visible;
                        callback();
                    });
                }
            ],
            function(){
                //finished
            });
    },
    /**
     *  Test: 9-003
     *  Expects:
     *      - Link between the AudioSteamFromWave and ChannelSelector will be removed.
     *      - The user can also do the deletion of link using the AUDIO terminal of the AudioStreamFromWave. The link in between will be removed.
     */
    "9-003": function (client) {
        async.series([
                //there should be Delete Link option
                function (callback) {
                    canvass.waitForElementPresent('@deleteLinkOption', 1000, function () {
                        callback();
                    })
                },
                //click Delete Link option
                function (callback) {
                    canvass.click('@deleteLinkOption', function(){
                        callback();
                    });
                },
                //create screenshot
                function (callback) {
                    compareScreenshot = client.page.compareScreenshot();
                    compareScreenshot.screenshot('cf_delete-9-003.png', client, function () {
                        callback();
                    });
                },
                //compare actual screenshot with expected
                function (callback) {
                    compareScreenshot.compare('cf_delete-9-003.png', 'cf_delete-9-003.png', function (result) {
                        client.assert.equal(result, true);
                        callback();
                    });
                }
            ],
            function(){
                //finished
            });
    },
    /**
     *  Test: 9-004
     *  Expects:
     *      - Link between the MultiFFT and ChannelSelector will be removed.
     *      - The user can also do the deletion of link using the INPUT terminal of the MultiFFT. The link in between will be removed.
     */
    "9-004": function (client) {
        async.series([
                //there should be Delete Link option
                function (callback) {
                    canvass.deleteLink(client, '#node_ChannelSelector_1_delim_OUTPUT > div', function(){
                        callback();
                    });
                },
                //pause
                function (callback) {
                    canvass.api.pause(2000, function () {
                        callback();
                    });
                },
                //create screenshot
                function (callback) {
                    compareScreenshot.screenshot('cf_delete-9-004.png', client, function () {
                        callback();
                    });
                },
                //compare actual screenshot with expected
                function (callback) {
                    compareScreenshot.compare('cf_delete-9-004.png', 'cf_delete-9-004.png', function (result) {
                        client.assert.equal(result, true);
                        callback();
                    });
                }
            ],
            function(){
                //finished
            });
    },
    /**
     *  Test: 9-005
     *  Expects:
     *      - Link between the MultiFFT and ChannelSelector will be drawn.
     *      - The user can also do the creation of link from the INPUT terminal of the MultiFFT to AUDIO terminal of AudioStreamFromWave. The link in between will be removed.
     */
    "9-005": function (client) {
        async.series([
                //connect nodeAudioStreamFromWave1 to nodeMultiFFT1
                function (callback) {
                    canvass.linkNodes(client, '#node_AudioStreamFromWave_1_delim_AUDIO > div', '#node_MultiFFT_1_delim_INPUT > div', 20, 10, function () {
                        callback();
                    });
                },
                //pause
                function (callback) {
                    canvass.api.pause(2000, function () {
                        callback();
                    });
                },
                //create screenshot
                function (callback) {
                    compareScreenshot.screenshot('cf_delete-9-005.png', client, function () {
                        callback();
                    });
                },
                //compare actual screenshot with expected
                function (callback) {
                    compareScreenshot.compare('cf_delete-9-005.png', 'cf_delete-9-005.png', function (result) {
                        client.assert.equal(result, true);
                        callback();
                    });
                }
            ],
            function(){
                //finished
            });
    },
    /**
     *  Test: 9-006
     *  Expects:
     *      - Network with the new link will be saved.
     */
    "9-006": function (client) {
        async.series([
                //save network
                function (callback) {
                    canvass.saveNetwork(client, 'sample_delete.n', function(){
                        callback();
                    });
                },
                //save sheet dialog should be closed
                function (callback){
                    canvass.expect.element('@saveSheetDialog').to.not.be.visible;
                }
            ],
            function(){
                //finished
            });
    },
    /**
     *  Test: 9-007
     *  Expects:
     *      - The saved network "sample_delete.n" will be loaded in the Hark Designer.
     *      - All the nodes are properly connected with each other. Links between AudioStreamFromWave and MultiFFT will be drawn.
     *      - There will be no links connected on the ChannelSelector node.
     */
    "9-007": function (client) {
        async.series([
                function (callback) {
                    //load sample_delete.n
                    canvass.loadNetwork(client, 'sample_delete.n', function () {
                        callback();
                    });
                },
                //file dialog should be closed
                function (callback){
                    canvass.expect.element('@fileDialog').to.not.be.visible;
                    callback();
                },
                //select LOOP sheet
                function (callback) {
                    canvass.selectSheet('@sheet1Tab', function () {
                        callback();
                    });
                },
                //check manually if changes are reflected
                function (callback) {
                    //screenshot
                    screenshot(client);
                    callback();
                }
            ],
            function(){
                //finished
            });
    },
    /**
     *  Test: 9-008
     *  Expects:
     *      - ChannelSelector node will be deleted in the sheet.
     */
    "9-008": function (client) {
        async.series([
                //select nodeChannelSelector1
                function (callback) {
                    canvass.selectNode('@nodeChannelSelector1', function () {
                        callback();
                    });
                },
                //nodeChannelSelector1 should be selected
                function (callback) {
                    canvass.expect.element('@nodeChannelSelector1').to.have.attribute('class').contains('nodeSelected');
                    callback();
                },
                //delete nodeChannelSelector1
                function (callback) {
                    canvass.deleteNode(client, '#node_ChannelSelector_1', 2, function () {
                        callback();
                    });
                },
                //nodeChannelSelector1 should be deleted
                function (callback){
                    canvass.expect.element('@nodeChannelSelector1').to.not.be.present;
                    callback();
                }
            ],
            function(){
                //finished
            });
    },
    /**
     *  Test: 9-009
     *  Expects:
     *      - Network with the deletion will be saved.
     */
    "9-009": function (client) {
        async.series([
                //save network
                function (callback) {
                    canvass.saveNetwork(client, 'sample_delete.n', function(){
                        callback();
                    });
                },
                //save sheet dialog should be closed
                function (callback){
                    canvass.expect.element('@saveSheetDialog').to.not.be.visible;
                }
            ],
            function(){
                //finished
            });
    },
    /**
     *  Test: 9-010
     *  Expects:
     *      - The saved network "sample_delete.n" will be loaded in the Hark Designer.
     *      - All the nodes are properly connected with each other.
     *      - There will be no ChannelSelector node in the LOOP sheet.
     */
    "9-010": function (client) {
        async.series([
                function (callback) {
                    //load sample_delete.n
                    canvass.loadNetwork(client, 'sample_delete.n', function () {
                        callback();
                    });
                },
                //file dialog should be closed
                function (callback){
                    canvass.expect.element('@fileDialog').to.not.be.visible;
                    callback();
                },
                //select LOOP sheet
                function (callback) {
                    canvass.selectSheet('@sheet1Tab', function () {
                        callback();
                    });
                },
                //nodeChannelSelector1 should not be present
                function (callback) {
                    canvass.expect.element('@nodeChannelSelector1').to.not.be.present;
                    callback();
                }
            ],
            function(){
                //finished
            });
    }
};