var async    = require('async');
var canvass;

module.exports = {
    '@tags': ['cf_theme'],
    before: function (client) {
        client
            .url(client.launchUrl)
            .waitForElementVisible('body', 1000);
    },

    after: function (client) {

        client.end();
    },
    /**
     *  Test: 12-001
     *  Expects:
     *      - Hark Designer will be loaded in the browser.
     *      - Nodes and links should be drawn in the sheet.
     */
    "12-001": function (client) {
        async.series([
                //load HARK Designer
                function(callback){
                    canvass = client.page.canvass();
                    canvass.navigate();

                    //check if body is now visible
                    canvass.expect.element('body').to.be.present;

                    //should have Main sheet
                    canvass.expect.element('@mainSheet').to.be.visible;

                    callback();
                },
                //setup fixture
                function (callback) {
                    canvass.loadFixture(client, function () {
                        callback();
                    });
                }
            ],
            function(){
                //finished
            });
    },
    /**
     *  Test: 12-002
     *  Expects:
     *      - LOOP node will change its color from Green to Red.
     *      - Theme colors:
     *          Fire     : Red
     *          Sea      : Blue
     *          Dark    : Black
     *          Classic : Green
     */
    "12-002": function (client) {
        async.series([
                //go to MAIN sheet
                function (callback) {
                    canvass.selectSheet('@mainSheetTab', function () {
                        callback();
                    });
                },
                //set Theme of nodeLOOP1 to fire
                function (callback) {
                    canvass.setTheme(client, '#node_LOOP_1', '@fireThemeOption', function () {
                        callback();
                    });
                },
                //check if node has fire theme
                function (callback) {
                    client.verify.cssClassPresent('#node_LOOP_1', 'nodeThemeFire');
                }
            ],
            function(){
                //finished
            });
    },
    /**
     *  Test: 12-003
     *  Expects:
     *      - DisplayLocalization node will change its color from Green to Blue.
     *      - Theme colors:
     *          Fire     : Red
     *          Sea      : Blue
     *          Dark    : Black
     *          Classic : Green
     */
    "12-003": function (client) {
        async.series([
                //go to LOOP sheet
                function (callback) {
                    canvass.selectSheet('@sheet1Tab', function () {
                        callback();
                    });
                },
                //set Theme of nodeDisplayLocalization1 to sea
                function (callback) {
                    canvass.setTheme(client, '#node_DisplayLocalization_1', '@seaThemeOption', function () {
                        callback();
                    });
                },
                //check if node has sea theme
                function (callback) {
                    client.verify.cssClassPresent('#node_DisplayLocalization_1', 'nodeThemeSea');
                }
            ],
            function(){
                //finished
            });
    },
    /**
     *  Test: 12-004
     *  Expects:
     *      - SourceTracker node will change its color from Green to Black.
     *      - Theme colors:
     *          Fire     : Red
     *          Sea      : Blue
     *          Dark    : Black
     *          Classic : Green
     */
    "12-004": function (client) {
        async.series([
                //go to LOOP sheet
                function (callback) {
                    canvass.selectSheet('@sheet1Tab', function () {
                        callback();
                    });
                },
                //set Theme of nodeSourceTracker to dark
                function (callback) {
                    canvass.setTheme(client, '#node_SourceTracker_1', '@darkThemeOption', function () {
                        callback();
                    });
                },
                //check if node has sea dark
                function (callback) {
                    client.verify.cssClassPresent('#node_SourceTracker_1', 'nodeThemeDark');
                }
            ],
            function(){
                //finished
            });
    },
    /**
     *  Test: 12-005
     *  Expects:
     *      - NormalizeMUSIC node will retain its color (Green).
     *      - Theme colors:
     *          Fire     : Red
     *          Sea      : Blue
     *          Dark    : Black
     *          Classic : Green
     */
    "12-005": function (client) {
        async.series([
                //go to LOOP sheet
                function (callback) {
                    canvass.selectSheet('@sheet1Tab', function () {
                        callback();
                    });
                },
                //set Theme of nodeNormalizeMUSIC to classic
                function (callback) {
                    canvass.setTheme(client, '#node_NormalizeMUSIC_1', '@classicThemeOption', function () {
                        callback();
                    });
                },
                //check if node has sea dark
                function (callback) {
                    client.verify.cssClassPresent('#node_NormalizeMUSIC_1', 'nodeThemeClassic');
                }
            ],
            function(){
                //finished
            });
    },
    /**
     *  Test: 12-006
     *  Expects:
     *      - Save sheet dialog should be closed
     */
    "12-006": function (client) {
        async.series([
                //save network as sample_theme.n
                function (callback) {
                    canvass.saveNetwork(client, 'sample_theme.n', function () {
                        callback();
                    });
                },
                //Save sheet dialog should be closed
                function(callback) {
                    canvass.expect.element('@saveSheetDialog').to.not.be.visible;
                    callback();
                }
            ],
            function(){
                //finished
            });
    },
    /**
     *  Test: 12-007
     *  Expects:
     *      - The saved network "sample_theme.n" will be loaded in the Hark Designer.
     *      - Nodes will be loaded with the set theme in configuration
     *      - Network nodes should be in this colors:
     *          LOOP: Green
     *          DisplayLocalization: Green
     *          SourceTracker: Green
     *          NormalizeMUSIC: Green
     *          Others: Green
     */
    "12-007": function (client) {
        async.series([
                //load sample_theme.n
                function (callback) {
                    canvass.loadNetwork(client, 'sample_theme.n', function () {
                        callback();
                    })
                },
                //check if LOOP is red
                function(callback) {
                    canvass.selectSheet('@mainSheetTab', function () {
                        client.verify.cssClassPresent('#node_LOOP_1', 'nodeThemeClassic');
                        callback();
                    });
                },
                //check if nodeDisplayLocalization is blue
                function(callback) {
                    canvass.selectSheet('@sheet1Tab', function () {
                        client.verify.cssClassPresent('#node_DisplayLocalization_1', 'nodeThemeClassic');
                        callback();
                    });
                },
                //check if nodeSourceTracker is black
                function(callback) {
                    canvass.selectSheet('@sheet1Tab', function () {
                        client.verify.cssClassPresent('#node_SourceTracker_1', 'nodeThemeClassic');
                        callback();
                    });
                },
                //check if nodeNormalizeMUSIC is green
                function(callback) {
                    canvass.selectSheet('@sheet1Tab', function () {
                        client.verify.cssClassPresent('#node_NormalizeMUSIC_1', 'nodeThemeClassic');
                        callback();
                    });
                },
                //check if other nodes in MAIN sheet are green
                function(callback) {
                    canvass.selectSheet('@mainSheetTab', function () {
                        client.verify.cssClassPresent('#node_Constant_1', 'nodeThemeClassic');
                        client.verify.cssClassPresent('#node_InputStream_1', 'nodeThemeClassic');
                        callback();
                    });
                },
                //check if other nodes in LOOP sheet are green
                function(callback) {
                    canvass.selectSheet('@sheet1Tab', function () {
                        client.verify.cssClassPresent('#node_AudioStreamFromWave_1', 'nodeThemeClassic');
                        client.verify.cssClassPresent('#node_ChannelSelector_1', 'nodeThemeClassic');
                        client.verify.cssClassPresent('#node_MultiFFT_1', 'nodeThemeClassic');
                        client.verify.cssClassPresent('#node_LocalizeMUSIC_1', 'nodeThemeClassic');
                        callback();
                    });
                }
            ],
            function(){
                //finished
            });
    }
};
