var async = require('async');
var del = require('del');

var defaultDir = '/usr/src/app/userdata/networks';

function checkBoundary(client, canvasId ,elementId, direction) {
    var boundary;
    client.getElementSize(canvasId, function(result) {
        client.getElementSize(elementId, function(elementSize) {
            if(direction === 'right') {
                boundary = parseInt((result.value.width - elementSize.value.width) / 10, 10) * 10;
                client.expect.element(elementId).to.have.attribute('style').contains('left: '+ boundary +'px;');
            } else if(direction === 'down') {
                boundary = parseInt((result.value.height - elementSize.value.height) / 10, 10) * 10;
                client.expect.element(elementId).to.have.attribute('style').contains('top: '+ boundary +'px;');
            }

        });
    });
}

module.exports = {
    '@tags': ['cl_arrowkeyNode'],

    before: function (client) {
        client
            .url('http://localhost:3000')
            .waitForElementVisible('#topmenu', 1000);
    },

    after: function (client) {
        // Clean up
        del.sync(defaultDir + '/*');

        client.end();
    },

    /**
     * Test: 1-029
     * Expects:
     * - Node should be dragged to the leftmost part of the canvass and should not be dragged outside the boundary.
     */
    "1-029" : function (client) {
        var canvassLayout = client.page.canvassLayout();

        async.series([
            function (callback) {
                canvassLayout.clearValue('@inputSearch');
                canvassLayout.setValue('@inputSearch', 'localize');
                callback();
            },
            function (callback) {
                canvassLayout.clickLocalizeMusic(function () {
                    callback();
                });
            },
            function (callback) {
                client.click('#node_LocalizeMUSIC_1');
                callback();
            },

            // Move node element to the leftmost part of the canvass
            function (callback) {
                for(var i = 0; i < 10; i++) {
                    client.keys([client.Keys.LEFT_ARROW]);
                }
                callback();
            },
            // Expect #node_LocalizeMUSIC_1 that it should not be go outside the LEFT boundary
            function (callback) {
                // 1-029
                client.expect.element('#node_LocalizeMUSIC_1').to.have.attribute('style').contains('left: 0px;');
                callback();
            }
        ], function (err, results) {

        })
    },

    /**
     * Test: 1-030
     * Expects:
     * - Node should be dragged to the leftmost part of the canvass and should not be dragged outside the boundary.
     */
    "1-030" : function (client) {
        var canvassLayout = client.page.canvassLayout();

        async.series([
            // Move node element to the topmost part of the canvass
            function (callback) {
                for(var i = 0; i < 15; i++) {
                    client.keys([client.Keys.UP_ARROW]);
                }
                callback();
            },
            // Expect #node_LocalizeMUSIC_1 that it should not be go outside the TOP boundary
            function (callback) {
                // 1-030
                client.expect.element('#node_LocalizeMUSIC_1').to.have.attribute('style').contains('top: 0px;');
                callback();
            }
        ], function (err, results) {

        })
    },

    /**
     * Test: 1-031
     * Expects:
     * - Node should be dragged to the leftmost part of the canvass and should not be dragged outside the boundary.
     */
    "1-031" : function (client) {
        var canvassLayout = client.page.canvassLayout();


        async.series([
            // Move node element to the rightmost part of the canvass
            function (callback) {
                for(var i = 0; i < 1100; i++) {
                    client.keys([client.Keys.RIGHT_ARROW]);
                }

                callback();
            },

            // Expect #node_LocalizeMUSIC_1 that it should not be go outside the RIGHT boundary
            function (callback) {
                // 1-031
                checkBoundary(client, '#canvas2_MAIN', '#node_LocalizeMUSIC_1', 'right');
                callback();
            }
        ], function (err, results) {

        })
    },

    /**
     * Test: 1-032
     * Expects:
     * - Node should be dragged to the leftmost part of the canvass and should not be dragged outside the boundary.
     */
    "1-032" : function (client) {
        var canvassLayout = client.page.canvassLayout();

        async.series([
            // Move node element to the bottommost part of the canvass
            function (callback) {
                for(var i = 0; i < 1100; i++) {
                    client.keys([client.Keys.DOWN_ARROW]);
                }
                callback();
            },
            // Expect #node_LocalizeMUSIC_1 that it should not be go outside the BOTTOM boundary
            function (callback) {
                // 1-032
                checkBoundary(client, '#canvas2_MAIN', '#node_LocalizeMUSIC_1', 'down');
                callback();
            }
        ], function (err, results) {

        })

    },

    /**
     * Test: 1-033
     * Expects:
     * - Selected nodes must not be dragged outside the bounds of the canvass
     * - Selected nodes must be visible in the canvass displayed.
     */
    "1-033" : function (client) {
        var canvassLayout = client.page.canvassLayout();

        async.series([
            // Clear network
            function (callback) {
                canvassLayout.clickClearBtn(function () {
                    callback();
                });
            },
            function (callback) {
                client.pause(3000);
                client.acceptAlert(function () {
                    callback();
                })
            },

            function (callback) {
                canvassLayout.clickLocalizeMusic(function () {
                    callback();
                });
            },

            function (callback) {
                canvassLayout.clickLocalizeMusic(function () {
                    callback();
                });
            },

            function (callback) {
                client.click('#node_LocalizeMUSIC_2');
                callback();
            },

            // Move node_LocalizeMUSIC_2 to the bottom right of node_LocalizeMUSIC_1
            function (callback) {
                for(var i = 0; i < 10; i++) {
                    client.keys([client.Keys.DOWN_ARROW]);
                }
                callback();
            },
            function (callback) {
                for(var i = 0; i < 10; i++) {
                    client.keys([client.Keys.RIGHT_ARROW]);
                }
                callback();
            },

            function (callback) {
                client
                    .moveToElement('#sheet_MAIN', 0, 0)
                    .mouseButtonDown('left')
                    .moveToElement('body', 350, 350)
                    .mouseButtonUp('left', function () {
                        callback();
                    });
            },

            // Move node element to the leftmost part of the canvass
            function (callback) {
                for(var i = 0; i < 15; i++) {
                    client.keys([client.Keys.LEFT_ARROW]);
                }
                callback();
            },
            // Expect #node_LocalizeMUSIC_1 that it should not be go outside the LEFT boundary
            function (callback) {
                // 1-033
                client.expect.element('#node_LocalizeMUSIC_1').to.have.attribute('style').contains('left: 0px;');
                callback();
            }
        ], function (err, results) {

        })

    },

    /**
     * Test: 1-034
     * Expects:
     * - Selected nodes must not be dragged outside the bounds of the canvass
     * - Selected nodes must be visible in the canvass displayed.
     */
    "1-034" : function (client) {
        var canvassLayout = client.page.canvassLayout();

        async.series([
            // Move node element to the topmost part of the canvass
            function (callback) {
                for(var i = 0; i < 15; i++) {
                    client.keys([client.Keys.UP_ARROW]);
                }
                callback();
            },
            // Expect #node_LocalizeMUSIC_1 that it should not be go outside the TOP boundary
            function (callback) {
                // 1-034
                client.expect.element('#node_LocalizeMUSIC_1').to.have.attribute('style').contains('top: 0px;');
                callback();
            }
        ], function (err, results) {

        })
    },

    /**
     * Test: 1-035
     * Expects:
     * - Selected nodes must not be dragged outside the bounds of the canvass
     * - Selected nodes must be visible in the canvass displayed.
     */
    "1-035" : function (client) {
        var canvassLayout = client.page.canvassLayout();

        async.series([
            // Move node element to the topmost part of the canvass
            function (callback) {
                for(var i = 0; i < 1100; i++) {
                    client.keys([client.Keys.RIGHT_ARROW])
                        .getLocationInView('#node_LocalizeMUSIC_2');
                }
                callback();
            },
            // Expect #node_LocalizeMUSIC_1 that it should not be go outside the RIGHT boundary
            function (callback) {
                // 1-035
                checkBoundary(client, '#canvas2_MAIN', '#node_LocalizeMUSIC_2', 'right');
                callback();
            }
        ], function (err, results) {

        })
    },

    /**
     * Test: 1-036
     * Expects:
     * - Selected nodes must not be dragged outside the bounds of the canvass
     * - Selected nodes must be visible in the canvass displayed.
     */
    "1-036" : function (client) {
        var canvassLayout = client.page.canvassLayout();

        async.series([
            // Move node element to the bottommost part of the canvass
            function (callback) {
                for(var i = 0; i < 1100; i++) {
                    client.keys([client.Keys.DOWN_ARROW])
                        .getLocationInView('#node_LocalizeMUSIC_2');
                }
                callback();
            },
            // Expect #node_LocalizeMUSIC_2 that it should not be go outside the BOTTOM boundary
            function (callback) {
                // 1-036
                checkBoundary(client, '#canvas2_MAIN', '#node_LocalizeMUSIC_2', 'down');
                callback();
            }
        ], function (err, results) {

        })
        client.end();
    }
};