var async = require('async');
var del = require('del');

var defaultDir = '/usr/src/app/userdata/networks';

module.exports = {
    '@tags': ['cl_canvassArrowkeyNote'],

    before: function (client) {
        client
            .url('http://localhost:3000')
            .waitForElementVisible('#topmenu', 1000);
    },

    after: function (client) {
        // Clean up
        del.sync(defaultDir + '/*');

        client.end();
    },

    /**
     * Test: 2-001
     * Expects:
     * - Hark Designer should display all available node groups according to their def files.
     */
    "2-001" : function (client) {
        var canvassLayout = client.page.canvassLayout();

        async.series([
            function (callback) {
                // 2-001
                for (var i = 1; i <= 33; i++) {
                    canvassLayout.expect.element('#nodelist > li:nth-child(' + i + ')').to.be.present;
                }
                callback();
            }
        ], function (err, results) {

        });
    },

    /**
     * Test: 2-002
     * Expects:
     * - A create note dialog will appear
     */
    "2-002" : function (client) {
        var canvassLayout = client.page.canvassLayout();

        async.series([
            function (callback) {
                client.click('#sheet_MAIN');
                callback();
            },
            function (callback) {
                client.moveToElement('#sheet_MAIN', 50, 50)
                    .mouseButtonClick('right', function () {
                        callback();
                    })
            },
            function (callback) {
                canvassLayout.clickAddNote(function () {
                    callback();
                })
            },
            function (callback) {
                // 2-002
                canvassLayout.expect.element('@addNoteDialog').to.be.visible;
                callback();
            }
        ], function (err, results) {

        });
    },

    /**
     * Test: 2-003
     * Expects:
     * - Note displaying "Sample text" will be added into the canvass
     */
    "2-003" : function (client) {
        var canvassLayout = client.page.canvassLayout();

        async.series([
            function (callback) {
                canvassLayout.clearValue('@addNoteTextArea');
                canvassLayout.setValue('@addNoteTextArea', 'Sample text', function () {
                    canvassLayout.click('@addNoteOk', function () {
                        callback();
                    })
                });
            },
            function (callback) {
                canvassLayout.expect.element('#note0').to.be.present;
                canvassLayout.expect.element('#note0').text.to.equal('Sample text');
                callback();
            }
        ], function (err, results) {

        });
    },

    /**
     * Test: 2-004
     * Expects:
     * - Selected note must not be dragged outside the bounds of the canvass
     */
    "2-004" : function (client) {
        var canvassLayout = client.page.canvassLayout();

        async.series([
            function (callback) {
                canvassLayout.click('#note0');
                callback();
            },
            function (callback) {
                for(var i = 0; i < 10; i++) {
                    client.keys([client.Keys.LEFT_ARROW]);
                }
                callback();
            },
            // Expect #note0 that it should not be go outside the LEFT boundary
            function (callback) {
                // 2-004
                client.expect.element('#note0').to.have.attribute('style').contains('left: 0px;');
                callback();
            }
        ], function (err, results) {

        });
    },

    /**
     * Test: 2-005
     * Expects:
     * - Selected note must not be dragged outside the bounds of the canvass
     */
    "2-005" : function (client) {
        var canvassLayout = client.page.canvassLayout();

        async.series([
            function (callback) {
                for(var i = 0; i < 5; i++) {
                    client.keys([client.Keys.UP_ARROW]);
                }
                callback();
            },
            // Expect #note0 that it should not be go outside the TOP boundary
            function (callback) {
                // 2-005
                client.expect.element('#note0').to.have.attribute('style').contains('top: 0px;');
                callback();
            }
        ], function (err, results) {

        });
    },

    /**
     * Test: 2-005
     * Expects:
     * - Selected note must not be dragged outside the bounds of the canvass
     */
    "2-005" : function (client) {
        var canvassLayout = client.page.canvassLayout();

        async.series([
            function (callback) {
                for(var i = 0; i < 5; i++) {
                    client.keys([client.Keys.UP_ARROW]);
                }
                callback();
            },
            // Expect #note0 that it should not be go outside the TOP boundary
            function (callback) {
                // 2-005
                client.expect.element('#note0').to.have.attribute('style').contains('top: 0px;');
                callback();
            }
        ], function (err, results) {

        });
    },

    /**
     * Test: 2-006
     * Expects:
     * - Selected note must not be dragged outside the bounds of the canvass
     */
    "2-006" : function (client) {
        var canvassLayout = client.page.canvassLayout();

        async.series([
            function (callback) {
                for(var i = 0; i < 1100; i++) {
                    client.keys([client.Keys.RIGHT_ARROW]);
                }
                callback();
            },
            // Expect #note0 that it should not be go outside the RIGHT boundary
            function (callback) {
                // 2-006
                var boundary;
                client.getElementSize('#canvas2_MAIN', function(result) {
                    client.getElementSize('#note0', function(noteResult) {
                        boundary = result.value.width - noteResult.value.width;
                        boundary = parseInt(boundary / 10, 10) * 10;
                        client.expect.element('#note0').to.have.attribute('style').contains('left: '+ boundary +'px;');
                    });
                });
                callback();
            }
        ], function (err, results) {

        });
    },

    /**
     * Test: 2-007
     * Expects:
     * - Selected note must not be dragged outside the bounds of the canvass
     */
    "2-007" : function (client) {
        var canvassLayout = client.page.canvassLayout();

        async.series([
            function (callback) {
                for(var i = 0; i < 1100; i++) {
                    client.keys([client.Keys.DOWN_ARROW])
                        .getLocationInView('#note0');
                }
                callback();
            },
            // Expect #note0 that it should not be go outside the BOTTOM boundary
            function (callback) {
                // 2-007
                var boundary;
                client.getElementSize('#canvas2_MAIN', function(result) {
                    client.getElementSize('#note0', function(noteResult) {
                        boundary = result.value.height - noteResult.value.height;
                        boundary = parseInt(boundary / 10, 10) * 10;
                        client.expect.element('#note0').to.have.attribute('style').contains('top: '+ boundary +'px;');
                    });
                });
                callback();
            }
        ], function (err, results) {

        });
        client.end();
    }
};