var async = require('async');

module.exports = {
    '@tags': ['nl_list'],

    before: function (client) {
        client
            .url('http://localhost:3000')
            .waitForElementVisible('#topmenu', 1000);
    },

    after: function (client) {
        client.end();
    },

    /**
     * Test: 1-001
     * Expects:
     * - Hark Designer should display all available node groups according to their def files.
     */
    '1-001' : function (client) {

        var node = client.page.node();

        async.series([
            function (callback) {
                node.expect.element('body').to.be.present;
                callback();
            },
            function (callback) {
                for(var i = 1; i < 34; i++) {
                    node.expect.element('#nodelist > li:nth-child('+ i +')').to.be.present;
                }
                callback();
            }
        ], function(err, results) {
        });
    },

    /**
     * Test: 1-002
     * Expects:
     *  - Node menu will be displayed
     *  - Menu will contain the nodes under NNet category
     */
    '1-002' : function (client) {

        var node = client.page.node();

        async.series([
            function (callback) {
                node.expect.element('@nodeNnet').to.be.present;
                callback();
            },
            function (callback) {
                node.clickNnet(function () {
                    callback();
                });
            },
            function (callback) {
                node.expect.element('@nNetUl').to.have.attribute('style').contains('display: block;');
                callback();
            }
        ], function(err, results) {
        });
    },

    /**
     * Test: 1-003
     * Expects:
     *  - "NNetExec" node will be added into the canvass
     */
    '1-003' : function (client) {

        var node = client.page.node();

        async.series([
            function (callback) {
                node.addNode(function () {
                    callback();
                });
            },
            function (callback) {
                node.expect.element('@nNetExecId').to.be.present;
                callback();
            }
        ], function(err, results) {
        });
    },

    /**
     * Test: 1-004
     * Expects:
     *  - Preferences dialog will be displayed
     */
    '1-004' : function (client) {

        var node = client.page.node();

        async.series([
            function (callback) {
                node.openPreferences(function () {
                    callback();
                });
            },
            function (callback) {
                node.expect.element('@preferencesDialog').to.have.attribute('style').contains('display: block;');
                callback();
            }
        ], function(err, results) {
        });
    },

    /**
     * Test: 1-005
     * Expects:
     *  - All available def files will be listed
     *  - All def files should be enabled
     */
    '1-005' : function (client) {

        var node = client.page.node();

        async.series([
            function (callback) {
                node.clickPackages(function () {
                    callback();
                });
            },
            function (callback) {
                for (var i = 1; i <= 8; i++) {
                    node.expect.element('#preferenceDialog > div.dialogMain > div.package > form > fieldset:nth-child(' + i + ') > select').to.be.enabled;
                }
                callback();
            }
        ], function(err, results) {
        });
    },

    /**
     * Test: 1-006
     * Expects:
     *  -  Exclamation icon will appear at the right side of the "Packages" option in the preferences dialog
     */
    '1-006' : function (client) {

        var node = client.page.node();

        async.series([
            function (callback) {
                node.disableNnetDef(function () {
                    callback();
                });
            },
            function (callback) {
                node.expect.element('@nnetDefField').to.have.attribute('class').contains('ui-icon-radio-off');
                callback();
            },
            function (callback) {
                node.expect.element('@nnetSelect').to.be.not.enabled;
                callback();
            },
            function (callback) {
                node.expect.element('@packagesSpan').to.have.attribute('style').contains('display: block;');
                callback();
            }
        ], function(err, results) {
        });
    },

    /**
     * Test: 1-007
     * Expects:
     * - Preferences dialog should disappear.
     * - In the node list, "NNet" node group is removed.
     */
    '1-007' : function (client) {

        var node = client.page.node();

        async.series([
            function (callback) {
                node.closePreferences(function () {
                    callback();
                });
            },
            function (callback) {
                node.expect.element('@preferencesDialog').to.have.attribute('style').contains('display: none;');
                callback();
            },
            function (callback) {
                node.expect.element('@nnetList').to.be.not.present;
                callback();
            }
        ], function(err, results) {
        });
    },

    /**
     * Test: 1-008
     * Expects:
     * - In the node list, "NNet" node group is listed.
     */
    '1-008' : function (client) {

        var node = client.page.node();

        async.series([
            function (callback) {
                node.openPreferences(function () {
                    callback();
                });
            },
            function (callback) {
                node.clickPackages(function () {
                    callback();
                });
            },
            function (callback) {
                node.enableNnetDef(function () {
                    callback();
                });
            },
            function (callback) {
                node.expect.element('@nnetDefField').to.have.attribute('class').contains('ui-icon-bullet');
                callback();
            },
            function (callback) {
                node.closePreferences(function () {
                    callback();
                });
            },
            function (callback) {
                node.expect.element('@nnetList').to.be.present;
                callback();
            }
        ], function(err, results) {
        });
    },

    /**
     * Test: 1-009
     * Expects:
     *  - "Hide list" button should disappear
     *  - "Show list" button is displayed
     *  - Node list should be hidden
     */
    '1-009' : function (client) {

        var node = client.page.node();

        async.series([
            function (callback) {
                node.expect.element('@hideShowBtn').to.be.present;
                callback();
            },
            function (callback) {
                node.clickHideList(function () {
                    callback();
                });
            },
            function (callback) {
                node.expect.element('@nodeList').to.have.attribute('style').contains('display: none;');
                node.expect.element('@hideShowSpan').text.to.equal('Show list');
                callback();
            }
        ], function(err, results) {
        });
    },

    /**
     * Test: 1-010
     * Expects:
     *  - "Show list" button should disappear
     *  - "Hide list" button is displayed
     *  - Node list is displayed
     */
    '1-010' : function (client) {

        var node = client.page.node();

        async.series([
            function (callback) {
                node.clickShowList(function () {
                    callback();
                });
            },
            function (callback) {
                node.expect.element('@nodeList').to.have.attribute('style').contains('display: block;');
                callback();
            },
            function (callback) {
                node.expect.element('@hideShowSpan').text.to.equal('Hide list');
                callback();
            }
        ], function(err, results) {
        });

        client.end();
    }
};