var async = require('async');
module.exports = {
    '@tags': ['pd_packages'],

    before: function (client) {
        client
            .url('http://localhost:3000')
            .waitForElementVisible('#topmenu', 1000);
    },

    after: function (client) {
        client.end();
    },

    /**
     * Test: 1-001
     * Expects:
     *  - Preferences dialog will appear with 6 buttons
     */
    '1-001': function (client) {
        var preferenceTest = client.page.preferenceManagement();

        async.series([
            function (callback) {
                preferenceTest.expect.element('body').to.be.present;
                callback();
            },
            function (callback) {
                preferenceTest.openDialog(function () {
                    callback();
                });
            },
            function (callback) {
                preferenceTest.expect.element('@preferenceDialog').to.be.visible;
                callback();
            },
            function (callback) {
                // 1-001
                preferenceTest.expect.element('@packagesBtn').to.be.visible;
                preferenceTest.expect.element('@displayBtn').to.be.visible;
                preferenceTest.expect.element('@versionBtn').to.be.visible;
                preferenceTest.expect.element('@resetBtn').to.be.visible;
                preferenceTest.expect.element('@tutorialBtn').to.be.visible;
                preferenceTest.expect.element('@environmentBtn').to.be.visible;
                callback();
            }
        ], function (err, stdout) {
        });
    },

    /**
     * Test: 1-002
     * Expects:
     *  - It should display the def files (e.g. libhark.def. libeffect, etc.)
     *  - The display should be in this order:
     *     [bullets] -  name_of_def  - dropdown menu for the preferences.
     *  - All buttons bullets should be enabled(solid circle) and clickable.
     */
    '1-002': function (client) {
        var preferenceTest = client.page.preferenceManagement();

        function checkPackages(package) {
            preferenceTest.expect.element(package).to.be.visible;
        }

        function checkBtnEnabled(element) {
            preferenceTest.expect.element(element).to.have.attribute('class').contains('ui-icon-bullet');
        }

        async.series([
            function (callback) {
                preferenceTest.clickPackagesButton(function () {
                    callback();
                });
            },
            function (callback) {
                // 1-002 Check if default packages are visible
                checkPackages('@package_nnet');
                checkPackages('@package_audioBlocks');
                checkPackages('@package_dataflow');
                checkPackages('@package_effects');
                checkPackages('@package_fuzzy');
                checkPackages('@package_hark');
                checkPackages('@package_hmm');
                checkPackages('@package_vq');

                checkBtnEnabled('@packageBtn_nnet');
                checkBtnEnabled('@packageBtn_audioBlocks');
                checkBtnEnabled('@packageBtn_dataflow');
                checkBtnEnabled('@packageBtn_effects');
                checkBtnEnabled('@packageBtn_fuzzy');
                checkBtnEnabled('@packageBtn_hark');
                checkBtnEnabled('@packageBtn_hmm');
                checkBtnEnabled('@packageBtn_vq');


                for(var i = 0; i > 9; i++) {
                    preferenceTest.expect.element('@packagesDialogValue' + i).to.be.enabled;
                }

                callback();
            }
        ], function (err, stdout) {
        });
    },

    /**
     * Test: 1-003
     * Expects:
     *  - hark.def. will be disabled, bullet should change from solid circle to
     *     hollow cirlce icon and dropdown should be disabled.
     *  - In the "Package" button, info icon (color blue) will be displayed.
     *  - Info icon indicates that not all packages are loaded.
     */
    '1-003': function (client) {
        var preferenceTest = client.page.preferenceManagement();

        async.series([
            function (callback) {
                preferenceTest.clickHarkDefBullet(function () {
                    callback();
                });
            },
            function (callback) {
                // 1-003
                preferenceTest.expect.element('@packageBtn_hark').to.have.attribute('class').contains('ui-icon-radio-off');
                callback();
            }
        ], function (err, stdout) {
        });
    },

    /**
     * Test: 1-004
     * Expects:
     *  - There will be no
     *     HARK:ASRIF
     *     HARK:FeatureExtraction
     *     HARK:AudioIO
     *     HARK:Localization
     *     HARK:Localization:CorrealationMatrix
     *     HARK:MFM
     *     HARK:MISC
     *    from the nodelist. It should be removed.
     */
    '1-004': function (client) {
        var preferenceTest = client.page.preferenceManagement();

        async.series([
            function (callback) {
                preferenceTest.closeDialog(function () {
                    callback();
                });
            },
            function (callback) {
                // 1-004
                for(var i = 1; i < 9; i++) {
                    preferenceTest.expect.element('@nodeHarkDef' + i).to.be.not.present;
                }
                callback();
            }
        ], function (err, stdout) {
        });
    },

    /**
     * Test: 1-005
     * Expects:
     *  - Preferences dialog will appear with 6 buttons namely:
     *     Packages, Display, Version, All Reset, Tutorial, Environment
     *    but this time in Packages button, info icon(color gray) will be displayed.
     *
     *  - Info icon will differ in color. If button is clicked, the color of icon will be in blue.
     *    If not clicked, the color will be in gray.
     */
    '1-005': function (client) {
        var preferenceTest = client.page.preferenceManagement();

        async.series([
            function (callback) {
                preferenceTest.openDialog(function () {
                    callback();
                });
            },
            function (callback) {
                preferenceTest.clickPackagesButton(function () {
                    callback();
                });
            },
            function (callback) {
                // 1-005
                preferenceTest.expect.element('@infoIcon').to.be.visible;
                callback();
            }
        ], function (err, stdout) {
        });
    },

    /**
     * Test: 1-006
     * Expects:
     *  - Hark Designer will load.
     *  - There will be a message in the console box saying "Not all packages are loaded.
     *     Click Preferences to check this issue." since libhark.def is disabled.
     */
    '1-006': function (client) {
        var preferenceTest = client.page.preferenceManagement();

        async.series([
            function (callback) {
                client.refresh()
                    .pause(2000)
                    .acceptAlert();
                callback();
            },
            function (callback) {
                client.pause(5000, function () {
                    callback();
                });
            },
            function (callback) {
                // 1-006
                preferenceTest.expect.element('@execStatus').to.have.text.which
                    .contain("Not all packages are loaded. Click Preferences to check this issue.");
                callback();
            }
        ], function (err, stdout) {
        });
    },

    /**
     * Test: 1-007
     * Expects:
     *  - SpeechRecognitionClient node or any node under the hark.def
     *     will be drawn in the canvass with a theme Sea(color blue).
     */
    '1-007': function (client) {
        var preferenceTest = client.page.preferenceManagement();

        async.series([
            function (callback) {
                preferenceTest.openDialog(function () {
                    callback();
                });
            },
            function (callback) {
                preferenceTest.clickPackagesButton(function () {
                    callback();
                });
            },
            function (callback) {
                preferenceTest.clickHarkDefBullet(function () {
                    callback();
                });
            },
            function (callback) {
                preferenceTest.setPackagesHarkValue(function () {
                    callback();
                })
            },
            function (callback) {
                preferenceTest.closeDialog(function () {
                    callback();
                })
            },
            function (callback) {
                preferenceTest.selectAsrif(function () {
                    callback();
                });
            },
            function (callback) {
                preferenceTest.expect.element('@speechRecognitionClient').to.be.present;
                preferenceTest.expect.element('@speechRecognitionSMNClient').to.be.present;
                callback();
            },
            function (callback) {
                preferenceTest.selectSpeechRecognizer(function () {
                    callback();
                })
            },
            function (callback) {
                // 1-007
                preferenceTest.expect.element('#node_SpeechRecognitionClient_1').to.have.attribute('class').contains('nodeThemeSea');
                callback();
            }
        ], function (err, stdout) {
        });
    },

    /**
     * Test: 1-008
     * Expects:
     *  - Reverb node or any node under libeffect.def will be drawn
     *    in the canvass with a theme Dark(color black).
     */
    '1-008': function (client) {
        var preferenceTest = client.page.preferenceManagement();

        async.series([
            function (callback) {
                preferenceTest.openDialog(function () {
                    callback();
                });
            },
            function (callback) {
                preferenceTest.clickPackagesButton(function () {
                    callback();
                });
            },
            function (callback) {
                preferenceTest.setPackagesEffectToDarkValue(function () {
                    callback();
                });
            },
            function (callback) {
                preferenceTest.closeDialog(function () {
                    callback();
                })
            },
            function (callback) {
                preferenceTest.selectEffect(function () {
                    callback();
                });
            },
            function (callback) {
                preferenceTest.expect.element('@reverb').to.be.visible;
                preferenceTest.expect.element('@saturate').to.be.visible;
                callback();
            },
            function (callback) {
                preferenceTest.click('@reverb', function () {
                    callback();
                });
            },
            function (callback) {
                // 1-008
                preferenceTest.expect.element('#node_Reverb_1').to.have.attribute('class').contains('nodeThemeDark');
                callback();
            }
        ], function (err, stdout) {
        });
    },

    /**
     * Test: 1-009
     * Expects:
     *  - FuzzyModelExec node or any node under libfuzzy.def
     *    will be drawn in the canvass with a theme Fire(color red).
     */
    '1-009': function (client) {
        var preferenceTest = client.page.preferenceManagement();

        async.series([
            function (callback) {
                preferenceTest.openDialog(function () {
                    callback();
                });
            },
            function (callback) {
                preferenceTest.clickPackagesButton(function () {
                    callback();
                });
            },
            function (callback) {
                preferenceTest.setValue('@packagesDialogValue5', 'Fire', function () {
                    callback();
                });
            },
            function (callback) {
                preferenceTest.closeDialog(function () {
                    callback();
                })
            },
            function (callback) {
                preferenceTest.click('@fuzzy', function () {
                    callback();
                });
            },
            function (callback) {
                // 1-009
                preferenceTest.expect.element('@fuzzyModelExec').to.be.visible;
                preferenceTest.expect.element('@fuzzyRule').to.be.visible;
                preferenceTest.expect.element('@fuzzySet').to.be.visible;
                preferenceTest.expect.element('@genericModel').to.be.visible;
                preferenceTest.expect.element('@inferenceModel').to.be.visible;
                preferenceTest.expect.element('@trapezoidalFunction').to.be.visible;
                preferenceTest.expect.element('@triangularFunction').to.be.visible;
                callback();
            },
            function (callback) {
                preferenceTest.click('@fuzzyModelExec', function () {
                    callback();
                });
            },
            function (callback) {
                preferenceTest.expect.element('@fuzzyModelExec').to.be.visible;
                callback();
            },
            function (callback) {
                preferenceTest.expect.element('#node_FuzzyModelExec_1').to.have.attribute('class').contains('nodeThemeFire');
                callback();
            },
        ], function (err, stdout) {
        });
    },

    /**
     * Test: 1-010
     * Expects:
     *  - Network will be saved.
     */
    '1-010': function (client) {
        var preferenceTest = client.page.preferenceManagement();

        async.series([
            function (callback) {
                preferenceTest.clickSaveButton(function () {
                    callback();
                });
            },
            function (callback) {
                preferenceTest.setFileName(function () {
                    callback();
                });
            },
            function (callback) {
                preferenceTest.saveFile(client, function () {
                    callback();
                })
            },
            function (callback) {
                preferenceTest.clickFileManagerButton(function () {
                    callback();
                });
            },
            function (callback) {
                preferenceTest.expect.element('@samplePackageNetwork').to.be.present;
                callback();
            }
        ], function (err, stdout) {
        });
    },

    /**
     * Test: 1-011
     * Expects:
     *  - The saved network "sample_theme.n" will be loaded in the Hark Designer.
     *  - Nodes with the corresponding theme will vary the colors based on the set theme.
     *  - Network nodes should be in this colors:
     *      SpeechRecognitionClient Blue
     *      Reverb: Black
     *      FuzzyModelExec: Red
     */
    '1-011': function (client) {
        var preferenceTest = client.page.preferenceManagement();

        async.series([
            function (callback) {
                client.pause(3000);
                callback();
            },
            function (callback) {
                preferenceTest.loadFile(function () {
                    client.pause(1000);
                    callback();
                })
            },
            function (callback) {
                client.acceptAlert();
                client.pause(1000);
                callback();
            },

            function (callback) {
                // 1-011
                preferenceTest.expect.element('#node_SpeechRecognitionClient_1').to.have.attribute('class').contains('nodeThemeSea');
                preferenceTest.expect.element('#node_Reverb_1').to.have.attribute('class').contains('nodeThemeDark');
                preferenceTest.expect.element('#node_FuzzyModelExec_1').to.have.attribute('class').contains('nodeThemeFire');
                callback();
            }
        ], function (err, stdout) {
        });

        client.end();
    }
}