// partial backbone

var ExecutionStatusDialogView = Backbone.View.extend({
    el: $("#executionStatusDialog"),
    initialize: function() {
        this.$el.dialog({
            autoOpen: false,
            modal: true,
            width: 700,
            height: 600
        });
    },
    events: {
        "click": "onclick"
    },
    onclick: function (event) {
        if (event.target.id === "closeButton") {
            this.$el.dialog("close");
            app.set("execstat", "");
        }
    },
});

var ExecutionModeDialogView = Backbone.View.extend({
    el: $("#executionModeDialog"),
    initialize: function() {
        console.log("initialize...");
        this.$el.dialog({
            autoOpen: false,
            modal: true,
        });
        app.set("executionMode", "local");
    },
    events: {
        "click": "onclick"
    },
    onclick: function (event) {
        if (event.target.id === "cancelButton") {
            this.$el.dialog("close");
        }
        else if (event.target.id === "okButton") {
            let currentFileName = app.get("currentFileName");
            if (currentFileName === "") {
                app.set("execstat", "Save the network in advance");
                return;
            }

            var that = this;
            let executionMode = app.get("executionMode");

            // run the process
            var executor = function (filename) {
                app.set("isRunning", true);

                $.ajax({
                    url: "/preference/environment",
                    method: "PUT",
                    data: userPreference.get("env")
                }).always(function (data) {
                        $.ajax({
                            url: `${rootPath}execution/${executionMode}/${currentFileName}`,
                            cache: false
                        }).done(function (data) {
                            console.log(data);
                            app.set("execstat", data);
                        }).fail(function (data, status, error) {
                            console.log(data, status, error);
                            app.set("execstat", "Error " + data.responseText);
                        });
                        setTimeout(polling, 1000);
                });
            };

            if (app.get("currentFileName") === "") {
                $("#saveFileDialog").dialog("dialogclose", executor);
                $("#saveFileDialog").dialog("open");
                $("#saveFilePerSheetDialog").dialog("dialogclose", executor);
                $("#saveFilePerSheetDialog").dialog("open");
            } else {
                $.ajax({
                    url: rootPath + "file/object/" + currentFileName,
                    type: "PUT",
                    data: network,
                }).done(function () {
                    app.set("execstat", "Saved");
                    updateFileList();
                }).fail(function (data) {
                    console.log("failed", data);
                    app.set("execstat", data.responseText);
                });
                executor();
            }
            this.$el.dialog("close");
        }
        else if (event.target.id === "modeLocal") {
            app.set("executionMode", "local");
        }
        else if (event.target.id === "modeDistribution") {
            app.set("executionMode", "distribution");
        }
        else if (event.target.id === "modeStartup") {
            app.set("executionMode", "startup");
        }
        else {
            //
        }
    },
});

var ExecuteButtonView = Backbone.View.extend({
    el: $("#execute"),
    initialize: function(){
        this.model.on("change", this.switchFilename, this);
        this.model.on("change", this.switchButtonText, this);
        this.switchFilename();
        this.switchButtonText();
    },
    events: {
        "click": "onclick"
    },
    switchButtonText: function(){
        this.$el.text(app.get("isRunning") ? "Stop" : "Execute");
    },
    onclick: function(event){
        if (!app.get("isRunning")) {
            $("#executionModeDialog").dialog("open");
        } else {
            console.log('Stopping execution.')
            $.ajax({
                url: rootPath + "abort"
            }).done(function (data) {
                console.log('Process stopped.')
            }).fail(function (data, status, error) {
                console.log(data, status, error);
                app.set("execstat", "Error " + data.responseText);
            });
            app.set("isRunning", false);
        }
    },
    switchFilename: function(){
        if (this.model.get("currentFileName") !== ""){
            this.$el.addClass("menuElement");
            this.$el.removeClass("menuElementDisabled");
        } else {
            this.$el.addClass("menuElementDisabled");
            this.$el.removeClass("menuElement");
        }
    }
});
function polling(){
    console.log("polling");
    var that = this;
    $.ajax({
        url: rootPath + "execstatus"
    }).done(function(data){
        app.set("execstat", data);
        if ("Process is running." === data){
            console.log("register again");
            setTimeout(that.polling, 1000);
        } else {
            app.set("isRunning", false);
        }
    });
}

var defHelpLUT = [
    "libaudio.def",
    "libeffect.def",
    "libflow.def",
    "libfuzzy.def",
    "libhmm.def",
    "libnnet.def",
    "libvq.def"
];

var ZoomView = Backbone.View.extend({
    el: $("#zoomSlider"),
    initialize: function(){
        this.listenTo(this.model, "change", this.update);
        this.render();
    },
    render: function(){
        var that = this;
        that.$el.slider({
            value: that.model.get("zoom"),
            min: 1,
            max: 10,
            slide: function(event, ui){
                // replace class & calculate the size ratio
                var baseNode = $(".node:visible").first();
                var prev = baseNode.width();
                console.log(that.model.get("zoom"));
                $(".node").removeClass("nodeSize" + that.model.get("zoom"));
                $(".node").addClass("nodeSize" + ui.value);
                that.model.set("zoom", ui.value);
                var aft = baseNode.width();
                var ratio = 1.0 * aft / prev;

                // update node.pos
                // Note: leftest and toppest is kept same
                _.each(_.groupBy(network.nodes, "sheet"), function(nodes){
                    var sheetname = nodes[0].sheet;

                    // leftest/toppest before moving
                    var offsets = _.map(nodes, function(n){return {top: n.pos[1], left: n.pos[0]};});
                    var leftestPrev = _.min(_.pluck(offsets, "left"));
                    var toppestPrev = _.min(_.pluck(offsets, "top"));

                    _.each(nodes, function(n){
                        n.pos[0] *= ratio;
                        n.pos[1] *= ratio;
                    });

                    // leftest/toppest after moving
                    offsets = _.map(nodes, function(n){return {top: n.pos[1], left: n.pos[0]};});
                    var leftestAft = _.min(_.pluck(offsets, "left"));
                    var toppestAft = _.min(_.pluck(offsets, "top"));

                    // shift nodes
                    _.each(nodes, function(n){
                        n.pos[0] = truncate(n.pos[0] + leftestPrev - leftestAft);
                        n.pos[1] = truncate(n.pos[1] + toppestPrev - toppestAft);
                    });

                    // update offset if the sheetname is the currentTab
                    if (sheetname === currentTab){
                        _.each(nodes, function(n){$("#" + n.id).css({left: n.pos[0], top: n.pos[1]});});
                        clearCanvas();
                        drawLinks();
                    }
                });
            }
        });
    },
    update: function(){
        this.$el.slider("value", this.model.get("zoom"));
    }
});

var UserPreferenceModel = Backbone.Model.extend({
    defaults: {
        size: "normal",
        packages: [],
        nodelistmin: false,
        colors  : [],
        link    : "Bezier",
        zoom    : 4,
        lang    : "en",
        showTooltip: "Yes",
        showCandidate: "Yes",
        reAlign: "fold",
        consoleLeft: "",
        consoleTop : "",
        env : "",
        origPackages: []
    },
    initialize: function(){
        this.on("change", this.save);
        this.on("change", this.updateDOMs);
        var that = this;
        //check if origPackages is present
        if(localStorage.userPref) {
            var origPackages = [];
            $.ajax({
                url: "/packagelist",
                method: "GET"
            }).done(function(data){
                _.each(data, function(pkg) {
                    origPackages.push(pkg);
                });
                var userPrefValue = JSON.parse(localStorage.userPref);
                userPrefValue.origPackages = origPackages;
                localStorage.setItem("userPref", JSON.stringify(userPrefValue));

                var value = JSON.parse(localStorage.userPref);
                var attrs = ["showTooltip",
                    "size",
                    "colors",
                    "link",
                    "zoom",
                    "lang",
                    "showCandidate",
                    "reAlign",
                    "consoleLeft",
                    "consoleTop",
                    "env",
                    "origPackages"];
                var i;
                for (i = 0; i < attrs.length; i++){
                    if (value[attrs[i]]) {
                        that.set(attrs[i], value[attrs[i]]);
                    }
                }
                that.setSelectValues();

                var pipes = [];

                for (i = 0; i < value.packages.length; i++){
                    that.addPackage(value.packages[i]);
                }

                $("#console").css("left", value.consoleLeft);
                $("#console").css("top",  value.consoleTop );
                $(document).tooltip();
                if (that.get("showTooltip") === "No"){
                    $(document).tooltip("disable");
                }
            });
        } else {
            // if no localStorage found, load all packages
            $.ajax({
                url: rootPath + "packagelist",
                method: "GET"
            }).done(function(data){
                _.each(data, function(p){
                    userPreference.addPackage(p);
                });
                userPreference.set("origPackages", data);
            });
        }

        $("#preferenceDialog .display button").button().click(function(){
            that.set("lang", "en");
            that.set("link", "Bezier");
            that.set("showTooltip", "Yes");
            that.set("showCandidate", "Yes");
            that.set("reAlign", "fold");
            that.setSelectValues();
            that.trigger("change");
        });
        var maxwidth = _.chain($("#preferenceDialog .display label"))
                        .map(function(l){return $(l).width();})
                        .max().value();
        _.each($("#preferenceDialog .display label"), function(l){
            $(l).css("width", (maxwidth + 20) + "px");
            $(l).css("display", "inline-block");
        });
    },
    setSelectValues: function(){
        $("#preferenceDialog .display select[name=lang]").val(this.get("lang"));
        $("#preferenceDialog .display select[name=line]").val(this.get("link"));
        $("#preferenceDialog .display select[name=tooltip]").val(this.get("showTooltip"));
        $("#preferenceDialog .display select[name=candidate]").val(this.get("showCandidate"));
        $("#preferenceDialog .display select[name=realign]").val(this.get("reAlign"));
    },
    updateDOMs: function(){
        if (typeof(userPreference) !== "undefined"){
            setupIntroJS();
        }
    },
    save: function(){
        localStorage.userPref = JSON.stringify(this.toJSON());
    },
    addPackage: function(def){
        var p = this.get("packages");
        var o = this.get("origPackages");

        if (_.indexOf(p, def) === -1){
            this.set("packages", p);
            loadNodeListAjax(def);
            p.push(def);
        }

        if (_.indexOf(o, def) === -1){
            this.set("origPackages", o);
            o.push(def);
        }
        this.save();
        nodelistSize();
    },
    delPackage: function(def){
        userPreference.set("packages",
            _.without(userPreference.get("packages"), def));
        nodedefs = _.reject(nodedefs, function(nodedef){
            return nodedef.def === def;
        });
        this.save();
        nodelistSize();
    }
});

var NowLoading = Backbone.View.extend({
    el: $(".nodeLoading"),
    initialize: function(){
        this.listenTo(this.model, "change", this.render);
    },
    render: function(){
        // if (_.isNumber(this.model.get("loadStage"))){
        //     this.$el.html("Now loading... " + this.model.get("loadStage") + "%");
        // }
        // else {
        //     this.$el.html("Now loading... " + this.model.get("loadStage"));
        // }
    }
});

var App = Backbone.Model.extend({
    defaults: {
        isRunning: false,
        execstat: "",
        loadStage: 0,
        currentFileName: "",
        clientOS: "",
        serverOS: ""
    }
});

var ConsoleView = Backbone.View.extend({
  el: $("#console"),
  prev: "",
  initialize: function() {
    this.listenTo(this.model, "change", this.render);
    this.$el.draggable({
      cancel: "#executeStatus",
      containment: "body",
      drag: function() {
        userPreference.set("consoleLeft", $(this).css("left"));
        userPreference.set("consoleTop", $(this).css("top"));
      }
    });
    this.$el.zIndex(10);
  },
  render: function() {
    const execStatus = this.model.get("execstat");
    const nonZeroExecCode = (execStatus.indexOf("<nonZeroExecCode>") !== -1) ? true : false;
    const userInterrupt = (execStatus.indexOf("<userInterrupt>") !== -1) ? true : false;
    var newtext = execStatus
      .replace(/</g, "&lt;")
      .replace(/>/g, "&gt;")
      .replace(/&lt;nonZeroExecCode&gt;/g, "")
      .replace(/&lt;\/nonZeroExecCode&gt;/g, "")
      .replace(/&lt;userInterrupt&gt;/g, "")
      .replace(/&lt;\/userInterrupt&gt;/g, "")
      .replace(/&lt;stderr&gt;/g, "<span style='color:#900'>")
      .replace(/&lt;\/stderr&gt;/g, "</span>")
      .replace(/&lt;stdout&gt;/g, "<span style='color:#000'>")
      .replace(/&lt;\/stdout&gt;/g, "</span>")
      .replace(/\n/g, "<br/>")
      .replace(/<br\/><br\/>/g, "<br/>");
    const errorMessage = (nonZeroExecCode || userInterrupt) ? newtext : undefined;
    console.log(newtext);
    if (newtext.length > 100) {
      var tmp = newtext.split("<br/>")[0];
      this.$el
        .find("#executeStatus")
        .html(tmp + "<span style='margin-left:10px;'>see more</span>");
      var seemore = $(this.$el.find("#executeStatus").find("span"));

      seemore.css("text-decoration", "underline");
      seemore.attr(
        "onclick",
        'window.open().document.body.innerHTML=decodeURIComponent("' +
          encodeURIComponent(newtext) +
          '");'
      );
      tmp +=
        "<span style=''" +
        " onclick='window.open().document.body.innerHTML=decodeURI(" +
        encodeURI(newtext) +
        ")'>see more</span>";
      newtext = tmp;
    } else {
      this.$el.find("#executeStatus").html(newtext);
    }

    if (nonZeroExecCode || userInterrupt) {
      $("#executionStatusDialog")
        .find("#errorMessage")
        .html(errorMessage);
      const title = (nonZeroExecCode) ? "Execution Error" : "User Interrupt";
      $("#executionStatusDialog").dialog({"title": title}).dialog("open");
    }

    if (this.prev !== execStatus) {
      if (this.model.get("execstat").indexOf("Error") !== -1) {
        this.$el
          .switchClass("consoleNoerror", "consoleError", 400, "linear")
          .delay(1000)
          .switchClass("consoleError", "consoleNoerror", 400, "linear");
      } else {
        this.$el.switchClass("consoleError", "consoleNoerror", 400, "linear");
      }
    }
    $("#saveFileDialog input[name=name]").val(
      this.model.get("currentFileName")
    );
    $("#saveFilePerSheetDialog input[name=name]").val(
      this.model.get("currentFileName")
    );
    $("#currentFileName").html(this.model.get("currentFileName"));
    this.prev = execStatus;
  }
});

var FileModel = Backbone.Model.extend({
    defaults: {
        type: "",
        name: "",
        date: new Date(),
        object: ""
    }
});

var FileCollection = Backbone.Collection.extend({
    url: '/filelist',
    model: FileModel,
});

// Start pagination
// Initialize pageable collection
var PageableFileCollection = Backbone.PageableCollection.extend({
    url: '/filelist',
    model: FileModel,
    state: {
        pageSize: 5
    },
    mode: "client" // page entirely on the client side
});
var pageableFileCollection = new PageableFileCollection();

// Initialize the paginator
var fileListPaginator = new Backgrid.Extension.Paginator({
    collection: pageableFileCollection
});

// Custom cell
var ActionCell = Backgrid.Cell.extend({
    events: {
        "click": "clickAction"
    },
    clickAction: function (e) {
        e.preventDefault();
    },
    className: "actionButtonSet",
    render: function () {
        var collection = this.model.collection;
        var model = this.model;
        var name = model.get("name");
        var that = this;

        var template = "<button class='delete'>Delete</button>";

        if(this.model.get("type") === 'network') {
            template += "<button class='load'>Load</button>";
        }

        template += "<a class='download'>Download</a>";

        // Renders the template
        this.$el.html(template);
        this.delegateEvents();

        // Converts into buttons and binds event click
        this.$el.find(".delete").button({
            text: false,
            icons: {primary: "ui-icon-closethick"}
        }).click(function () {
            if (confirm("Are you sure to delete file " + name + "?")){
                // Gets active directory
                deleteFile(name);
            }
        });

        this.$el.find(".load").button({
            text: false,
            icons: {primary: "ui-icon-folder-open"}
        }).click(function(){
            if (confirm("Current network will be cleaned. OK?")){
                // Gets active directory
                loadFile(name);
            }
        });

        this.$el.find(".download").button({
            text: false,
            icons: {primary: "ui-icon-arrowthickstop-1-s"}
        }).attr("href", "dl/network/" + encodeURIComponent(name)) // Encodes filename
            .click(function (e) {
                // Gets active directory
                window.open($(this).attr("href"),"_blank");
            });

        return this;
    }
});

// Set up a grid to use the pageable collection
var fileListGrid = new Backgrid.Grid({
    collection: pageableFileCollection,
    columns: [{
        name: "name",
        label: "Name",
        cell: "string",
        editable: "false",
    }, {
        name: "type",
        label: "Type",
        cell: "string",
        editable: "false",
    }, {
        name: "",
        cell: ActionCell,
    }],
    emptyText: "No Records Found",
});

// Initialize a client-side filter to filter on the client
// mode pageable collection's cache.
var fileListFilter = new Backgrid.Extension.ClientSideFilter({
    collection: pageableFileCollection,
    fields: ['name','type']
});
// End pagination

// Comments this since fileListGrid is used
/*var FileView = Backbone.View.extend({
    tagName: "tr",
    className: function(){
        return this.model.get("type");
    },
    events: {
        "mouseover": "onMouseover",
        "mouseout" : "onMouseout",
        "mousedown": function(e){return false;}
    },
    onMouseover: function(e){
        // this.$el.addClass("ui-widget");
        // this.$el.addClass("ui-state-default");
        // this.$el.addClass("ui-corner-all");
    },
    onMouseout: function(e){
        // this.$el.removeClass("ui-widget");
        // this.$el.removeClass("ui-state-default");
        // this.$el.removeClass("ui-corner-all");
    },
    initialize: function(){
        this.listenTo(this.model, "remove", this.remove);
    },
    render: function(){
        var str = "";
        str += "<td><button class='delete' name='<%= name %>'>Delete</button></td>" +
               "<td class='type'><%= type %></td>" +
               "<td class='name'><%= name %></td>";
        //display appropriate buttons for files which doesn't have upload error
        if (this.model.get("type") !== "error"){
            str += "<td class='buttonset'>" +
                "    <a class='download' href='dl/network/<%= name %>' target='_blank'>Download</a>";
        }
        if (this.model.get("type") === "network"){
            str += "    <button class='load' name='<%= name %>'>Load</button>";
        }
        str += "</td>";

        this.$el.append(_.template(str,this.model.toJSON()));
        // console.log(_.template(str,this.model.toJSON()));

        this.$el.find(".download").button({
            text: false,
            icons: {primary: "ui-icon-arrowthickstop-1-s"}
        });
        this.$el.find(".load").button({
            text: false,
            icons: {primary: "ui-icon-folder-open"}
        }).click(function(){
            if (confirm("Current network will be cleaned. OK?")){
                loadFile($(this).attr("name"));
            }
        });
        this.$el.find(".delete").button({
            text: false,
            icons: {primary: "ui-icon-closethick"}
        }).click(function(){
            if (confirm("Are you sure to delete file" + $(this).attr("name") + "?")){
                deleteFile($(this).attr("name"));
            }
        });

        this.$el.find(".load").tooltip({position: {my: 'left center', at: 'right+5 center'}});
        this.$el.find(".download").tooltip({position: {my: 'left center', at: 'right+5 center'}});
        this.$el.find(".delete").tooltip({position: {my: 'right center', at: 'left-5 center'}});

        this.$el.css("list-style-type", "none");
        return this;
    }

});*/
var debug;
var FileManagerView = Backbone.View.extend({
    el: $("#fileManagerDialog"),
    initialize: function(){
        var that = this;
        var table = that.$el.children(".dialogMain");
        var  store = [];
        $("#fileProgress").progressbar();
        this.$el.find(".reload").button({
            text: false,
            icons: {primary: "ui-icon-arrowrefresh-1-s"}
        }).click(function(e){that.$el.find("#datapath").trigger("apply");});

        $('#fileUploadFile').fileupload({
            // dataType: "json",
            url: function(){ return $("#fileUploadFile").attr("data-url");},
            done: function(e, data){
                if (data.textStatus === 'success') {
                    _.each(store, function (f,i) {
                        if (data.files[0].name === f) {
                            store.splice(i,1);
                        }
                    });
                }
                data.context.unbind().remove();
            },
            always: function () {
                if (store.length <= 0) { // Update only once
                    updateFileList();
                }
            },
            progressall: function(e, data){
                var progress = parseInt(data.loaded / data.total * 100, 10);
                if (progress === 100){
                    $('#fileProgress').css("display", "none");
                } else {
                    $('#fileProgress').css("display", "block");
                }
                $("#fileProgress").progressbar("value", progress);
            },
            add: function(e, data){
                // Comments this since PageableFileCollection is used
                //if (_.pluck(fileCollection.toJSON(), "name").indexOf(data.files[0].name) !== -1){
                //    if (!confirm("Overwrite " + data.files[0].name + "?")){
                //        $("#fileManagerLoading").css("display", "none");
                //        return;
                //    }
                //}
                var flag = 0;
                var nameArr = _.pluck(fileCollection.toJSON(), "name");

                function checkType(filename){
                    var ext = filename.split(".").slice( -1 )[0];
                    if (ext === "n"){
                        return "network";
                    }
                    if (_.indexOf(["dat", "tff"], ext) !== -1){
                        return "tf";
                    }
                    if (_.indexOf(["zip"], ext) !== -1){
                        return "zip";
                    }
                    if (_.indexOf(["wav", "raw"], ext) !== -1){
                        return "audio";
                    }
                    return "others";
                }

                // Validates for duplicate file and overwrite files
                if (nameArr.length > 0) {
                    _.each(nameArr, function (n) {
                        if (n.toUpperCase() === data.files[0].name.toUpperCase()) {
                            var cnfrm = confirm("Overwrite "+ n + " with " + data.files[0].name + "?");
                            if (cnfrm) {
                                flag = 1;
                            }
                            else {
                                $("#fileManagerLoading").css("display", "none");
                                flag = 0;
                            }
                        }
                        else {
                            flag = 1;
                        }
                    });
                }
                else {
                    flag = 1;
                }

                if (flag >= 1) {
                    store.push(data.files[0].name);
                    data.context =
                        $(_.template($("#fileUploadTemplate").html(), data).trim())
                            .appendTo("#uploadList");
                    var kind = checkType(data.files[0].name);
                    data.url = "/file/" + kind + "/" + data.files[0].name + "/" + data.files[0].size;
                    data.submit();
                    data.context.unbind().remove();
                }
            },
            dropZone: $("#fileManagerDialog")
        });

        var setDefaultRoot = function(){
            $.ajax({
                url: rootPath + "datapath",
                method: "GET"
            }).done(function(res){
                // Initialize global variable base directory
                base = res;
                $("#datapath").val(base);
                localStorage.setItem('basePath', base);
            });
        };
        // add data path
        if(localStorage.basePath) {
            $.ajax({
                url: '/dirlist/set_dir',
                type: 'POST',
                data: { dir : localStorage.basePath },
            }).done(function (res) {
                base = res.root;
                $("#datapath").val(base);
                localStorage.setItem('basePath', base);
            }).fail(function (res) {
                setDefaultRoot();
            });
        } else {
            setDefaultRoot();
        }

        $("#datapath").on("click", function(){
            this.setSelectionRange(this.value.lastIndexOf("/"), this.value.length);
        });

        // setup dialog
        this.$el.dialog({ autoOpen: false,
                          modal: true,
                          minWidth: 710 });

        $("#fileManagerDialog .dialogMain").on("mousedown", function(){return false;});

        // Icon for opening directory
        this.$el.find("#openDir").button({
            text: false,
            icons: {primary: "ui-icon-arrowreturnthick-1-n"},
        }).blur();

        // Icon for resetting to home directory
        this.$el.find("#homeDir").button({
            text: false,
            icons: {primary: "ui-icon-home"},
        }).blur();

        // Render pagination
        var dialog = this.$el.find(".dialogMain");
        var filter = dialog.find("#filter");
        var grid = dialog.find("#grid");
        var paginator = dialog.find("#paginator");

        // Render the filter
        dialog.before(fileListFilter.render().$el);
        // Render the grid
        grid.append(fileListGrid.render().$el);
        // Render the paginator
        paginator.append(fileListPaginator.render().$el);

        // Comments this since PageableFileCollection is used
        // this.listenTo(this.collection, "all", function(e, b){console.log(e, b.toJSON());});
        // this.listenTo(this.collection, "add", this.addFile)
        // this.listenTo(this.collection, "reset", this.reset);

        // Comments this since PageableFileCollection is used
        // this.listenTo(this.collection, "add", this.update);
        // this.listenTo(this.collection, "reset", this.update);
        // this.listenTo(this.collection, "delete", this.update);

        //override close event of dialog to perform clearing of search query
        this.$el.on("dialogclose", function(){
            $("#fileManagerDialog").find("input[name='q']").val("");
        });
        this.on("close", this.close);
    },
    close: function(){
        this.$el.dialog("close");
    },
    // Comments this since PageableFileCollection is used
   /* arrangeSize: function(type){
        _.each($("#file_" + type), function(d){
            var filenames = $(d).find("span.name");
            if (filenames.length === 0){return;}

            var width = _.chain(filenames)
                         .map(function(s){return $(s).width();})
                         .max().value();
            if (width !== -Infinity){
                _.each(filenames, function(f){$(f).css("width", width + "px");});
            }
        });
    },
    reset: function(resetted, m){
        _.each(m.previousModels, function(m){
            m.trigger("remove");
            m.destroy();
        });
    },
    addFile: function(file){
        var view = (new FileView({model: file})).render();
        $(this.$el.find(".dialogMain")).append(view.el);
    },
    update: function(file){
        var that = this;
        var exists = _.chain(this.collection.toJSON()).groupBy("type").keys().value();
        _.each(this.$el.find(".ui-tabs-anchor"), function(a, i){
            if (_.indexOf(exists, $(a).attr("type")) === -1){
                $("#file_" + $(a).attr("type")).children(".nofile").text("no files");
            } else {
                $("#file_" + $(a).attr("type")).children(".nofile").text("");
            }
        });
    }*/
});

// File tree dialog backbone view
var FileTreeView = Backbone.View.extend ({
    el: $("#fileTreeDialog"),
    initialize: function () {
        // Initialize file tree dialog
        this.$el.dialog({
            autoOpen: false,
            modal: true,
            width: 400,
            height: 600,
            open: function () {
                $('.ui-dialog:button').blur();
                $('.ui-dialog').find("#fileTree").blur();
            },
            buttons: {
                "Select": function () {
                    var cleanpath = function (dir) {
                        dir = dir.replace(/\/|\\/g, '|');
                        dir = _.without(dir.split('|'), "");
                        dir = dir.join('/');
                        dir = dir.trim();
                        return dir;
                    };

                    // Update datapath value
                    var fileTree = $('#fileTree').jstree(true);
                    var datapath = $("#fileManagerDialog").find("#datapath");
                    if (fileTree.get_selected() && fileTree.get_selected().length > 0) {
                        if (!datapath.val()) {
                            datapath.val(fileTree.get_node("/").text);
                        }
                        var dir = datapath.val() + '/' + cleanpath(fileTree.get_selected()[0]);
                        datapath.val(dir);
                        datapath.trigger("apply");
                    }
                    $(this).dialog("close");
                },
                "Cancel": function () {
                    $(this).dialog("close");
                }
            }
        });
        this.$el.prev().find(".ui-dialog-titlebar-close span.ui-button-text").css({padding: 0});

        // Initialize jstree plugin
        this.$el.find("#fileTree").jstree({
            core: {
                'check_callback' : function(o, n, p, i, m) {
                    return true;
                },
                'force_text' : true,
                'themes' : {
                    'responsive' : false,
                    'variant' : 'small',
                },
                'multiple' : false,
            },
            'sort' : function(a, b) {
                return this.get_type(a) === this.get_type(b) ? (this.get_text(a) > this.get_text(b) ? 1 : -1) : (this.get_type(a) >= this.get_type(b) ? 1 : -1);
            },
            'types' : {
                'default' : { 'icon' : 'folder' },
            },
            'unique' : {
                'duplicate' : function (name, counter) {
                    return name + ' ' + counter;
                }
            },
            'plugins' : ['state','sort','types','unique']
        });
    }
});

var SheetsView = Backbone.View.extend({
    el: $("#sheets"),
    initialize: function(){
        this.listenTo(this.model, "change", this.toggleMinimize);
        this.toggleMinimize();
    },
    toggleMinimize: function(event, prev){
        if (this.model.get("nodelistmin")){
            this.$el.css("margin-left", "0px");
        } else {
            $("#sidebar").css("width", "auto");
            $("#sheets").css("margin-left", $("#sidebar").width() + "px");
            $("#sidebar").css("width", $("#sidebar").width() + "px");
            // this.$el.css("margin-left", "200px");
        }
    }
});

var NodelistView = Backbone.View.extend({
    el: $("#sidebar"),
    initialize: function(){
        this.listenTo(this.model, "change", this.toggleMinimize);
        this.toggleMinimize();
    },
    toggleMinimize: function(event){
        if (this.model.get("nodelistmin")){
            this.$el.css("display", "none");
        } else {
            this.$el.css("display", "block");
        }
    }

});
var MinimizeNodelistButton = Backbone.View.extend({
    el: $("#minimizeNodelistButton"),
    events: {
        "click": "onclick"
    },
    initialize: function(){
        this.listenTo(this.model, "change", this.switchIcon);
        this.switchIcon();
    },
    onclick: function(event){
        this.model.set("nodelistmin", !this.model.get("nodelistmin"));
    },
    switchIcon: function(event){
        if (this.model.get("nodelistmin") === false){
            this.$el.button({label: "Hide list", text: false});
        } else {
            this.$el.button({label: "Show list", text: false});
        }
    }

});
var zoomView;
var executeModeDialogView;
var executeButtonView;
var fileManagerView;
var fileTreeView;
var consoleView;
var nowLoadingView;
var sheetsView;
var nodelistView;
var minimizeNodelistButton;
var app;
var userPreference;
var fileCollection;

function setupAllBackbone() {
    app = new App();
    userPreference = new UserPreferenceModel();
    fileCollection     = new FileCollection();
    fileCollection.fetch({
        url: rootPath + "filelist",
        reset: true,
        add: true
    });

    zoomView = new ZoomView({model: userPreference});
    executeModeDialogView = new ExecutionModeDialogView();
    executeErrorDialogView = new ExecutionStatusDialogView();
    executeButtonView = new ExecuteButtonView({model: app});
    fileManagerView = new FileManagerView({collection: fileCollection});
    fileTreeView = new FileTreeView();
    consoleView = new ConsoleView({model: app});
    nowLoadingView = new NowLoading({model: app});
    sheetsView = new SheetsView({model: userPreference});
    nodelistView = new NodelistView({model: userPreference});
    minimizeNodelistButton = new MinimizeNodelistButton({model: userPreference});
}

/* Global variables */
var prevQuery = "";
var network = newNetwork();
var nodedefs = [];
var currentTab = "MAIN";
var currentLinkFrom = false;
var isDraggingStarted = false;
var isSearching = false;
var addSheetDialog;
var base;

var rootPath = $("#rootPath").attr("root") || "/";
var isSelenium = $("#rootPath").attr("selenium");

// delimiter related definitions
// used to connect node_id and terminal name. e.g., node_Synthesize_1[delim]OUTPUT
var delim = "_delim_";
function getBeforeDelim(str){return str.substring(0, str.lastIndexOf(delim));}
function getAfterDelim(str) {return str.substring(str.lastIndexOf(delim) + delim.length);}

$(document).ready(function() {
    // is logged in?
    // if (localStorage.loggedin !== "true"){
    //     window.onbeforeunload = undefined;
    //     alert("you are not logged in");
    //     window.location = "http://localhost:3000/login";
    //     return;
    // }

    if (isSelenium === "right"){
        window.onbeforeunload = function(event){
            event = event || window.event;
            if (event){
                event.returnValue = "Do you really want to change the page?";
            }
            return "Do you really want to change the page?";
        };
    }

    //retrieve find error
    var findError = $("body").attr("findError");
    //remove user preferences if there are errors
    if (findError !== "0") {
        localStorage.clear();
    }

    // jQuery UI widgets setup
    $("#sidebar").accordion({
        active: 1,
        icons: false
    });
    $("#nodelist").menu();
    $("#flatNodelist").menu();
    $("#filelist").menu();
    $(document).on("focus", "#nodelistSearch input", function(){
        $(".nodeSelected").removeClass("nodeSelected");
        $(".noteSelected").removeClass("noteSelected");
    });

    setupAllBackbone();
    $.ajax({
        url: rootPath + "os"
    }).done(function(data){
        app.set("serverOS", data);
    });
    setupSheets();
    setupSheetEvents();
    setupNodeKeyBindings();
    setupSaveFile();
    setupSavePerSheet();
    setupTerminalRoles();
    setupParameterValidation();
    setupButtons();
    setupStatus();
    setupPreference();
    setupFileManager();
    setupFileTree();
    setupIntroJS();
    if (userPreference.get("env")){
        $.ajax({
            url: "/preference/environment",
            method: "PUT",
            data: userPreference.get("env")});
    }
    if ($("#logout").length){
       $("#logout").button().click(function(){
           $.ajax({
                url: "/logout"
            }).always(function(data){
                console.log("logout", data);
                window.location = "/login";
            });
        });
    }

    _.delay(clearCanvas, 400);
    $(".loading").delay(200).fadeOut("slow")
                 .queue(function() {
                     // $(".loading").remove();
                     $("#nodelistSearch input").focus();
                     nowLoadingView.unbind().remove();
                     //display errors in finding def files
                     if (findError === "1") {
                         //for linux
                         app.set("execstat", "Error: 'find' command, for Linux, is not found. Please check the search path on your system.");
                     }
                     else if (findError === "2") {
                         //for windows
                         app.set("execstat", "Error: 'findstr' command, for Windows, is not found. Please check the search path on your system.");
                     }

    if(localStorage.length > 1) localStorage.removeItem("clipboard");

                     if (localStorage.userPref) {
                         //retrieve local storage for user preferences
                         var value = JSON.parse(localStorage.userPref);
                         //validate if all packages are enabled
                         if (value.origPackages && value.packages.length !== value.origPackages.length) {
                             //display warning message
                             app.set("execstat", "Not all packages are loaded. Click Preferences to check this issue.");
                         }
                     }
                 });
    keepalive();
});
var interval = 5000;
function keepalive(){
    var msg = "Server connection failed";
    $.ajax({
        url: rootPath + "keepalive",
        cache: false
    }).done(function(data){
            interval = 5000;
            if (app.get("execstat") === msg){
                app.set("execstat", "");
                app.set("execstat", "Error: Server connection established");
            }
            setTimeout(keepalive, interval);
    }).fail(function(data){
            interval = _.min([interval * 2, 60000]); // max 1min
            app.set("execstat", "");
            app.set("execstat", "Error: " + msg);
            setTimeout(keepalive, interval);
    });
}
var timer = false;
$(window).resize(function(){
    // console.log("resizeing");
    // if (timer !== false){
    //     clearTimeout(timer);
    // }
    // timer = setTimeout(function(){
        updateSheetsSize();
        clearCanvas();
        drawLinks();
    // }, 200);
});

function setupTerminalRoles(){
    $("#terminalRoleNameDialog").dialog({
        autoOpen: false,
        modal: true,
        buttons : {
            OK: function(){
                // var clickedTermId = $(this).attr("clickedTermId");
                // _.find(network.terminalRoles, function(role){
                //     return clickedTermId ===
                //            role.nodeId + delim + role.termName;
                // }).name = $("#terminalRoleName").val();
                newTerminalRole({
                    nodeId  : $(this).attr("clickedNodeId"),
                    termName: $(this).attr("clickedTermName"),
                    name    : $("#terminalRoleName").val(),
                    role    : $(this).attr("role"),
                    sheet   : currentTab,
                    object_type: $(this).attr("ioobj_type"),
                    description: $(this).attr("ioobj_description")
                });
                $(this).dialog("close");
            },
            Cancel: function(){
                $("#terminalRoleName").val("");
                $(this).dialog("close");
            }
        }
    });
    $("#terminalRoleNameDialog").find("form").submit(function(event){
        $($('#terminalRoleNameDialog').siblings('.ui-dialog-buttonpane').find("button")[0]).click();
        event.preventDefault();
        event.stopPropagation();
    });
}

/* netwrok data structure construction */
function reset(){
    currentTab = "MAIN";
    currentLinkFrom = false;
    app.set("currentFileName", "");

    // remove all Dynamic nodedefs
    nodedefs = _.reject(nodedefs, function(def){
        return def.category === "Dynamic";
    });
    app.set("execstat", "");

    // remove all sheets except MAIN
    $("#sheets").tabs({active: 0});
    _.each(_.keys(network.sheets),
        function(sheetname){
            if (sheetname !== "MAIN"){
                removeSheet("sheet_" + sheetname, true);
            }
    });

    // remove all nodes in MAIN
    _.each(network.nodes, function(n){removeNode(n.id, false, true);});

    delete network.notes;

    // remove all notes
    $(".note").remove();

    clearCanvas("canvas_MAIN");
    $("#sheet_" + currentTab).scrollTop(0);
    $("#sheet_" + currentTab).scrollLeft(0);
}

function newNetwork(){
    return {"nodes": [],
            "links": [],
            "terminalRoles": [],
            "sheets": {"MAIN": {name: "MAIN", type: "subnet"}},
            "notes" : []
            };
}

function nodeId2nodename(nodeId){
    var pre = "node_";
    var post = "_";
    return nodeId.substring(nodeId.indexOf(pre) + pre.length,
                            nodeId.lastIndexOf(post));
}

function loadFile(filename){
    $("#fileManagerLoading").css("display", "inline");
    $.ajax({
        url: rootPath + "file/" + encodeURIComponent(filename), // Encodes the filename
        method: "GET",
        dataType: "JSON"
    }).done(function(data){
        if (data.type !== "network"){
            app.set("execstat", "Data type is not network");
            return;
        }
        reset();
        app.set("execstat", "Clearing finished. loading...");
        var result = _parseData(data);

        // produces appropriate message
        if (result.length > 0) {
            var text = "";
            _.each(result, function (r,key) {
                if (key !== result.length-1) text += "'"+ r +"', ";
                else text += "'"+ r +"'";
            });

            app.set("execstat", "Error: Loading finished, but 'subnet_param' of " + text + " have error(s).");
        }
        else app.set("execstat", "Loading finished.");

        //check if data has main sheet
        var sheets = data.object.sheets;
        if(!_.findWhere(sheets, {name: "MAIN"})) {
            data.object.sheets.MAIN =  {
                                                "name": "MAIN",
                                                "type": "subnet"
                                            };
        }

        app.set("execstat", "Loading finished.");
    }).fail(function (data){
        updateFileList();
        $("#fileManagerDialog").dialog("close");
        app.set("execstat", "Loading failed: " + data.responseText);
        console.log(data);
    }).always(function(data){
        $("#fileManagerLoading").css("display", "none");
    });
    function _parseData(data){
        network = data.object;

        pruneNetwork(network);
        app.set("currentFileName", data.name);

        var newsheets = {};

        // add sheet
        _.each(network.sheets, function(sheet){
            if (sheet.name !== "MAIN"){
                addSheetDOM(sheet);
                addDynamicNode(sheet.name);
            }
            newsheets[sheet.name] = sheet;
        });

        network.sheets = newsheets;

        // network pruning
        network.nodes = _.map(network.nodes, function(n){
            var def = _.findObject(nodedefs, "name", n.name) || {inputs: [], outputs: []};
            if (typeof(def.name) === "undefined"){
                console.log("undefined node found:" + JSON.stringify(def));
            }
            n.def = def.def || "Dynamic";
            n.theme = def.theme || "Classic";
            n.category = def.category || "Undefined";
            //remove spaces from sheet name
            n.sheet = n.sheet.replace(/ /g, "");

            // prune inputs/outputs
            n.inputs  = _.uniq(_.union(def.inputs,  n.inputs  || []), function(io){return io.name;});
            n.outputs = _.uniq(_.union(def.outputs, n.outputs || []), function(io){return io.name;});

            // update parameters if exists in the network.
            if (n.def === "Dynamic"){
                n.parameters = n.parameters || [];
            } else if (def.parameters){
                n.parameters = _.map(def.parameters, function(dp){
                    var param = _.find(n.parameters, function(np){return np.name === dp.name;});
                    if (param) {
                        _.extend(param, dp, {value: param.value, type: param.type});
                    } else {
                        param = dp;
                    }
                    return param;
                });
            }
            // prune parameters
            _.each(n.parameters, function(p){
                p.condition = p.condition || "";

                //checking if parameter set exist in list
                if ((p.list !== undefined || p.list !== []) && p.list.length > 0) {
                    if (!_.contains(p.list, p.value)) {
                        p.list.splice(0,0,p.value);
                    }
                }

                p.list = p.list || [];
            });

             return n;
        });

        // network subnet_param validation
        var validateSubnet = function (nodes) {
            var normalN = [],
                dynamicN = [],
                arr = [];

            // stores in array whether normal or dynamic node
            _.each(nodes, function (n) {
                if (n.category !== 'Dynamic') normalN.push(n);
                else dynamicN.push(n);
            });

            // loops and validates 'subnet_param' of normal and dynamic node
            _.each(normalN, function (n) {
                _.each (dynamicN, function (d) {
                    _.each(n.parameters, function (nParam) {
                        _.each (d.parameters, function(dParam) {
                            if ((nParam.value === dParam.value) && (n.sheet === d.sheet) && (nParam.type === 'subnet_param')) arr.push(n.id);
                            if ((dParam.value === nParam.value) && (d.sheet === n.sheet) && (dParam.type === 'subnet_param')) arr.push(d.id);
                        });
                    });
                });
            });

            // returns array of node ids if condition meets
            return arr;
        };

        var subnet = validateSubnet(network.nodes);

        // add subnet_param (follow updateParams())
        if(subnet.length === 0) {	// if no error
            _.each(network.nodes, function(n) {
                subnetParamMaintain(n.id);
            });
        }

        _.each(network.links, function(l){
            var addLinkedIOs = function (nodeId, termName, io){
                _.chain(network.nodes)
                 .filter(function(n){return n.id === nodeId;})
                 .each(function(n){
                    // if links connects nonexistent term, add it to the node
                    if (_.pluck(n[io], "name").indexOf(termName) === -1){
                        n[io].push({
                            description: io + "added by the user",
                            name: termName,
                            type: "any"
                        });
                    }
                 });
             };
            addLinkedIOs(l.fromNodeId, l.fromTermName, "outputs");
            addLinkedIOs(l.toNodeId,   l.toTermName,   "inputs");

            //correct link sheet names
            l.sheet = l.sheet.replace(/ /g, '');
        });

        // terminalRoles to output
        _.each(network.terminalRoles, function(r){
            var addTerminalRolesIO = function (nodeId, termName, io){
                _.chain(network.nodes)
                 .filter(function(n){return n.id === nodeId;})
                 .each(function(n){
                    // if links connects nonexistent term, add it to the node
                    if (_.pluck(n[io], "name").indexOf(termName) === -1){
                        console.log(io, n[io]);
                        n[io].push({
                            description: io + "added by the user",
                            name: termName,
                            type: "any"
                        });
                    }
                });
            };
            if (r.role === "input" || r.role === "output"){
                addTerminalRolesIO(r.nodeId, r.termName, r.role + "s");
            }

            //correct terminalroles sheet names
            r.sheet = r.sheet.replace(/ /g, '');
        });

        // dom
        _.each(network.nodes, function(n){
            addNodeDOM(n);
        });

        // Add dom for note
        _.each(network.notes, function(n){
           addNoteDOM(n);
        });

        // add terminalRoles for non-MAIN sheet
        updateDynamicNodeDOM(network);
        _.chain(network.terminalRoles)
         .reject(function(r){return r.sheet === "MAIN";})
         .each(function(r){
            updateTerminalRoles(r, "add");
        });

        // add terminalRoles for MAIN sheet
        _.chain(network.terminalRoles)
         .filter(function(r){return r.sheet === "MAIN";})
         .each(function(r){
            updateTerminalRoles(r, "add", true);
         });

        _.each(_.keys(network.sheets), function(sheet){
            sheet = sheet.replace(/ /g, "");
            clearCanvas("canvas_" + sheet);
            drawLinks("canvas_" + sheet);
        });

        updateDynamicNodeDOM(network);

        $("#fileManagerLoading").css("display", "none");
        _.chain(network.sheets).keys().each(function(s){adjustScroll(s);});

        //DOM position
        console.log("DOM position update");
        _.chain(network.node).filter(function(n){return n.sheet === currentTab;})
         .each(function(n){$("#" + n.id).css({left: n.pos[0], top: n.pos[1]});});

        _.chain(network.note).filter(function(n){return n.sheet === currentTab;})
            .each(function(n){$("#" + n.noteId).css({left: n.pos[0], top: n.pos[1]});});
        fileManagerView.trigger("close");

        return subnet;
    }
}

function deleteFile(filename){
    $.ajax({
        url: rootPath + "file/" + encodeURIComponent(filename), // Encodes filename
        method: "DELETE"
    }).done(function(data){
        app.set("execstat", "Deleted");
        console.log(data);
        updateFileList();
    });
}

/* setup nodes events */
function addNode(nodename, sheetname){
    // memo: nodecount starts from 1.
    var samenames = _.filter(network.nodes, function(n){
        return n.name === nodename;
    });
    console.log(samenames);

    var nodecount = 1;
    if (samenames.length > 0){
        var maxindex = _.chain(samenames).pluck("index").max().value() + 2;

        nodecount = _.find(_.range(1, maxindex), function(candidate){
            if (_.indexOf(_.pluck(samenames, "index"), candidate) === -1) {
                return true;
            }
        });
    }
    // _.each(network.nodes, function(node){
    //     return node.name == nodename;
    // }).length + 1;
    var nodeId = "node_" + nodename + "_" + nodecount;
    var nodedef = _.findObject(nodedefs, "name", nodename);

    // add to the data structure
    newNode({
        name: nodename,
        id: nodeId,
        sheet: sheetname,
        index: nodecount,
        pos: undefined,
        parameters: JSON.parse(JSON.stringify(nodedef.parameters)), // copy
        inputs: JSON.parse(JSON.stringify(nodedef.inputs)),
        outputs: JSON.parse(JSON.stringify(nodedef.outputs)),
        def: nodedef.def,
        theme: nodedef.theme,
        category: nodedef.category
    });
    var addedNode = _.last(network.nodes);

    return addNodeDOM(addedNode);
}

// Add note DOM
function addNoteDOM(noteobj) {
    var noteId   = noteobj.noteId;
    var sheetname = noteobj.sheet;

    if(noteobj.index === null) {
        noteobj.index = network.notes.indexOf(noteobj);
        noteobj.noteId = "note" + noteobj.index;
        noteId = noteobj.noteId;
    }

    // initialize a position if undefined.
    var margin = 100;
    var offset = 10;
    noteobj.pos = noteobj.pos || ([margin, margin]);
    // move the note if another note exists at the same place
    // limit to 1000 loop (I dont use infinite loop for safety)
    var rejectfunc = function(n){return n.noteId === noteobj.noteId;};
    var findfunc  = function(n){return n.sheet === noteobj.sheet && n.pos[0] === noteobj.pos[0] && n.pos[1] === noteobj.pos[1];};
    for (var i = 0; i < 1000; i++){
        if (_.chain(network.notes).reject(rejectfunc).find(findfunc).value()){
            noteobj.pos[0] += offset;
            noteobj.pos[1] += offset;
        } else {
            break;
        }
    }

    // make sure position % 10 = 0
    noteobj.pos[0] = truncate(noteobj.pos[0]);
    noteobj.pos[1] = truncate(noteobj.pos[1]);

    $("#canvas2_" + sheetname).append(
        _.template($("#noteTemplate").html(), {note: noteobj}));

    var notedom = $("#" + noteId);

    if(noteobj.visible === 0) {
        notedom.css("display", "none");
    }

    notedom.css({left: noteobj.pos[0], top: noteobj.pos[1]});
    notedom.children("textarea").resizable();

    notedom.draggable({
        cancel: "", //deny default(":input")
        containment: "#canvas2_" + sheetname,
        scroll: true,
        grid: [10, 10],
        opacity: 0.8,
        drag: function(e, ui){
            var clickedNoteId = ui.helper.attr("id");
            var prevPos = _.findObject(network.notes, "noteId", clickedNoteId).pos;
            var currPos = ui.position;
            var diff = {top:  prevPos[1] - currPos.top,
                left: prevPos[0] - currPos.left};

            ui.position = {left:truncate(currPos.left), top: truncate(currPos.top)};
            _.each(network.notes, function(n){
                if (n.noteId === clickedNoteId){
                    n.pos = [ui.position.left, ui.position.top];
                }
            });

            _.chain(network.nodes)
                .filter(function(n){return $("#" + n.id).hasClass("nodeSelected");})
                .each(function(n){
                    n.pos[0] = truncate(n.pos[0] - diff.left);
                    n.pos[1] = truncate(n.pos[1] - diff.top);
                    $("#" + n.id).css({left: n.pos[0], top: n.pos[1]});
                });
            _.chain(network.notes)
                .filter(function(n){return $("#" + n.noteId).hasClass("noteSelected");})
                .reject(function(n){return n.noteId === clickedNoteId;})
                .each(function(n){
                    n.pos[0] = truncate(n.pos[0] - diff.left);
                    n.pos[1] = truncate(n.pos[1] - diff.top);
                    $("#" + n.noteId).css({left: n.pos[0], top: n.pos[1]});
                });
        },
        start: function(){
            if (!$(this).hasClass("noteSelected")){
                $(".nodeSelected").removeClass("nodeSelected");
                $(".noteSelected").removeClass("noteSelected");
                $(this).addClass("noteSelected");
            }
        },
    });
}

/**
 * initNote (id, modal) - initialize note dialog
 *
 * @param id - id of note
 * @param modal {boolean} - sets 'modal' to true or false
 * @returns {jQuery|Mixed|HTMLElement}
 */
function initNote (id, modal) {
    // Add note dialog instance
    var $note = $("#textCont_" + id);
    var $prevTextarea = $('#' + id).find('textarea');

    $note.dialog({
        autoOpen: false,
        modal: modal,
        buttons : {
            OK: function () {
                var $this = $(this);
                var $thisTextarea =  $this.find('textarea');
                var textValue = $thisTextarea.val();

                _.findWhere(network.notes,{noteId: id}).text = textValue;

                $prevTextarea.text(textValue);

                $(this).dialog('close');
            },
            Cancel: function () {
                $(this).dialog('close');

                if (modal) {
                    $("#" + id).remove();
                    removeNote(id);
                }
            }
        },
        open: function () {
            var $this = $(this);
            var $thisTextarea =  $this.find('textarea');

            $thisTextarea.val($prevTextarea.val());
        },
        close: function (event) {
            // Remove note dom element on 'escape' button and close button
            if ((event.toElement || event.key) && modal) {
                $("#" + id).remove();
                removeNote(id);
            }
        }
    });

    return $note;
}

function addNodeDOM(nodeobj, noAnimation){
    var nodeId   = nodeobj.id;
    var nodename = nodeobj.name;
    var nodedom, nodeparamdom;
    var sheetname  = nodeobj.sheet;
    var existingTerminalRoles = _.filter(network.terminalRoles, function(r){return r.nodeId === nodeId;});

    // initialize a position if undefined.
    var margin = 100;
    var offset = 10;
    nodeobj.pos = nodeobj.pos || [margin, margin];

    // move the node if another node exists at the same place
    // limit to 1000 loop (I dont use infinite loop for safety)
    var rejectfunc = function(n){return n.id === nodeobj.id;};
    var findfunc  = function(n){return n.sheet === nodeobj.sheet && n.pos[0] === nodeobj.pos[0] && n.pos[1] === nodeobj.pos[1];};
    for (var i = 0; i < 1000; i++){
        if (_.chain(network.nodes).reject(rejectfunc).find(findfunc).value()){
            nodeobj.pos[0] += offset;
            nodeobj.pos[1] += offset;
        } else {
            break;
        }
    }

    // make sure position % 10 = 0
    nodeobj.pos[0] = truncate(nodeobj.pos[0]);
    nodeobj.pos[1] = truncate(nodeobj.pos[1]);

    //add to HTML structure
    $("#canvas2_" + sheetname).append(
        _.template($("#nodeTemplate").html(), {node: nodeobj, delim: delim}));
    nodedom = $("#" + nodeId);

    nodedom.css({left: nodeobj.pos[0], top: nodeobj.pos[1], "display": "none"});

    var nodeIoTemplate = $("#nodeIOTemplate");

    nodedom.children(".interminals").html(
        _.template(nodeIoTemplate.html(), {node: nodeobj, type: "in", term: existingTerminalRoles})
    );
    nodedom.children(".outterminals").html(
        _.template(nodeIoTemplate.html(), {node: nodeobj, type: "out", term: existingTerminalRoles})
    );

    var html = "";
    if (nodeobj.parameters.length === 0){
        html = "<tr><td>No parameters exist.</td></tr>";
    } else {
        html = listNodeParameter(nodeobj.parameters);
    }
    $("#" + nodeId + delim + "parametersForm table").html(html);

    if (noAnimation){
        nodedom.addClass("nodeSize" + userPreference.get("zoom"))
               .css("display", "block");
    } else {
        nodedom.addClass("nodeSize" + userPreference.get("zoom"))
               .fadeIn();
    }

    // register event callbacks
    nodedom.draggable({
        cancel: ".terminalcircle",
        containment: "#canvas2_" + sheetname,
        scroll: true,
        grid: [10, 10],
        opacity: 0.8,
        drag: function(event, ui){
            var clickedNodeId = ui.helper.attr("id");
            var prevPos = _.findObject(network.nodes, "id", clickedNodeId).pos;
            var currPos = ui.position;
            var diff = {top:  prevPos[1] - currPos.top,
                        left: prevPos[0] - currPos.left};

            ui.position = {left:truncate(currPos.left), top: truncate(currPos.top)};
            _.each(network.nodes, function(n){
                if (n.id === clickedNodeId){
                    n.pos = [ui.position.left, ui.position.top];
                }
            });

            _.chain(network.nodes)
                .filter(function(n){return $("#" + n.id).hasClass("nodeSelected");})
                .reject(function(n){return n.id === clickedNodeId;})
                .each(function(n){
                    n.pos[0] = truncate(n.pos[0] - diff.left);
                    n.pos[1] = truncate(n.pos[1] - diff.top);
                    $("#" + n.id).css({left: n.pos[0], top: n.pos[1]});
                });
            _.chain(network.notes)
                .filter(function(n){return $("#" + n.noteId).hasClass("noteSelected");})
                .each(function(n){
                    n.pos[0] = truncate(n.pos[0] - diff.left);
                    n.pos[1] = truncate(n.pos[1] - diff.top);
                    $("#" + n.noteId).css({left: n.pos[0], top: n.pos[1]});
                });
        },
        start: function(){
            $("#nodelist").unbind("keydown");
            if (!$(this).hasClass("nodeSelected")){
                $(".nodeSelected").removeClass("nodeSelected");
                $(".noteSelected").removeClass("noteSelected");
                $(this).addClass("nodeSelected");
            }
            // lighten link on dragging
            clearCanvas();
            drawLinks(undefined, undefined, "#ccc");
        }
    });

    nodeparamdom = $("#" + nodeId + delim + "parameters");
    nodeparamdom.dialog({
        autoOpen: false,
        show: "fade",
        hide: "fade",
        isExpanded: false,
        // height: 0.7 * ($("body").height() - $("#topmenu").height()),
        maxHeight: 0.9 * ($("body").height() - $("#topmenu").height()),
        buttons : {
            OK: function(){
                updateParams($(this));
                // temporary fix prevent Uncaught Error
                //$(this).dialog("close");
                $("#" + nodeId + delim + "parameters").dialog("close");
            },
            Cancel: function(){
                var tgtId = getBeforeDelim($(this).attr("id"));
                var prevParam = jQuery.data(this, "prevParam");
                _.each(network.nodes, function(n){
                    if (n.id === tgtId){
                        n.parameters = prevParam;
                    }
                });
                $(this).dialog("close");
            }
        },
        open: function(){
            var nodeId = getBeforeDelim($(this).attr("id"));
            var nodeobj = _.findObject(network.nodes, "id", nodeId);

            jQuery.data(this, "prevParam", JSON.parse(JSON.stringify(nodeobj.parameters)));

            var html = "";
            if (nodeobj.parameters.length === 0){
                html = "<tr><td>No parameters exist.</td></tr>";
            } else {
                html = listNodeParameter(nodeobj.parameters);
            }
            $(this).find("table").html(html);
            if (!this.isExpanded){
                $(this).dialog("option", "width", nodeparamdom.find("table").width() + 100);
                this.isExpanded = true;
            }
            console.log(nodeparamdom.children(".type"));
            _.each($("#" + nodeId + delim + "parameters .type select"), function(type){
                if ($(type).val() !== "subnet_param"){$(type).trigger("change");}
            });
            _.each($("#" + nodeId + delim + "parameters .value select"), function(v){
                $(v).trigger("change");
            });
            _.each($("#" + nodeId + delim + "parameters value input"), function(v){
                $(v).trigger("change");
            });
        }
    });
    $("#" + nodeId + delim + "parametersForm").submit(function(event){
        updateParams($(this));
        nodeparamdom.dialog("close");
        event.preventDefault();
        event.stopPropagation();
    });

    // force value validation
    nodeparamdom.children(".value select").trigger("change");
    nodeparamdom.tooltip({position: {my: 'left center', at: 'right+5 center'}});

    return nodedom;
}

function listNodeParameter(nodeParams) {
    const paramMap = _.reduce(nodeParams, (res, val) => {
        const { name } = val;
        const [left, right] = name.split(".");
        const category = right ? left : "Node";
        res[category] = res[category] || [];
        res[category].push(val);
        return res;
    },
        {}
    );

    const html = _.reduce(_.keys(paramMap), (res, val) => {
        res += _.template($("#nodeParameterHeader").html(), { nodeHeader: val });
        _.each(paramMap[val], (param) => {
            res += _.template($("#nodeParameterTemplate").html(), { parameter: param });
        });
        return res;
    },
        ""
    );

    return html;
}

function setupParameterValidation(){
    $(document).on("change keyup", ".parameters input", function(event){
        var nodeId = getBeforeDelim($(this).parents("form").attr("id"));
        var paramname = $($(this).parent().siblings()[0]).children("label").text().trim();

        var nodeobj = _.findObject(network.nodes, "id", nodeId);
        var paramobj = JSON.parse(JSON.stringify(_.findObject(nodeobj.parameters, "name", paramname)));
        var oldparam = JSON.parse(JSON.stringify(_.findObject(nodeobj.parameters, "name", paramname)));
        paramobj.value = $(this).val();

        var result = testParameter(paramobj);
        if (result){
            $(this).addClass("error");
            raise(result, paramobj);
        } else {
            $(this).removeClass("error");
        }
    });

    // for list-type parameter values. Check the conditions and hide/show the other parameters.
    $(document).on("change", ".parameters .value select", function(event){
        // get useful info
        var nodeId = getBeforeDelim($(this).parents("form").attr("id"));
        var nodeobj = _.findObject(network.nodes, "id", nodeId);

        // update parameters
        var paramName = $($(this).parent().siblings()[0]).children("label").text().trim();
        _.findObject(nodeobj.parameters, "name", paramName).value = $(this).val();

        // control display = none/table-row
        var objdompair = _.map(nodeobj.parameters, function(p){
            return [p, _.find($("#" + nodeId + delim + "parametersForm tr"), function(tr){
                return $(tr).children(".name").text() === p.name;
            })];
        });

        _.chain(objdompair)
         .filter(function(pair){return pair[0].condition !== "";})
         .each(function(pair){
            var obj = pair[0];
            var dom = $(pair[1]);

            if (!obj){
                alert("ERROR: obj is undefined this should not happen");
                return;
            }
            if (!obj.condition){
                alert("ERROR: obj.condition is undefined this should not happen" + JSON.stringify(obj));
                return;
            }

            var condname = obj.condition.split("=")[0];
            var condvalue = obj.condition.split("=")[1].split(":");

            // hide parameter if condition is not satisfied
            var paramobj = _.findObject(nodeobj.parameters, "name", condname);
            if (paramobj && condvalue.indexOf(paramobj.value) === -1){
                dom.css("display", "none");
            } else {
                dom.css("display", "table-row");
            }
        });
    });
}

function updateParams(formobj){
    var nodeId = getBeforeDelim(formobj.attr("id"));
    var nodeobj = _.findObject(network.nodes, "id", nodeId);
    var sheet = nodeobj.sheet;
    var errNode = {};
    var errStore = [];
    var categorizeErr = function () { // Stores error in object
        errNode = {};
        _.each(errStore, function (n) {
            if (!_.has(errNode, n.id)) {
                errNode[n.id] = [];
                errNode[n.id].push({type: n.type, paramList: []});
            }
            else {
                _.each(errNode[n.id], function (p) {
                    if (p.type !== n.type) {
                        errNode[n.id].push({type: n.type, paramList: []});
                    }
                });
            }

            if (errNode[n.id].length > 0) {
                _.each(errNode[n.id], function(p) {
                    if (n.type === p.type) {
                        p.paramList.push(n.name);
                    }
                });
            }
        });
    };

    _.each(nodeobj.parameters, function(p){
        var prevVal = p.value;
        p.value = $("#" + nodeId + delim + "parametersForm [name='" + p.name + "']").val();
        // Change if..else statement to switch
        switch (p.type) {
            case "file":
                p.type = "string";
                break;
            case "float": // Add validation for exponential input in float
                var regex = new RegExp(/(^(\+|\-)?(\.)?(\d+\.\d+|\d+)\e(\+|\-)?\d+$)|(^(\+|\-)?(\.)?\d+(\.\d+)?)$/);
                if (!regex.test(p.value) && !_.isEmpty(p.value)) {
                    p.value = prevVal;
                    errStore.push({type: p.type, id: nodeobj.id, name: p.name});
                }
                break;
            case "subnet_param":
                if (nodeobj.category !== 'Dynamic') { // Normal node
                    if (!checkSubnetName(p.value, sheet)) {
                        errStore.push({type: p.type, id: nodeobj.id, name: p.name});
                    }
                }
                else { // Dynamic node
                    if (!checkSubnetName(p.value, sheet, true)) {
                        errStore.push({type: p.type, id: nodeobj.id, name: p.name});
                    }
                }
                break;
        }
    });

    if (errStore.length <= 0) { // Add a condition to determine whether to update the node or not
        subnetParamMaintain(nodeId);
        updateDynamicNodeDOM(network, true);
    }
    else {
        categorizeErr();
        var errMsg = "";
        _.each(errNode, function (val, key) { // Processes errors stored in object to display appropriate error messages
            _.each(val, function (p) {
                if (p.type === "float") {
                    errMsg += "Error: Parameter value(s): "+ p.paramList.join(", ") +" in "+ key +" is not float nor exponential value \n";
                }
                if (p.type === "subnet_param") {
                    errMsg += "Error: Parameter value(s): "+ p.paramList.join(", ") +" in "+ key +" already exists \n";
                }
            });
        });

        app.set('execstat', errMsg);
    }
}

/**
 * checkSubnetName()
 *      validates 'subnet_param' input from normal or dynamic node
 *      that meets the condition of both respectively
 *
 * @param subnet - inputted 'subnet_param' value
 * @param sheet - sheet name of where the node is associated node
 * @param isDynamic - flag parameter to distinguish normal or dynamic node input
 * @returns {boolean}
 */
function checkSubnetName (subnet, sheet, isDynamic) {
    // sets default
    isDynamic = typeof isDynamic !== 'undefined' ? isDynamic : false;
    sheet = typeof sheet !== 'undefined' ? sheet : "MAIN";

    var temp = [];
    var nodes = network.nodes;
    var count = 0;

    // normal node 'subnet_param' input, stores dynamic node 'subnet_param' name
    if (!isDynamic) {
        var dNodes = _.where(nodes, {category: "Dynamic", sheet: sheet});
        _.each(dNodes, function (d) {
            var dNode = _.where(d.parameters, {type: "subnet_param"});
            _.each(dNode, function (p) {
                temp.push(p.value);
            });
        });
    }
    else { // dynamic node 'subnet_param' input, stores normal node 'subnet_param' value
        var nNodes = _.filter(nodes, function (obj) {
            if ((obj.category !== 'Dynamic') && (obj.sheet === sheet)) return obj;
        });

        _.each(nNodes, function (n) {
            var nNode = _.where(n.parameters, {type: "subnet_param"});
            _.each(nNode, function (p) {
                temp.push(p.value);
            });
        });
    }

    //  loops through and validates value
    _.each(temp, function (name) {
        if (name.trim().toUpperCase() === subnet.trim().toUpperCase()) count++;
    });

    // returns appropriate boolean. true if has no error, else false
    return count === 0 ? true : false;
}

function removeNode(nodeId, DOMonly, noAnimation, calledFromUpdateDynamicNodeDom){
    if (!DOMonly){
        _.chain(network.terminalRoles)
         .filter(function(r){return r.nodeId === nodeId;})
         .each(function(r){
            updateTerminalRoles(r, "del");
        });

        network.nodes = _.reject(network.nodes, function(node){
            return nodeId === node.id;
        });

        network.links = _.reject(network.links, function(link){
            return nodeId === link.fromNodeId |
                   nodeId === link.toNodeId;
        });
    }

    function _destroy(nodeId){

        //remove node DOM
        $("#" + nodeId).draggable("destroy");
        _.chain($(".node")).filter(function(n){return $(n).attr("id") === nodeId;})
         .each(function(n){$(n).unbind().remove();});

        //remove parameter dialog
        $("#" + nodeId + delim + "parameters").dialog("destroy");
        $("#" + nodeId + delim + "parameters").unbind().remove();

        removeNonExistingSubnetParam();

        if (!DOMonly){
            clearCanvas();
            drawLinks();
        }
    }
    var nodeObj = $("#" + nodeId);
    if (!noAnimation && nodeObj.css("display") === "block"){
        var imitation = nodeObj.clone(false)
            .addClass("nodeDeleted").removeClass("node").removeClass("nodeSelected").attr("id", nodeId + "_deleted");
        imitation.insertAfter(nodeObj);
        _destroy(nodeId);
        imitation.fadeOut(function(){$(this).remove();});
        adjustScroll();
    } else {
        _destroy(nodeId);
    }
}

/**
 * removeNote(id) - removes note dom and note obj and destroy note dialog instance
 *
 * @param id - id of note dialog instance
 */
function removeNote (id) {
    // Remove note obj
    network.notes = _.reject(network.notes, function(c){
        return c.noteId === id;
    });

    // Remove note dom
    $("#" + id).remove();

    // Remove dialog instance
    $("#textCont_" + id).dialog('destroy');
}

function addCanvasLayer() {
    var id = "canvas_" + currentTab + delim + "layer1";
    var canvas = document.getElementById(id);
    if(canvas) {
        var ctx = canvas.getContext("2d");
        ctx.clearRect(0, 0, canvas.width, canvas.height);
    } else {
        canvas = $("<canvas class='sheet' id='" + id + "' style='pointer-events: none;' />");
        $("#sheet_" + currentTab).append(canvas);
        canvas.ready(function() {
            canvas.attr("height", canvas.css("height"));
            canvas.attr("width", canvas.css("width"));
        });
    }
    return id;
}

function removeCanvasLayer() {
    $("canvas[id$='" + delim + "layer1']").remove();
}

function getSelectedObjs() {
    var activeSheet = getActiveSheet();

    var selectedNodeObjs = _.map($("#sheet_" + activeSheet).find(".nodeSelected"), function(n){
        var id = $(n).children(".nodeid").text().trim();
        return _.findObject(network.nodes, "id", id);
    });

    var selectedLinks = _.filter(network.links, function(l){
        var nodeIds = _.pluck(selectedNodeObjs, "id");
        //console.log(nodeIds);
        return nodeIds.indexOf(l.fromNodeId) !== -1 && nodeIds.indexOf(l.toNodeId) !== -1;
    });

    var selectedNoteObjs = _.map($("#sheet_" + activeSheet).find(".noteSelected"), function(n){
        var name = $(n).attr("id");
        return _.findObject(network.notes, "noteId", name);
    });

    return {nodes: selectedNodeObjs, links: selectedLinks, notes: selectedNoteObjs};
}

function addObjs(objs, sheet, offset) {
    if(!sheet) { sheet = getActiveSheet(); }
    if(!offset) { offset = {}; }
    if (!offset.left) { offset.left = 0; }
    if (!offset.top) { offset.top = 0; }

    var newObjs = JSON.parse(JSON.stringify(objs));
    _.each(newObjs.nodes, function (add) {
        var oldid = add.id;

        var newnode = JSON.parse(JSON.stringify(add));
        var nodecount = _.filter(network.nodes, function (node) {
                return node.name == newnode.name;
            }).length + 1;

        nodecount = _.find(_.range(nodecount), function (c) {
                if (_.find(network.nodes, function (n) {
                        console.log(n.id, "node_" + newnode.name + "_" + (c + 1));
                        return n.id === "node_" + newnode.name + "_" + (c + 1);
                    }) === undefined) {
                    return true;
                } else {
                    return false;
                }
            }) + 1;
        newnode.id = "node_" + newnode.name + "_" + nodecount;
        newnode.index = nodecount;
        newnode.sheet = sheet;

        newObjs.links = _.map(newObjs.links, function (l) {
            if (l.fromNodeId === oldid) {
                l.fromNodeId = newnode.id;
            }
            if (l.toNodeId === oldid) {
                l.toNodeId = newnode.id;
            }
            return l;
        });

        newnode.pos[0] += offset.left;
        newnode.pos[1] += offset.top;

        network.nodes.push(newnode);
        addNodeDOM(newnode);
        $("#" + newnode.id).addClass("nodeSelected");
    });

    _.each(newObjs.notes, function (n) {
        var newnote = JSON.parse(JSON.stringify(n));
        var index = 0;
        var noteId;

        if (_.has(network, "notes") === false) {
            network.notes = [];
            noteId = "note" + index;
        } else {
            if (_.last(network.notes) === undefined) {
                index = 0;
            } else {
                var lastIndex = _.max(_.pluck(network.notes, "index"));
                index = lastIndex + 1;
            }
            noteId = "note" + index;
        }

        newnote.index = index;
        newnote.noteId = noteId;
        newnote.sheet = sheet;

        newnote.pos[0] += offset.left;
        newnote.pos[1] += offset.top;

        network.notes.push(newnote);
        addNoteDOM(newnote);
        $("#" + newnote.noteId).addClass("noteSelected");
    });

    _.each(newObjs.links, function (l) {
        l.sheet = sheet;
        network.links.push(l);
    });

    drawLinks();
}

function pasteClipboard(toPos) {
    $(".node").removeClass("nodeSelected");
    $(".note").removeClass("noteSelected");

    var clipboard;
    try {
        clipboard = JSON.parse(localStorage.getItem('clipboard'));
    } catch (e) {
        console.log(e);
        return;
    }
    if(!clipboard) {
        return;
    }


    var activesheet = getActiveSheet();
    var clipboardPosArray = [];
    _.each(clipboard.nodes, function(n){
        clipboardPosArray.push(n.pos);
    });
    _.each(clipboard.notes, function(n){
        clipboardPosArray.push(n.pos);
    });

    var clipboardPosList = _.map(clipboardPosArray, function(n) { return {left: n[0], top: n[1]};});
    var leftest = _.min(_.pluck(clipboardPosList, "left"));
    var toppest = _.min(_.pluck(clipboardPosList, "top"));
    
    var offset = {left:0, top:0};
    if (toPos && toPos.left && toPos.top) {
        // Pasteの動作
        // 貼り付け先指定なし(Ctrl+V): コピー元と同じ座標に貼り付ける
        // 貼り付け先指定あり(ContextMenuから): クリップボードのオブジェクトの左上を指定座標に合わせる(ここでの処理)
        offset.left = truncate(toPos.left) - leftest;
        offset.top = truncate(toPos.top) - toppest;
        leftest = truncate(toPos.left);
        toppest = truncate(toPos.top);
    }
    addObjs(clipboard, activesheet, offset);

    adjustScroll(activesheet, function(n){return n.hasClass("nodeSelected") || n.hasClass("noteSelected");});
}

function copyObjsToClipboard(objs) {
    localStorage.setItem('clipboard', JSON.stringify(objs));
}

function deleteObjs(objs) {
    _.each(objs.nodes, function(selectedNodeObj){
        removeNode(selectedNodeObj.id);
    });
    _.each(objs.notes, function(n){
        removeNote(n.noteId);
    });
}

function cutObjsToClipboard(objs) {
    $("#nodelistSearch input").blur();
    copyObjsToClipboard(objs);
    deleteObjs(objs);
}

function setupSheetEvents() {
    // parameter change
    $(document).on("change", ".parameters td.type select", function(event){
        var nodeId = getBeforeDelim($($(this).parents("div")[0]).attr("id"));
        var paramName = $(this).attr("name").replace("_type", "");

        var nodeobj = _.findObject(network.nodes, "id", nodeId);
        var paramobj = _.findObject(nodeobj.parameters, "name", paramName);
        var newtype = $(this).val();
        var oldtype = paramobj.type;
        var oldvalue = paramobj.value;

        // if no types changed, just quit.
        paramobj.type = newtype;
        if (newtype === "subnet_param"){
            paramobj.value = paramName;
        }

        //recording of original type and original list
        if (paramobj.originalType === undefined) {
            paramobj.originalType = oldtype;
            paramobj.originalList = paramobj.list;
        }
        if (newtype !== paramobj.originalType) {
            paramobj.list = [];
        }
        else {
            paramobj.list = paramobj.originalList;
        }

        // update parameter
        var nodeIndex = _.indexOf(network.nodes, nodeobj);
        var paramIndex = _.indexOf(nodeobj.parameters, paramobj);
        network.nodes[nodeIndex].parameters[paramIndex] = JSON.parse(JSON.stringify(paramobj));

        // check if selected parameter type is invalid - for Parameters with '.' only
        if (paramobj.name.includes('.')) {
            testSelection(paramobj, newtype);            
        } 

        // // rendering: keep old values
        // var renderParams = JSON.parse(JSON.stringify(nodeobj.parameters));
        // var funcRenderParams = function(formValue){
        //     _.find(renderParams, function(p){
        //         if (p.name === $(formValue).attr("name"))
        //             p.value = $(formValue).val();
        //     });
        // };
        // _.each($(".paramDialog .value input"), funcRenderParams);
        // _.each($(".paramDialog .value select"), funcRenderParams);
        // renderParams[paramIndex] = JSON.parse(JSON.stringify(paramobj));
        $(this).parentsUntil("tr").parent().html(
            _.template($("#nodeParameterTemplate").html(), {parameter: paramobj})
             .replace("<tr>", "").replace("</tr>", "")
        );

        // if type === select, update options
        if (newtype === "file"){
            var node = $("#" + nodeId + delim + "parameters" + " .value select[name=" + paramName + "]");
            node.html("");
            _.chain(fileCollection.toJSON())
             .groupBy("type").map(function(gr){return _.sortBy(gr, "name");}).flatten()
             .each(function(f){
                var name = f.name;
                if (oldvalue === name){
                    node.append("<option value='" + name + "' selected>" + name + "</option>");
                } else {
                    node.append("<option value='" + name + "'>" + name + "</option>");
                }
            });
            if (fileCollection.toJSON().length === 0){
                node.append("<option value='no files found' selected>no files found</option>");
            }
        }

        $(".parameters .value select").trigger("change");
    });

    // add node from menu
    $(document).on("click", ".nodeElement", function(event){
        var nodeObj = addNode($(this).text(), currentTab);
        adjustScroll(currentTab, function(n){return n.attr("id") === nodeObj.attr("id");});
    });

    $(document).on("mousedown", ".terminalcircle", function(event){
        if (currentLinkFrom === false){
            currentLinkFrom = $(this).closest(".terminal");
        }
        isDraggingStarted = false;
        $("#sheet_" + currentTab).selectable("destroy");
        event.stopPropagation();
    });
    $(document).on("mouseup", "body", function(event){
        currentLinkFrom = false;
        removeCanvasLayer();
        setupSelectable(currentTab);
        clearCanvas();
        drawLinks();
    });
    $(document).on("mouseup", ".terminal", function(event){
        if (currentLinkFrom === false){
            return;
        }
        if (isDraggingStarted === false){
            return;
        }

        // connect a link
        var fromTerminal = (currentLinkFrom.parent().hasClass("interminals")) ? "interminals" : "outterminals";
        if (fromTerminal === "outterminals"){
            newLink({
                fromNodeId: currentLinkFrom.parents(".node").attr("id"),
                fromTermName: currentLinkFrom.attr("name"),
                toNodeId: $(this).parents(".node").attr("id"),
                toTermName: $(this).attr("name"),
                sheet: currentTab
            });
        } else {
            newLink({
                toNodeId: currentLinkFrom.parents(".node").attr("id"),
                toTermName: currentLinkFrom.attr("name"),
                fromNodeId: $(this).parents(".node").attr("id"),
                fromTermName: $(this).attr("name"),
                sheet: currentTab
            });
        }
        currentLinkFrom = false;
        isDraggingStarted = false;
        removeCanvasLayer();
        clearCanvas();
        drawLinks();
    });

    $(document).on("mousemove", ".sheetContainer div.sheet", function(event){
        if (currentLinkFrom !== false){
            //clearCanvas();
            //drawLinks();
            var canvasid = addCanvasLayer();
            if(userPreference.get("showCandidate") === "Yes"){
                drawCandidateLinks(canvasid);
            }
            drawCurrentLine(canvasid, event);
            isDraggingStarted = true;
        }
    });

    $(document).on("dblclick", ".node", function(event){
        fileCollection.fetch({reset:true}); // Fetches data without pagination
        var clickedNodeId = $(this).attr("id");
        $("#" + clickedNodeId + delim + "parameters").dialog("open");
        //event.preventDefault();
        event.stopPropagation();
    });

    $(document).on("click", ".node", function(event) {
        if (event.ctrlKey) {
            // toggle selected
            if ($(this).hasClass("nodeSelected")) {
                $(this).removeClass("nodeSelected");
            } else {
                $(this).addClass("nodeSelected");
            }
        } else if (!$(this).hasClass("nodeSelected")) {
            $(".nodeSelected").removeClass("nodeSelected");
            $(this).addClass("nodeSelected");
            $(".noteSelected").removeClass("noteSelected");
        }
        //event.preventDefault();
        event.stopPropagation();
    });

    $(document).on("mouseover", ".note", function(event){
        $(this).addClass("noteHover");
    });

    $(document).on("click", ".note", function(event){
        if (event.ctrlKey) {
            // toggle selected
            if ($(this).hasClass("noteSelected")) {
                $(this).removeClass("noteSelected");
            } else {
                $(this).addClass("noteSelected");
            }
        } else if (!$(this).hasClass("noteSelected")) {
            $(".nodeSelected").removeClass("nodeSelected");
            $(".noteSelected").removeClass("noteSelected");
            $(this).addClass("noteSelected");
            //$(".noteDrag").removeClass("noteDrag");
            //$(this).addClass("noteDrag");
        }
        //event.preventDefault();
        event.stopPropagation();
    });

    $(document).on("dblclick", ".note", function(event){
        var clickedNodeId = $(this).attr("id");
        var $note = initNote(clickedNodeId, false);

        $note.dialog('open');

        event.stopPropagation();
    });

    // context menu for sheetTab
    var selected = "";
    $("#sheets").contextMenu({
        selector: ".sheetTab",
        build: function($trigger, event){
            var sheetname_ = $trigger.find("a").attr("href").replace("#sheet_", "");
            var types = ["Change the name", "Change to subnet", "Change to iterator", "Delete the sheet"/*, "Save the sheet"*/];
            var items = {};
            _.each(types, function(type){
                items[type] = {name: type, icon: "edit"};
            });
            items["Delete the sheet"] = {name: "Delete the sheet", icon: "delete"};
            //items["Save the sheet"] = {name: "Save the sheet", icon: "save"};

            _.each(_.keys(items), function(key){
                if (key.indexOf(network.sheets[sheetname_].type) !== -1){
                    delete items[key];
                }
            });

            if ($($trigger).attr("aria-controls") === "sheet_MAIN"){
                return {};
            } else {
                return {
                    callback: function(key, options){
                        var sheetname = $trigger.find("a").attr("href");
                        sheetname = sheetname.substr(sheetname.indexOf("#sheet_") + "#sheet_".length);
                        if (key === "Change the name"){
                            selected = sheetname;
                            $("#sheetTabDialog").dialog("open");
                        } else if (key === "Delete the sheet"){
                           if (confirm("Are you sure to delete " + sheetname + "?")){
                               removeSheet("sheet_" + sheetname, false, function(){
                                   //click the last sheet
                                   var sheetTabs = $(".sheetTab");
                                   var a = $(sheetTabs[sheetTabs.length - 1]).find("a");
                                   $(a).click();
                               });
                           }
                        } else if (key === "Save the sheet"){
                            setupSavePerSheet(sheetname_);
                            $("#saveFilePerSheetDialog").dialog("open");
                        } else {
                            network.sheets[sheetname].type = key.replace("Change to ", "");
                            $(".sheetTab[aria-controls='sheet_" + sheetname + "']").attr(
                                "title", 'Sheet type: ' + network.sheets[sheetname].type);
                            $(".sheetTab[aria-controls='sheet_" + sheetname + "'] a")
                                .removeClass("sheetType_iterator")
                                .removeClass("sheetType_subnet")
                                .addClass("sheetType_" + network.sheets[sheetname].type);
                            }
                    },
                    items: items
                };
            }
        }
    });

    $("#sheetTabDialogForm").submit(function(event){
        $($("#sheetTabDialog").siblings(".ui-dialog-buttonpane").find(".ui-button")[0]).trigger("click");
        event.preventDefault();
        event.stopPropagation();
    });
    $("#sheetTabDialog").dialog({
        autoOpen: false,
        modal: true,
        buttons :  {
            "OK": function(){
                var newname = $(this).find("input").val();
                if (newname === ""){
                    app.set("execstat", "Error sheet name cannot be empty");
                    return;
                }

                if (_.keys(network.sheets).indexOf(newname) !== -1){
                    app.set("execstat", "Error: Sheet name is duplicated");
                    return;
                }

                if(newname.match(/[^A-Za-z0-9\-_\ ]/)) {
                    app.set("execstat", "Error: Sheet name contains invalid characters.");
                    return;
                }


                // rename node.sheet and dynamic nodes
                _.each(network.nodes, function(n){if (n.sheet === selected){n.sheet = newname;}});
                _.each(network.nodes, function(n){
                    if (n.name === selected){
                        var oldid = n.id;
                        n.id = "node_" + newname + "_" + n.index;
                        n.name = newname;

                        $("#" + oldid + " .nodename").text(newname);
                        $("#" + oldid + " .nodeid").text(n.id);
                        _.each($("#" + oldid + " .terminal"), function(t){
                            $(t).attr("id", n.id + delim + getAfterDelim($(t).attr("id")));
                        });
                        _.each($("[aria-describedby=" + oldid + delim + "parameters]"), function(d){
                            $(d).attr("aria-describedby", n.id + delim + "parameters");
                            $(d).find(".ui-dialog-title").text("Parameters of " + n.id);
                            $(d).find("#" + oldid + delim + "parameters").attr("id", n.id + delim + "parameters");
                            $(d).find("#" + oldid + delim + "parametersForm").attr("id", n.id + delim + "parametersForm");
                        });
                        $("[aria-controls=sheet_" + selected + "]").attr("aria-controls", "sheet_" + newname);
                        $("#" + oldid).attr("id", n.id);
                        $("#" + oldid).draggable("refresh");
                    }
                });

                //rename links
                _.each(network.links, function(l){if (l.sheet === selected){l.sheet = newname;}});
                _.each(network.links, function(l){
                    var regexp = "node_" + selected;
                    if (l.fromNodeId.match(regexp)){
                        l.fromNodeId = "node_" + newname + "_" + l.fromNodeId.split("_").slice( -1 );
                    }
                    if (l.toNodeId.match(regexp)){
                        l.toNodeId = "node_" + newname + "_" + l.toNodeId.split("_").slice( -1 );
                    }
                });

                //rename terminalRoles
                _.each(network.terminalRoles, function(l){if (l.sheet === selected){l.sheet = newname;}});
                _.each(network.terminalRoles, function(l){
                    var regexp = "node_" + selected;
                    if (l.nodeId.match(regexp)){
                        l.nodeId = "node_" + newname + "_" + l.nodeId.split("_").slice( -1 );
                    }
                });

                //rename Sheets
                network.sheets[newname] = {
                    name: newname,
                    type: network.sheets[selected].type
                };
                delete network.sheets[selected];

                //rename nodedefs
                _.find(nodedefs,function(d){return d.name === selected;}).name = newname;

                $("#sheet_" + selected).attr("id", "sheet_" + newname);
                $("#canvas_" + selected).attr("id", "canvas_" + newname);
                $("a[href=#sheet_" + selected + "]").text(newname);
                $("a[href=#sheet_" + selected + "]").attr("href", "#sheet_" + newname);
                if (currentTab === selected) {currentTab = newname;}
                $("#sheets").tabs("refresh");

                updateNodeList();
                sheetsView.toggleMinimize();
                $(this).dialog("close");

            },
            "Cancel": function(){
                $(this).dialog("close");
            }
        }
    });

    $("#setProcessDialog").dialog({
        title: "Process name",
        autoOpen: false,
        buttons: [{
            text: "OK",
            click: function(event){
                var clickedNodeId = $("#setProcessDialog").attr("isOneclick");
                var nodes = (clickedNodeId) ? $("#" + clickedNodeId) : $(".nodeSelected");
                var process = $("#setProcessName").val();

                _.each(nodes, function(node){
                    _.each(network.nodes, function(n){
                        if($(node).attr("id") === n.id){
                            n.process = process;
                        }
                    });
                });
                $(this).dialog("close");}
        },
        {
            text: "Cancel",
            click: function(event){$(this).dialog("close");}
        }]
    });

    // context menu for sheet
    $("#sheets").contextMenu({
        selector: ".sheetContainer div.sheet",
        reposition: false,
        build: function($trigger, event){
            // make contextmenu items
            // common menu items for sheet
            var items = {
                "cut": {name: "Cut", icon: "cut"},
                "copy": {name: "Copy", icon: "copy"},
                "paste": {
                    name: "Paste", icon: "paste", disabled: function (key, opt) {
                        return localStorage.getItem('clipboard') === null;
                    }
                },
                "sep1": "---------",
                "delete": {name: "Delete", icon: "delete"},
                "sep2": "---------",
                "createSheet": {
                    name: "Create sheet from selected", icon: "add", disabled: function (key, opt) {
                        return $(".nodeSelected, .noteSelected").length === 0;
                    }
                },
            };
            this.onObject = event.target.closest(".node, .note");
            var onObjectType = "sheet";
            if (this.onObject) {
                this.onObject = $(this.onObject);
                if (this.onObject.hasClass("node")) {
                    onObjectType = "node";
                    this.onObject.addClass("nodeSelected");
                } else if (this.onObject.hasClass("note")) {
                    onObjectType = "note";
                    this.onObject.addClass("noteSelected");
                }
            }
            // menu items for object
            if (onObjectType === "node" || onObjectType === "note") {
                _.extend(items, {
                    "sep3": "---------",
                    "property": {name: "Property", icon: "edit"}
                });
            }
            // menu items for node
            if (onObjectType === "node") {
                _.extend(items, {
                    "sep4": "---------",
                    "addInput":  {name: "Add Input", icon: "add"},
                    "addOutput":  {name: "Add Output", icon: "add"},
                    "showHelp" : {name: "Help", icon: "help"},
                    "theme": {name: "Set theme",
                        "items": {
                            "theme-Fire": {"name": "Fire"},
                            "theme-Sea" : {"name": "Sea" },
                            "theme-Dark": {"name": "Dark"},
                            "theme-Classic": {"name": "Classic"}
                        }
                    }
                });
            }
            // menu items for sheet
            if (onObjectType === "sheet") {
                _.extend(items, {
                    "sep5": "---------",
                    "addNote": {name: "Add Note", icon: "add"}
                });
            }
            return {
                callback: function(key, options) {
                    var selected = getSelectedObjs();
                    var clickedObjId = options.onObject && options.onObject.attr("id");
                    var cursorOffset = options.$menu.offset();
                    var canvasOffset = $("#canvas2_" + getActiveSheet()).offset();
                    var cursorPos = {"left": truncate(cursorOffset.left - canvasOffset.left),
                        "top": truncate(cursorOffset.top - canvasOffset.top)};
                    switch (key) {
                        case "cut":
                            cutObjsToClipboard(selected);
                            break;
                        case "copy":
                            copyObjsToClipboard(selected);
                            break;
                        case "paste":
                            pasteClipboard(cursorPos);
                            break;
                        case "delete":
                            deleteObjs(selected);
                            break;
                        case "createSheet":
                            addSheetDialog.dialog("option", "nodesForCreateSheet", selected);
                            addSheetDialog.dialog("open");
                            break;
                        case "property":
                            if (options.onObject && options.onObject.hasClass("node")) {
                                $("#" + clickedObjId + delim + "parameters").dialog("open");
                            } else if (options.onObject && options.onObject.hasClass("note")) {
                                $("#" + clickedObjId + " textarea").trigger("dblclick");
                            }
                            break;
                        case "addInput":
                            $("#AddIODialog").attr("clickedNodeId", clickedObjId)
                                .attr("title", "Add Input")
                                .dialog("option", "title", "Add Input")
                                .dialog("open");
                            break;
                        case "addOutput":
                            $("#AddIODialog").attr("clickedNodeId", clickedObjId)
                                .attr("title", "Add Output")
                                .dialog("option", "title", "Add Output")
                                .dialog("open");
                            break;
                        case "showHelp":
                            var clickedNodeObj = _.findObject(network.nodes, "id", clickedObjId);
                            var clickedNodeName = nodeId2nodename(clickedObjId);
                            var clickedNodeCat  = nodeId2category(clickedObjId);
                            var clickedNodeDefName = clickedNodeObj.def;
                            var urltmpl = "http://www.hark.jp/document/hark-document-LANG/subsec-REPLACE.html";
                            if (clickedNodeCat.indexOf("HARK") !== -1) {
                                urltmpl = urltmpl.replace("REPLACE", clickedNodeName)
                                    .replace("LANG", userPreference.get("lang"));
                            } else if (_.contains(defHelpLUT, clickedNodeDefName)){
                                urltmpl = "http://flowdesigner.sourceforge.net/doc/nodes.html#" + clickedNodeName;
                            } else {
                                alert("No help for this node.");
                                break;
                            }
                            //show confirmation if user wants to redirect to external help site
                            if(confirm("Hark Designer is opening the web page of " + urltmpl + ". OK?")){
                                window.open(urltmpl,"Help: " + clickedNodeName);
                            }
                            break;
                        case "theme-Fire":
                        case "theme-Sea":
                        case "theme-Dark":
                        case "theme-Classic":
                            var theme = key.slice(6);
                            var nodes = $(".nodeSelected");
                            _.each(nodes, function(n){
                                $(n).removeClass("nodeThemeFire")
                                    .removeClass("nodeThemeSea")
                                    .removeClass("nodeThemeDark")
                                    .removeClass("nodeThemeClassic")
                                    .addClass("nodeTheme" + theme);
                                _.each(network.nodes, function(n2){
                                    if (n2.id === $(n).attr("id")){
                                        n2.theme = theme;
                                    }
                                });
                            });
                            break;
                        case "addNote":
                            var activesheet = getActiveSheet();
                            var index = 0;
                            var noteId;
                            var newNoteObj={};

                            if(_.has(network, "notes") === false) {
                                network.notes = [];
                                noteId = "note";
                            } else {
                                if( _.last(network.notes) === undefined) {
                                    index = 0;
                                } else {
                                    var lastIndex = _.last(network.notes);
                                    index = lastIndex.index + 1;
                                }
                                noteId = "note" + index;
                            }

                            newNoteObj.pos = [cursorPos.left, cursorPos.top];
                            newNoteObj.visible = 1;
                            newNoteObj.sheet = activesheet;
                            newNoteObj.index = index;
                            newNoteObj.noteId = noteId;

                            network.notes.push(newNoteObj);

                            addNoteDOM(newNoteObj);

                            var $note = initNote(noteId, true);

                            $note.dialog('open');

                            break;
                    }
                },
                items: items
            };
        }
    });

    //context menu for input terminals
    $("#sheets").contextMenu({
        selector: ".terminal",
        trigger: isSelenium,
        build: function($trigger, event){
            var items = {};
            var termid = $trigger.attr("id");
            var terminalSide = ($trigger.parent().hasClass("interminals")) ? "Input" : "Output";
            var termName = getAfterDelim(termid);
            var nodeobj = _.findObject(network.nodes, "id", getBeforeDelim(termid));

            // if the terminal is already input, set "unset" menu
            // otherwise, set "set as input" menu
            var isIOAlreadySet = _.find(network.terminalRoles, function(role){
                return role.role !== "condition" && role.nodeId + delim + role.termName === termid;});
            if (isIOAlreadySet){
                items.unset = {name: "Delete " + terminalSide, icon: "delete"};
            } else {
                items.set = {name: "Set as " + terminalSide, icon: "add"};
            }

            var isLinkExists = _.find(network.links, function(link){
                return termid === link.fromNodeId + delim + link.fromTermName ||
                       termid === link.toNodeId   + delim + link.toTermName;
                   });
            if (isLinkExists){items.unlink = {name: "Delete the link", icon: "delete"};}

            var dsc = _.findObject(nodeobj[terminalSide.toLowerCase() + "s"],
                                     "name", termName).description;
            if (dsc && dsc.indexOf("added by the user") !== -1 || dsc.indexOf("Dynamically added") !== -1){
                items.deleteIO = {name: "Delete this " + terminalSide, icon: "delete"};
            }

            //get sheetname
            var sheetname = getActiveSheet();

            if (terminalSide == "Output"){
                var isConditionAlreadySet = _.find(
                    network.terminalRoles,
                    function(role){
                        return role.role === "condition" && role.nodeId + delim + role.termName == termid;
                    });
                if (isConditionAlreadySet){
                    items.deleteCondition = {name: "Delete Condition", icon: "delete"};
                }
                //add menu for setting as condition only if sheet is not subnet
                else if(network.sheets[sheetname].type !== "subnet") {
                    items.setCondition    = {name: "Set as Condition", icon: "add"};
                }
            }

            return {
                callback:
                function (key, options){
                    var clickedTermId = $(this).attr("id");
                    var clickedNodeId = getBeforeDelim(clickedTermId);
                    var clickedNodeObj = _.findObject(network.nodes, "id", clickedNodeId);
                    //var clickedNodeName = nodeId2nodename(clickedNodeId);
                    var clickedTermName = getAfterDelim(clickedTermId);
                    var delTarget;
                    switch (key){
                        case "unset":
                            delTarget = _.find(network.terminalRoles,
                                function(role){return role.nodeId + delim + role.termName === termid && role.role !== "condition";});
                            updateTerminalRoles(delTarget, "del");

                            if (currentTab !== "MAIN"){
                                var dynamic = _.findObject(network.nodes, "name", currentTab);
                                network.links = _.reject(network.links,
                                    function(link){return link.fromNodeId === dynamic.id && link.fromTermName === clickedTermName;});
                                network.links = _.reject(network.links,
                                    function(link){return link.toNodeId === dynamic.id && link.toTermName === clickedTermName;});
                            }
                            break;
                        case "set":
                            var ioobj = (terminalSide === "Input") ? clickedNodeObj.inputs : clickedNodeObj.outputs;
                            ioobj = _.findObject(ioobj, "name", clickedTermName);
                            $("#terminalRoleNameDialog")
                                .attr("clickedNodeId", clickedNodeId)
                                .attr("clickedTermName", clickedTermName)
                                .attr("role", terminalSide.toLowerCase())
                                .attr("ioobj_type", ioobj.type)
                                .attr("ioobj_description", ioobj.description);

                            var numNames = _.chain(network.terminalRoles)
                                .filter(function(r){return r.sheet === currentTab;})
                                .filter(function(r){return r.name.replace(/_[0-9]$/, "") === clickedTermName;})
                                .value();
                            if (numNames.length === 0){
                                $("#terminalRoleName").val(clickedTermName);
                            } else {
                                var filterfunc = function(r){return r.sheet === currentTab;};
                                var findfunc   = function(r){return r.name === candidate;};
                                for (var i = 0; i < numNames.length + 1; i++){
                                    var candidate = (i === 0) ? clickedTermName : (clickedTermName + "_" + i);
                                    if (_.chain(network.terminalRoles)
                                        .filter(filterfunc)
                                        .find(findfunc).value() === undefined){
                                        $("#terminalRoleName").val(candidate);
                                        break;
                                    }
                                }
                            }
                            $("#terminalRoleNameDialog").dialog("open");

                            break;
                        case "unlink":
                            network.links = _.reject(network.links, function(link){
                                return clickedTermId === link.fromNodeId + delim + link.fromTermName ||
                                       clickedTermId === link.toNodeId   + delim + link.toTermName;
                            });
                            clearCanvas();
                            drawLinks();
                            break;
                        case "deleteCondition":
                            delTarget = _.find(network.terminalRoles,
                                function(role){return role.nodeId + delim + role.termName === termid;});
                            updateTerminalRoles(delTarget, "del");
                            break;
                        case "setCondition":
                            newTerminalRole({
                                nodeId: clickedNodeId,
                                termName: clickedTermName,
                                name: "CONDITION",
                                role: "condition",
                                sheet: currentTab
                            });
                            break;
                        case "deleteIO":
                            clickedNodeObj[terminalSide.toLowerCase() + "s"] = _.reject(
                                clickedNodeObj[terminalSide.toLowerCase() + "s"], function(t){
                                return t.name === clickedTermName;
                            });
                            network.terminalRoles = _.reject(network.terminalRoles, function(r){
                                return r.nodeId === clickedNodeId &&
                                       r.termName === clickedTermName;
                            });

                            $("#" + clickedTermId).remove();

                            //remove the links from network object
                            network.links = _.reject(network.links, function(link){
                                return clickedTermId === link.fromNodeId + delim + link.fromTermName ||
                                    clickedTermId === link.toNodeId   + delim + link.toTermName;
                            });
                        break;

                    }
                    updateDynamicNodeDOM(network);
                    clearCanvas();
                    drawLinks();
                },
                items: items};
        }
    });

    $(document).on("click", ".sheetContainer", function(event){
        $(event.currentTarget).focus();
    });
}

function nodeId2category(nodeId){
    var nodename = nodeId.substring(nodeId.indexOf("_") + 1,
                                    nodeId.lastIndexOf("_"));
    var def = _.findObject(nodedefs, "name", nodename);
    if (def){
        return def.category;
    } else {
        return "Undefined";
    }
}

function updateTerminalRoles(role, command, stopUpdateDOM){
    // command in ["add", "del"]

    var obj;
    if (command === "add"){
        if (role.sheet !== "MAIN"){
            var newio = {
                name: role.name,
                type: "any",
                description: "Dynamic"
            };
            updateDynamicNode(role.sheet, role.role, newio);

            if (role.role !== "condition"){
                // add to defs
                _.chain(nodedefs)
                 .filter(function(def){return def.name === role.sheet;})
                 .each(function(def){
                    if (_.find(def[role.role + "s"], function(io){return io.name === role.name;}) === undefined){
                        def[role.role + "s"].push(newio);
                    }
                    });

                 // add to nodes
                 _.chain(network.nodes)
                  .filter(function(n){return n.name === role.sheet;})
                  .each(function(n){
                    if (_.find(n[role.role + "s"], function(io){return io.name === role.name;}) === undefined){
                        n[role.role + "s"].push(newio);
                    }
                  });
            }
        }
        obj = $("#" + role.nodeId + delim + role.termName + " .terminalRoleName");
        if (obj.text().trim().replace(/[\W]+/g, " ").split(" ").indexOf(role.name) === -1){
            obj.text(obj.text() + " " + role.name);
        }
    } else if (command === "del"){
        if (role.role !== "condition"){
            // remove from defs
            _.chain(nodedefs)
             .filter(function(def){return def.name === role.sheet;})
             .each(function(def){
                def.outputs = _.reject(def.outputs, function(o){return o.name === role.termName;});
                def.inputs  = _.reject(def.inputs,  function(i){return i.name === role.termName;});
             });

            // remove from nodes
            _.chain(network.nodes)
             .filter(function(n){return n.name === role.sheet;})
             .each(function(n){
                n.outputs = _.reject(n.outputs, function(o){return o.name === role.name;});
                n.inputs  = _.reject(n.inputs,  function(i){return i.name === role.name;});
            });

            // remove terminalRoles
            var tmp = _.chain(network.terminalRoles)
                   .reject(function(r){
                    if (nodeId2nodename(r.nodeId) === role.sheet && r.termName === role.name){
                        updateTerminalRoles(r, "del");
                        return true;
                    } else {
                        return false;
                    }
                }).value();
            tmp = network.terminalRoles;

            network.links = _.reject(network.links, function(l){
                return (nodeId2nodename(l.fromNodeId) === role.sheet && l.fromTermName === role.name) ||
                       (nodeId2nodename(l.toNodeId)   === role.sheet && l.toTermName   === role.name);
            });
            // _.each(network.terminalRoles, function(r){
                // if (nodeId2nodename(r.nodeId) === role.sheet && r.termName === role.name)
                    // updateTerminalRoles(r, "del");
              // });
        }

        //remove from terminalRoles
        network.terminalRoles = _.reject(network.terminalRoles, function(r){
            return r === role;
        });
        obj = $("#" + role.nodeId + delim + role.termName + " .terminalRoleName");
        obj.text(obj.text().replace(role.name, ""));
    } else {
        console.log("updateTerminalRoles: unknown command: " + command + ". Command must be add or del.");
    }

    if (stopUpdateDOM !== true || role.sheet === "MAIN"){
        updateDynamicNodeDOM(network);
    }
}

function updateDynamicNodeDOM(networkobj, forceClose){
    // if (role && command){
    //     console.log(role, command);
    //     if (command === "add" && role.role !== "condition"){
    //         var selector = "#" + role.nodeId + " " + role.role.replace("put", "") + "terminals";
    //         var tmpl = '<die class="terminal" description="<%- description %>" name="<%= name %>" type="<%- object_type %>" id="<%= nodeId %>' + delim + '<%= name %>" ><div class="terminalcircle"><span class="terminalRoleName"></span></div><%= name %></div>';
    //         console.log(selector);
    //         console.log(_.template(tmpl, role));
    //         $(selector).append(_.template(tmpl, role));
    //     }
    //     else if (command === "del" && role.role !== "condition"){

    //     }
    //     return;
    // }

    // remove and add
    var updatedNodeNames = _.chain(networkobj.nodes)
                            .groupBy("sheet")
                            .keys()
                            .reject(function(k){return k === "MAIN";})
                            .value();
    _.chain(networkobj.nodes)
     .filter(function(n){return updatedNodeNames.indexOf(n.name) !== -1;})
     .each(function(n){
        var sel = "#" + n.id + delim + "parameters";
        var isOpen = $(sel).dialog("isOpen");
        removeNode(n.id, true, true, true, true);
        addNodeDOM(n, true);
        if (isOpen && !forceClose){
            $(sel).dialog("open");
        }

        _.chain(networkobj.terminalRoles)
         .filter(function(r){return r.nodeId === n.id;})
         .each(function(r){
            var obj = $("#" + r.nodeId + delim + r.termName + " .terminalRoleName");
            if (obj.text().trim().replace(/[\W]+/g, " ").split(" ").indexOf(r.name) === -1){
                obj.text(obj.text() + " " + r.name);
            }
         });
     });
}

/* drawing functions.*/
function clearCanvas(canvasid){
    canvasid = canvasid || "canvas_" + currentTab;
    var curcanvas = document.getElementById(canvasid);
    if(curcanvas) {
        var ctx = curcanvas.getContext("2d");
        ctx.fillStyle = "#f8f8f8";
        ctx.fillRect(0, 0, curcanvas.width, curcanvas.height);
    }
}

function truncate(pos){
    return parseInt(pos / 10, 10) * 10.0;
}

function Bezier(poses, context){
/*http://ruiueyama.tumblr.com/post/11197882224*/
    function interp(a, b, f){return a + f * (b - a);}
    var split = 15;

    context.moveTo(poses[0].left, poses[0].top);
    for (var i = 0; i < split; i++){
        var p1 = {left: interp(poses[0].left, poses[1].left, 1.0 * i / split),
                  top : interp(poses[0].top,  poses[1].top,  1.0 * i / split)};
        var p2 = {left: interp(poses[1].left, poses[2].left, 1.0 * i / split),
                  top : interp(poses[1].top,  poses[2].top,  1.0 * i / split)};

        context.lineTo(interp(p1.left, p2.left, 1.0 * i / split),
                       interp(p1.top,  p2.top,  1.0 * i / split));
    }
    context.lineTo(poses[2].left, poses[2].top);
}

function intelligentLink(frompos, topos, context, mtop, mleft){
    var mode = userPreference.get("link");
    mleft = mleft || (frompos.left + topos.left) / 2.0;
    mtop  = mtop || (frompos.top + topos.top) / 2.0;
    var margin = Math.min(Math.abs(frompos.left - topos.left), 20);
    var poses = [
        frompos,
        {left: frompos.left + margin, top: frompos.top},
        {left: frompos.left + margin, top: (frompos.top + mtop) / 2.0},
        {left: frompos.left + margin, top: mtop},
        {left: mleft,                 top: mtop},
        {left: topos.left - margin,   top: mtop},
        {left: topos.left - margin,   top: (topos.top + mtop) / 2.0},
        {left: topos.left - margin,   top: topos.top},
        topos
    ];

    if (mode === "Bezier"){
        if (frompos.left < topos.left){
            var m1 = {left: mleft, top: frompos.top};
            var m2 = {left: mleft, top: mtop};
            var m3 = {left: mleft, top: topos.top};

            Bezier([frompos, m1, m2], context);
            Bezier([m2, m3, topos], context);
        } else {
            Bezier([poses[0], poses[1], poses[2]], context);
            Bezier([poses[2], poses[3], poses[4]], context);
            Bezier([poses[4], poses[5], poses[6]], context);
            Bezier([poses[6], poses[7], poses[8]], context);
        }
    } else if (mode === "Lines"){
        if (frompos.left < topos.left){
            context.moveTo(frompos.left, frompos.top);
            context.lineTo(mleft, frompos.top);
            context.lineTo(mleft, topos.top);
            context.lineTo(topos.left, topos.top);
        } else {
            context.moveTo(poses[0].left, poses[0].top);
            for (var i = 1; i < poses.length; i++){
                context.lineTo(poses[i].left, poses[i].top);
            }
        }
    }
}

function drawCandidateLinks(canvasid){
    var targetSide = (currentLinkFrom.parent().hasClass("interminals")) ? "outputs" : "inputs";
    var mySide = (currentLinkFrom.parent().hasClass("interminals")) ? "inputs" : "outputs";
    var curFromId = currentLinkFrom.attr("id");

    var currentLinkFromType = _.findObject(network.nodes, "id", getBeforeDelim(curFromId))[mySide];
    currentLinkFromType = _.findObject(currentLinkFromType,
                                       "name", getAfterDelim(curFromId)).type;
    var candidates = _.chain(network.nodes)
                        .filter(function(n){return n.sheet === currentTab;})
                        .map(function(n){
                            return _.chain(n[targetSide])
                                    .filter(function(term){return isTypesMatch(term.type, currentLinkFromType);})
                                    .map(function(term){
                                        if (mySide === "outputs"){
                                            return {
                                                 fromNodeId: getBeforeDelim(curFromId),
                                                 fromTermName: getAfterDelim(curFromId),
                                                 toNodeId  : n.id,
                                                 toTermName: term.name,
                                                 sheet: currentTab
                                            };
                                        } else {
                                            return {
                                                 fromNodeId  : n.id,
                                                 fromTermName: term.name,
                                                 toNodeId: getBeforeDelim(curFromId),
                                                 toTermName: getAfterDelim(curFromId),
                                                 sheet: currentTab
                                            };
                                        }
                            }).value();
                        }).flatten().value();
    candidates = _.filter(candidates, function(c){return testLink(c) === "";});
    drawLinks(canvasid, candidates, "#888", 1, true);
}

function drawLinks(canvasid, links, style, width, isDotted){
    links = links || network.links;
    style = style || "black";
    width = width || 2;
    canvasid = canvasid || "canvas_" + currentTab;

    var curcanvas = document.getElementById(canvasid);
    var context = curcanvas.getContext("2d");
    context.strokeStyle = style;
    context.lineWidth = width;
    if (isDotted){
        if (context.setLineDash) {context.setLineDash([3, 3]);}
        context.mozDash = [3, 3];
    } else {
        if (context.setLineDash) {context.setLineDash([1, 0]);}
        context.mozDash = [1, 0];
    }

    context.beginPath();
    _.each(links, function(link){
        var tmpl = "#<%= id %><%= delim %><%= link %> .terminalcircle:first";
        var fromNodeId  = _.template(tmpl, {id: link.fromNodeId, link: link.fromTermName, delim: delim});
        var toNodeId    = _.template(tmpl, {id: link.toNodeId, link: link.toTermName, delim: delim});
        var fromTerminal = $(fromNodeId);
        var toTerminal = $(toNodeId);
        if (fromTerminal.length === 0){return;}
        if (toTerminal.length === 0){return;}
        var fromNode = $("#" + link.fromNodeId);
        var toNode = $("#" + link.toNodeId);
        if (fromNode.length === 0){return;}
        if (toNode.length === 0){return;}

        var sc = fromTerminal.closest(".sheetContainer div.sheet");
        var frompos = {
            left: fromTerminal.offset().left -
                  sc.offset().left + 6,
            top:  fromTerminal.offset().top -
                  sc.offset().top + 6
        };
        var topos   = {
            left: toTerminal.offset().left -
                  sc.offset().left + 6,
            top:  toTerminal.offset().top -
                  sc.offset().top + 6
        };

        var fromnodepos = fromNode.position();
        fromnodepos.bottom = fromnodepos.top + fromNode.height();
        fromnodepos.right = fromnodepos.left + fromNode.width();
        var tonodepos = toNode.position();
        tonodepos.bottom = tonodepos.top + toNode.height();

        if (fromnodepos.right > tonodepos.left) {
            if (fromnodepos.bottom < tonodepos.top) {
                intelligentLink(frompos, topos, context,
                    (fromnodepos.bottom + tonodepos.top) / 2.0);
            } else if (tonodepos.bottom < fromnodepos.top){
                intelligentLink(frompos, topos, context,
                    (tonodepos.bottom + fromnodepos.top) / 2.0);
            }
        } else {
            intelligentLink(frompos, topos, context);
        }
    });
    context.stroke();
    context.closePath();
}

function crossBrowserMousePos(event){
    /*http://www.jacklmoore.com/notes/mouse-position/*/
    var target = event.currentTarget || event.srcElement,
        rect = target.getBoundingClientRect(),
        offsetX = event.clientX - rect.left,
        offsetY = event.clientY - rect.top;
    return {left: offsetX, top: offsetY};
}

function drawCurrentLine(canvasid, event){
    canvasid = canvasid || "canvas_" + currentTab;
    var curcanvas = document.getElementById(canvasid);
    var context = curcanvas.getContext("2d");

    var circle = currentLinkFrom.find(".terminalcircle:first");
    var sc = circle.closest(".sheetContainer div.sheet");

    var termpos = {
        left: circle.offset().left - sc.offset().left + 6,
        top : circle.offset().top  - sc.offset().top + 6
    };
    var mousepos = crossBrowserMousePos(event);

    // set line style (black, width=1, dotted)
    context.strokeStyle = "black";
    context.lineWidth = 1;
    if (context.setLineDash) {context.setLineDash([3, 3]);}
    context.mozDash = [3, 3];

    context.beginPath();
    var mySide = (currentLinkFrom.parent().hasClass("interminals")) ? "inputs" : "outputs";
    if (mySide === "inputs"){
        intelligentLink(mousepos, termpos, context);
    } else {
        intelligentLink(termpos, mousepos, context);
    }
    context.stroke();
    context.closePath();
}

function relativePos(of, from){
    if (of === undefined || from === undefined){
        return undefined;
    }
    // of and from must be a jQuery object $(...)
    var pad = {};
    pad.top = from.css("padding-top").slice(0, -2);
    pad.left = from.css("padding-left").slice(0, -2);

    var mid = {};
    mid.top = of.css("height").substr(0, 2) / 2.0;
    mid.left = of.css("width").substr(0, 2) / 2.0;

    var pos = {};
    pos.top  = - from.position().top  + of.position().top - pad.top + mid.top;
    pos.left = - from.position().left + of.position().left - pad.left + mid.left;

    return pos;
}

/* setup #sheets */
function addDynamicNode(sheetname){
    var flag = 0;

    _.each(nodedefs, function(n){ // loops through nodedefs array
        var nodeName = n.name;

        if (nodeName.toUpperCase() === sheetname.toUpperCase()) // compares existing nodes
            flag = 1;
    });

    if (flag >= 1) { // checks if it has error
        app.set("execstat", "Error: '" + sheetname + "' sheet already exists"); // sends error msg
        return false;
    }
    else { // proceeds to adding dynamic node
        nodedefs.push({
            name:        sheetname,
            category:    "Dynamic",
            description: "Dynamic node",
            inputs:      [],
            outputs:     [],
            parameters:  [],
            def:         "Dynamic",
            theme:       "Classic"
        });
        updateNodeList();

        // returns true if has no errors
        return true;
    }
}

function updateDynamicNode(sheetname, type, option){
    // update inputs/outputs/parameters of dynamic node
    if (sheetname === "MAIN"){return;}

    var def = _.findObject(nodedefs, "name", sheetname);

    if (type === "input" || type === "output"){
        if (def[type + "s"] &&
           _.find(def[type + "s"], function(i){return i.name === option.name;}) === undefined){
            def[type + "s"].push(option);
        }
        _.each(network.nodes, function(n){
            if (n.name === sheetname &&
               _.find(n[type + "s"], function(i){return i.name === option.name;}) === undefined){
                n[type + "s"] = n[type + "s"] || [];
                n[type + "s"].push(option);
            }
        });
    }
}

function addSheet(sheetname, sheettype){
    var result = false;
    var title = sheetname;
    //ensuring that sheetname doesn't contain spaces
    sheetname = sheetname.replace(/ /g, "");
    // new sheets can be created if 'addDynamicNode' function returns true
    if (addDynamicNode(sheetname)){
        result = newSheet({name: sheetname, type: sheettype, title: title});
        addSheetDOM(network.sheets[sheetname]);
    }

    return result;
}

function getSheetId(sheetname){
    var index;
    $("#sheets .ui-tabs-nav li[role=tab]").each(function(_index, tab){
        if($(tab).attr("aria-controls") === "sheet_" + sheetname) {
            index = _index;
            return false;
        }
    });
    return index;
}

//function for retrieving currently active sheet tab
function getActiveSheet() {
    var sheetname = "";
    var sheetList = $("#sheets").find(".sheetTab");
    _.each(sheetList, function(sheet) {
        if ($(sheet).attr("aria-selected") === "true") {
            sheetname = $(sheet).attr("aria-controls").replace("sheet_", "");
        }
    });

    return sheetname;
}

//function for retrieving title of tab
function getSheetTitle(sheetname) {
    var sheettitle = "";
    var sheetList = $("#sheets").find(".sheetTab");
    _.each(sheetList, function(sheet) {
        if ($(sheet).attr("aria-controls").replace("sheet_", "") === sheetname) {
            sheettitle = $(sheet).find("a").text();
        }
    });

    return sheettitle;
}

//function for removing spaces in sheet names
function correctSheetNames() {
    _.each(network.nodes, function(n){
        //remove spaces
        n.sheet = n.sheet.replace(/ /g, "");
    });

    //rename sheet in links
    _.each(network.links, function(l){
        //remove spaces
        l.sheet = l.sheet.replace(/ /g, "");
    });

    //rename sheet in terminal roles
    _.each(network.terminalRoles, function(t){
        //remove spaces
        t.sheet = t.sheet.replace(/ /g, "");
    });

    //rename sheet in notes
    _.each(network.notes, function(nt){
        //remove spaces
        nt.sheet = nt.sheet.replace(/ /g, "");
    });

    var newsheets = {};
    //rename sheets
    _.each(network.sheets, function(s){
        //replace all sheet name to sheet title to include spaces
        s.name = s.name.replace(/ /g, "");
        newsheets[s.name] = s;
    });

    network.sheets = newsheets;

}

//function for reverting to original sheet names
function revertSheetNames() {
    //rename sheet in nodes
    _.each(network.nodes, function(n){
        //replace all sheet name to sheet title to include spaces
        n.sheet = getSheetTitle(n.sheet);
    });

    //rename sheet in links
    _.each(network.links, function(l){
        //replace all sheet name to sheet title to include spaces
        l.sheet = getSheetTitle(l.sheet);
    });

    //rename sheet in terminal roles
    _.each(network.terminalRoles, function(t){
        //replace all sheet name to sheet title to include spaces
        t.sheet = getSheetTitle(t.sheet);
    });

    //rename sheet in notes
    _.each(network.notes, function(nt){
        //replace all sheet name to sheet title to include spaces
        nt.sheet = getSheetTitle(nt.sheet);
    });

    var newsheets = {};
    //rename sheets
    _.each(network.sheets, function(s){
        //replace all sheet name to sheet title to include spaces
        var index = s.title || s.name;
        s.name = index;
        newsheets[index] = s;
    });

    network.sheets = newsheets;
}

function addSheetDOM(sheet){
    var sheetname = sheet.name.replace(/ /g, "");

    var template =
        "<li class='sheetTab'>" +
        "<a href='#sheet_<%= name %>' class='sheetType_<%= type %>'>" +
        "<%= title %>" +
        // "<br> <span class='sheetType'>(<%= type %>)</span>" +
        "<span class='ui-icon ui-icon-close' style='display:inline-block;' role='presentation'></span>" +
        "</a>" +
        "</li>";
    if (sheet.title === undefined) {
        sheet.title = sheet.name;
    }

    sheet.name = sheetname;
    var li = _.template(template, sheet);
    var tabs = $("#sheets");

    tabs.find(".ui-tabs-nav").append(li);
    tabs.append(
        "<div id='sheet_" + sheetname + "' class='sheetContainer' tabindex='0'>" +
        "    <canvas class='sheet' id='canvas_" + sheetname + "'></canvas>" +
        "    <div class='sheet' id='canvas2_" + sheetname + "'></div>" +
        "</div>");
    tabs.tabs("refresh");
    setupSelectable(sheetname);
    setupCanvasEvent(sheetname);
    // update network data structure.
    updateSheetsSize();
    clearCanvas("canvas_" + sheetname);

    tabs.tabs({active: getSheetId(sheetname)});
    $(".sheetTab[aria-controls='sheet_" + sheet.name + "']").tooltip(
        {position: {my: 'center bottom', at: 'center top-5'}});
    setTimeout(function(){sheetsView.toggleMinimize();}, 100);
}

var flagSheetTab = 0;
var sheetTabs = $("#sheets ul");
var sheetTabHeight = null;

function removeSheet(sheetname, noAnimation, afterCallback){
    var obj = $("a[href=#" + sheetname + "]").closest("li");

    if (noAnimation){
        obj.remove();
        $.when($("#" + sheetname).unbind().remove()).then( updateCurrentTab() );
    } else {
        obj.fadeOut(function(){obj.unbind().remove();});
        $("#" + sheetname).fadeOut(function(){
            $.when($("#" + sheetname).unbind().remove()).then(function(){
                updateCurrentTab();

                if  (afterCallback) {
                    afterCallback();
                }
            });

        });
    }

    // from sheet_??? --> ???
    sheetname = sheetname.substring(sheetname.indexOf("_") + 1);
    delete network.sheets[sheetname];

    //update network data structure.
    var removedNodes = _.filter(network.nodes,
        function(node){return node.sheet == sheetname;});

    //get all dynamic nodes
    var dynamicNodes = _.filter(network.nodes,
        function(node){return node.name == sheetname;});

    // remove the link if its fromNodeId or toNodeId is in removedNodes.
    network.links = _.reject(network.links,
        function(link){
            return _.find(removedNodes,function(n){return n.id === link.fromNodeId || n === link.toNodeId;});});

    //remove the link if its fromNodeId or toNodeId is in dynamicNodes.
    network.links = _.reject(network.links,
        function(l){
            return _.find(dynamicNodes,function(n){return n.id === l.fromNodeId || n === l.toNodeId;});});

    // remove dynamic node and the nodes in the removed sheet.
    _.chain(network.nodes)
     .filter(function(n){return n.sheet === sheetname || n.name === sheetname;})
     .each(function(n){removeNode(n.id, false, noAnimation);});

    // remove from the nodedef
    $("#flatNodelist [key=" + sheetname + "]").unbind().remove();
    $(_.chain($("#nodelist [role=menuitem]"))
       .find(function(n){return $(n).text().trim() === sheetname;})
       .value()).parent().unbind().remove();

    // removes dynamic node from the nodedefs array
    var nodeName = _.findWhere(nodedefs, {category: "Dynamic", name: sheetname}); // finds the node
    var index = _.indexOf(nodedefs, nodeName); // gets the index
    if (index > -1) {
        nodedefs.splice(index, 1); // removes the node
        updateNodeList(); // updates nodelist
    }

    $("#sheet").tabs("refresh");

    // Check if has scroll
    if (flagSheetTab >= 1) {
        setTimeout(function() {
            if (sheetTabs.get(0).scrollWidth <= sheetTabs.innerWidth()) {
                sheetTabs.attr('style', 'height:' + sheetTabHeight + 'px');
                flagSheetTab = 0;
            }
        }, 500);
    }

    //drawLinks();
}

function setupSheets(){
    $("#sheets").tabs({
        activate: function(event, ui){
            currentTab = ui.newTab.text().replace(/ /g, "");
            _.chain(network.nodes)
             .filter(function(n){return n.sheet === currentTab;})
             .each(function(n){$("#" + n.id).css({left: n.pos[0], top: n.pos[1]});});

            _.chain(network.notes)
                .filter(function(n){return n.sheet === currentTab;})
                .each(function(n){$("#" + n.noteId).css({left: n.pos[0], top: n.pos[1]});});

            updateSheetsSize();
            clearCanvas();
            drawLinks();

            $(".nodeSelected").removeClass("nodeSelected");
            $(".noteSelected").removeClass("noteSelected");
        }
    });
    // $("#sheets .ui-tabs-nav").css("overflow-y", "hidden").css("overflow-x", "scroll");

    $("#addSheetButton").button({
        text: false,
        icons: {primary: "ui-icon-plusthick"}
    }).click(function(){addSheetDialog.dialog("open");});

    $("#reshapeButton").button({
        text: false,
        icons: {primary: "ui-icon-arrowrefresh-1-s"}
    }).click(function(){
        var postdata = { sheets: {}, nodes: [], links: [], preference: {} };
        _.each(_.keys(network.sheets), function(sheetname){
            // var sheetname = $("#sheets .ui-tabs-active").text();
            $("#sheet_" + sheetname).scrollTop(0);
            $("#sheet_" + sheetname).scrollLeft(0);
            // All sheets' style change temporarily into display:block for measuring nodes size
            $("#sheet_" + sheetname).css("display","block");
            // Get sheet display size
            postdata.sheets[sheetname] = network.sheets[sheetname];
            postdata.sheets[sheetname].dispWidth = $("#sheet_" + sheetname).width();
            postdata.sheets[sheetname].dispHeight = $("#sheet_" + sheetname).height();
        });
        var prevActiveTab = $("#sheets").tabs("option", "active");

        // Measure node size
        _.each(network.nodes, function(node){
            var newNode = { id: node.id, sheet: node.sheet };
            var nodeElm = $("#"+node.id);
            newNode.dispWidth = nodeElm.width();
            newNode.dispHeight = nodeElm.height();
            // get InputTerminalRole width
            var itrWidth = 0;
            nodeElm.find(".interminals span.terminalRoleName").each(function(index, elm){
                var _width = $(elm).width();
                if (itrWidth < _width) {
                    itrWidth = _width;
                }
            });
            if (itrWidth > 0) {
                newNode.inputTermRoleWidth = itrWidth;
            }
            postdata.nodes.push(newNode);
            //console.log(node.id+": w "+node.dispWidth+", h "+node.dispHeight);
        });

        // Revert sheet style
        _.each(_.keys(network.sheets), function(sheetname){
            if (sheetname !== currentTab)
                $("#sheet_" + sheetname).css("display","none");
            //console.log(sheetname+".style.display="+$("#sheet_" + sheetname).css("display"));
        });

        _.each(network.links, function(link){
            postdata.links.push(link);
        });

        postdata.preference.reAlign = userPreference.get("reAlign");

        $.ajax({
            type: "PUT",
            url: rootPath + "reshapedNodes",
            dataType: "json",
            data: postdata
        }).done(function(reshaped, dataType){
            if (!reshaped.nodes){return;}
            //network = reshaped;
            _.each(network.nodes, function(n){
                var reshapedNode = _.findObject(reshaped.nodes, "id", n.id);
                if (reshapedNode) {
                    n.pos[0] = truncate(reshapedNode.pos[0]) + 60;
                    n.pos[1] = truncate(reshapedNode.pos[1])+ 90;
                }
            }, reshaped);
            _.each(network.nodes, function(n){
                console.log(n.id, n.pos[0], n.pos[1]);
            });
            _.each(_.range(_.keys(network.sheets).length), function(id){
                $("#sheet").tabs({active: id});
                var sheetname = $("#sheets .ui-tabs-active").text();
                $("#sheet_" + sheetname).scrollTop(0);
                $("#sheet_" + sheetname).scrollLeft(0);
                _.delay(function(){
                    _.chain(network.nodes)
                     .filter(function(n){return n.sheet === sheetname;})
                     .each(function(n){
                        $("#" + n.id).css({left: n.pos[0], top: n.pos[1]});
                     });
                    clearCanvas("canvas_" + sheetname);
                    drawLinks("canvas_" + sheetname);
                }, 300);
            });
            $("#sheets").tabs({active: prevActiveTab});
        }).fail(function(data){
            alert("you need to install Graphviz.");
            // $("#reshapeButton").tooltip();
            // $("#reshapeButton").tooltip("destroy");
            // $("#reshapeButton").remove();
        });
    });

    $("#AddIODialog").dialog({
        autoOpen: false,
        modal: true,
        buttons : {
            Add: function(){
                var clickedNodeId = $(this).attr("clickedNodeId");
                var clickedNodeObj = _.findObject(network.nodes, "id", clickedNodeId);
                var terminalRolesOfClickedNode =
                    _.chain(network.terminalRoles)
                     .filter(function(n){return n.nodeId === clickedNodeId;})
                     .value();
                var type = ($(this).attr("title").indexOf("Input") === -1) ? "outputs" : "inputs";
                var value = $("#IOName").val();

                if (_.pickArray(clickedNodeObj.inputs,  "name").indexOf(value) !== -1 ||
                   _.pickArray(clickedNodeObj.outputs, "name").indexOf(value) !== -1){
                    raise(clickedNodeId + " names are duplicated.", value);
                } else {
                    clickedNodeObj[type].push({
                        description: "Dynamically added " + type,
                        name: value,
                        type: "any"
                    });
                    type = (type == "inputs") ? "in" : "out";
                    $("#" + clickedNodeId + " ." + type + "terminals").html(
                        _.template($("#nodeIOTemplate").html(),
                                   {node: clickedNodeObj, type: type, term: terminalRolesOfClickedNode})
                    );
                }
                $(this).dialog("close");
            },
            Cancel: function(){$(this).dialog("close");}
        }
    });
    addSheetDialog = $("#addSheetDialog").dialog({
        autoOpen: false,
        modal: true,
        buttons : {
            Add: function(){
                var sheetname = $("#addSheetName").val();

                //validate if sheet name has special characters except space, hyphen, and underscore
                if(!sheetname.match(/[^A-Za-z0-9\-_\ ]/)) {
                    var sheettype = $("#addSheetType").val();

                    var result = false;
                    result = addSheet(sheetname, sheettype);
                    //add validation if create sheet or create sheet from node
                    if (result) {
                        var options = $(this).dialog("option");
                        if (options.nodesForCreateSheet) {
                            // create sheet from selected
                            $(".node").removeClass("nodeSelected");
                            $(".note").removeClass("noteSelected");

                            addObjs(options.nodesForCreateSheet, sheetname);
                            drawLinks();
                            $(this).dialog("option", {nodesForCreateSheet: undefined});
                        }

                        $("#addSheetName").val("");
                        $(this).dialog("close");
                    }

                } else {
                    app.set("execstat", "Error: Sheet name contains invalid characters.");
                }

                // Check if has scroll
                if (flagSheetTab === 0) {
                    var scrollWidth = 17;
                    sheetTabHeight = Math.ceil(sheetTabs.height());

                    if (sheetTabs.get(0).scrollWidth > sheetTabs.innerWidth()) {
                        sheetTabs.attr('style', 'height:' + parseInt(sheetTabHeight + scrollWidth) + 'px');
                        flagSheetTab = 1;
                    }
                }
            },
            Cancel: function(){
                $("#addSheetName").val("");
                $(this).dialog("close");
            }
        }
    });

    $("#addSheetDialog").find("form").submit(function(event){
        $($('#addSheetDialog').siblings('.ui-dialog-buttonpane').find("button")[0]).click();
        addSheetDialog.dialog("close");
        event.preventDefault();
        event.stopPropagation();
    });

    $("#AddIODialog").find("form").submit(function(event){
        addSheet($("#addSheetName").val(), $("#addSheetType").val());
        addSheetDialog.dialog("close");
        event.preventDefault();
        event.stopPropagation();
    });

    $("#sheets").on("click","span.ui-icon-close",  function(){
        var sheetname = $(this).closest("li").attr("aria-controls");
        if (confirm("Are you sure to remove " + sheetname + "?")){
            removeSheet(sheetname, false, function(){
                //click the last sheet
                var sheetTabs = $(".sheetTab");
                var a = $(sheetTabs[sheetTabs.length - 1]).find("a");
                $(a).click();
            });
        }
    });
    $("#sheets").on("click", function(){
        $("#nodelistSearch input").blur();
    });
    setupSelectable("MAIN");
    setupCanvasEvent("MAIN");
}

function setupSelectable(sheetname){
    $("#sheet_" + sheetname).selectable({
        filter: ".node, .note",
        cancel: "#nodelist",
        distance: 30,
        selecting: function(e, ui){
            var targetObj = $(ui.selecting);
            if (targetObj.hasClass("node")) {
                targetObj.addClass("nodeSelected");
                // $(".nodeSelected").draggable();
            } else if (targetObj.hasClass("note")) {
                targetObj.addClass("noteSelected");
            }
        },
        unselecting: function(e, ui){
            $(ui.unselecting).removeClass("nodeSelected");
            $(ui.unselecting).removeClass("noteSelected");
        },
        start: function(e, ui){
            $(".nodeSelected").removeClass("nodeSelected");
            $(".noteSelected").removeClass("noteSelected");
        }
    });
}

function setupCanvasEvent(sheetname){
    $("#canvas2_" + sheetname).on("click", function(e){
        if (!e.target.closest(".node,.note")) {
            $(".nodeSelected").removeClass("nodeSelected");
            $(".noteSelected").removeClass("noteSelected");
        }
    });
}

/* Load and parse .def file for nodelist*/
function loadNodeListAjax(defname, update){
    rootPath = rootPath || "";
    var tmpdef = defname.replace(".def", "");

    var url = "";

    if (typeof update === 'boolean' && update) {
        url = rootPath + "def/" + defname + "?f=" + update;
    }
    else {
        url = rootPath + "def/" + defname;
    }

    // load .def file
    $.ajax({
        type: "GET",
        url: url,
        dataType: "json",
        beforeSend: function(){
            $(".nodeLoading").append(
                "<div class ='def def" + tmpdef + "'>" + defname + "...</div>");
            $(".nodeLoading").height(
                parseInt($(".nodeLoading").height(), 10) + 20);
        }
    }).done(function(data){
        _.each(data, parseDefFile);
        // parseDefFile(xml);
        setSearchEvents();
        userPreference.addPackage(defname);

        // update display for new nodelist
        updateNodeList();
        $("#sidebar").accordion("refresh");
        $("#sidebar").css("z-index", 1000);
        $("#sheets .ui-tabs-nav").css("height", "43px");
        updateSheetsSize();

        $(".nodeLoading .def" + tmpdef).text(defname + "... done.");
        // $(".nodeLoading").height(parseInt($(".nodeLoading").height()) - 18)
        updateNodeList();
    }).fail(function(data){
        alert("Reading " + defname + " failed.\n You can load from Preference -> Packages.");
        userPreference.delPackage(defname);
    });

    //xml parser
    function parseDefFile(def){
        var nodeobj = $(this);
        if (_.indexOf(_.pluck(nodedefs, "name"), def.name) !== -1){
            return;
        }

        nodedefs.push({
            name:        def.name,
            category:    def.category,
            description: def.description.replace(/[ \n\r]*$/g, ""),
            inputs:      [],
            outputs:     [],
            parameters:  [],
            def:         defname,
            theme:       "Classic"
        });
        _.each(def.inputs, function(i){
            _.last(nodedefs).inputs.push({
                name: i.name,
                type: i.object_type,
                description: i.description.replace(/[ \n\r]*$/g, "")
            });
        });
        _.each(def.outputs, function(o){
            _.last(nodedefs).outputs.push({
                name: o.name,
                type: o.object_type,
                description: o.description.replace(/[ \n\r]*$/g, "")
            });
        });
        _.each(def.parameters, function(p){
            var paramobj = $(this);
            var paramlist = [];
            newParameter(p, _.last(nodedefs));
        });
    }
}
function updateNodeList(){
    var nodelistdom = $("#nodelist");
    var flatnodelistdom = $("#flatNodelist");
    var nodedefsPerCategory = _.groupBy(nodedefs, "category");
    var c_tmpl = '<li type="category" key="<%= category %>" category="<%= category %>"><a href="#"><%= category %></a></li>';
    var n_tmpl = '<li key="<%= name %>" class="nodeElement" title="<%= name %>: <%= description %>"><a href="#"><%= name %></a></li>';
    var f_tmpl = '<li key="<%= name %>" category="<%= category %>" class="nodeElement" title="<%= description %>"><%= name %></li>';

    var isFlatShown = flatnodelistdom.css('display') == 'block';
    // existing nodes and categories
    var datCats = _.sortBy(_.keys(nodedefsPerCategory),
                           function(c){return c;});
    var domCats = _.sortBy(_.map(nodelistdom.children("li"), function(c){return $(c).attr("key");}),
                           function(c){return c;});
    var newCats = _.reject(datCats, function(c){return _.contains(domCats, c);});
    var extCats = _.reject(datCats, function(c){return _.contains(newCats, c);});
    var delCats = _.reject(domCats, function(c){return _.contains(datCats, c);});

    // delete removed categories
    _.each(delCats, function(c){
        nodelistdom.find("[key='" + c + "']").unbind().remove();
        flatnodelistdom.find("[key='" + c + "']").unbind().remove();
        flatnodelistdom.find("[category='" + c + "']").unbind().remove();
    });

    // add new category
    _.each(newCats, function(category){
        nodelistdom.append(_.template(c_tmpl, {category: category}));
        nodelistdom.children("li:last").append("<ul></ul>");
        _.each(_.sortBy(nodedefsPerCategory[category], function(n){return n;}), function(n){
            nodelistdom.find("ul:last").append(_.template(n_tmpl, n));
        });

        flatnodelistdom.append((isFlatShown) ? $(_.template(c_tmpl, {category: category})).hide() : _.template(c_tmpl, {category: category}));
        flatnodelistdom.children("li:last").append("<ul></ul>");
        _.each(_.sortBy(nodedefsPerCategory[category], function(n){return n;}), function(n){
            flatnodelistdom.find("ul:last").append(_.template(n_tmpl, n));
            flatnodelistdom.append((isFlatShown) ? $(_.template(f_tmpl, n)).hide() : _.template(f_tmpl, n));
        });
    });

    // update existing category
    _.each(extCats, function(category){
        var catdom = nodelistdom.find("li[type=category][key='" + category + "'] ul");
        var datName = _.pluck(nodedefsPerCategory[category], "name");
        var domName = _.map(catdom.children("li"), function(l){return $(l).attr("key");});

        if (_.all(_.map(_.zip(datName.sort(), domName.sort()), function(pair){return pair[0] === pair[1];}))){
            return;
        }

        catdom.html("");
        _.each(_.sortBy(nodedefsPerCategory[category], function(n){return n;}),
               function(n){catdom.append(_.template(n_tmpl, n));});

        catdom = flatnodelistdom.find("li[type=category][key='" + category + "'] ul");
        catdom.html("");
        _.each(_.sortBy(nodedefsPerCategory[category], function(n){return n;}),
            function(n){catdom.append(_.template(n_tmpl, n));});

        flatnodelistdom.children(".nodeElement[category='" + category + "']").unbind().remove();
        _.each(_.sortBy(nodedefsPerCategory[category], function(n){return n;}),
            function(n){flatnodelistdom.append((isFlatShown) ? $(_.template(f_tmpl, n)).hide() : _.template(f_tmpl, n));});
    });
    nodelistdom.menu("refresh");
    flatnodelistdom.menu("refresh");
    if(isFlatShown) {
        $("#nodelistSearch input").trigger("keyup", {refresh: true});
    }
    $("#flatnodelist .nodeElement").tooltip({position: {my: 'left center', at: 'right+5 center'}});
    $("#nodelist .nodeElement").tooltip({position: {my: 'left center', at: 'right+5 center'}});

    nodelistSize();
}
function nodelistSize(){
    $("#sidebar").css("width", "auto");
    $("#sheets").css("margin-left", $("#sidebar").width() + "px");
    $("#sidebar").css("width", $("#sidebar").width() + "px");
}
var isSearchEventsAdded = false;
function setSearchEvents(){
    if (isSearchEventsAdded){
        return;
    } else {isSearchEventsAdded = true;}
    //search
    // when the Enter key is pressed
    $("#nodelistSearch").submit(function(event){
        var node = $("#flatNodelist li.active");
        if (node.length > 0){
            $(node).click();
        }
        $("#flatNodelist li").removeClass("active");
        event.preventDefault();
        event.stopPropagation();
        return false;
    });

    //incremental search
    var isPrevMin;
    $("#nodelistSearch input").keyup(function(event, data){
        var doRefresh = data && data.refresh;
        var query = $(this).val().toLowerCase();
        if (event.keyCode === 13){
            userPreference.set("nodelistmin", false);
        }
        if (!doRefresh) {
            if ([13, 37, 38, 39, 40].indexOf(event.keyCode) !== -1){
                return;
            }

            if (!query && prevQuery === query){
                return;
            }
        }

        $(".nodeElement").removeClass("active");
        $(".nodeSelected").removeClass("nodeSelected");

        var nodelist = [];
        if (query === ""){
            isSearching = false;
            $("#nodelist").show();
            $("#flatNodelist").hide();
            nodelistSize();
            userPreference.set("nodelistmin", isPrevMin);
            isPrevMin = undefined;
        } else {
            if (isPrevMin === undefined){
                isPrevMin = userPreference.get("nodelistmin");
            }
            userPreference.set("nodelistmin", false);
            $("#nodelist").hide();
            $("#flatNodelist").show();
            //search

            var searchresults = [];
            var otherresults = [];
            //retrieve search results
            $("#flatNodelist > li").each(function(){
                var key = $(this).attr("key").toLowerCase();
                var pos = key.indexOf(query);
                if (pos === 0){
                    searchresults.push({
                        "key": key,
                        "elem": this,
                        "pos": pos
                    });
                    $(this).show();
                } else if (pos > -1) {
                    otherresults.push({
                        "key": key,
                        "elem": this,
                        "pos": pos
                    });
                    $(this).show();
                } else {
                    $(this).hide();
                }
            });
            //sort alphabetically
            searchresults = _.sortBy(searchresults, "key").reverse();
            otherresults = _.sortBy(otherresults, "key").reverse();
            //append other results to search results
            searchresults = _.union(otherresults, searchresults);
            //display results
            _.each(searchresults, function(res){
                var elem = res.elem;
                var sttag = '<span style="font-weight:bold;">';
                var edtag = "</span><!--searching-->";
                var html = $(elem).html();

                var pos = res.pos;
                if (pos > -1){
                    if (html.indexOf("</ul>") === -1){
                        html = html.replace(sttag, "").replace(edtag, "");
                        var st = pos, ed = pos + query.length;
                        html = html.substring(0, st) + sttag +
                            html.substring(st, ed) + edtag +
                            html.substring(ed);
                        $(elem).html(html);
                    }
                    $("#flatNodelist").prepend(elem);
                }
            });
            if ($("#flatNodelist > li:visible").length === 0){
                $("#nodeNotFound").show();
            } else {
                $("#nodeNotFound").hide();
            }
        }
        nodelistSize();
        prevQuery = query;
    });
    $(document).on("mouseover", ".nodeElement", function(){$(".nodeElement").removeClass("active"); $(this).addClass("active");});
    $(document).on("mouseout",  ".nodeElement", function(){$(this).removeClass("active");});
}

function data2graphviz(){
    var out = "";
    out += "digraph network{\n";
    _.chain(network.links)
     .each(function(link){
        out += link.fromNodeId + "->" + link.toNodeId + ";\n";
    });
    out += "}\n";
}

/**** network manipulation functions with verification ****/
function raise(message, obj){
    if (!message){
        app.set("execstat", "");
        return;
    }
    console.log("Error: " + message + " (" + JSON.stringify(obj) + ")");
    app.set("execstat", "Error: " + message);
}

function hasMandatory(obj, mandatory){
    if (_.chain(mandatory)
        .map(function(m){return _.has(obj, m);})
        .every().value() === false){
        return false;
    }
    return true;
}

function isTypesMatch(type1, type2){
    if (type1 === "any" || type2 === "any"){
        return true;
    }
    return type1 === type2;
}

function newNode(node, tmpnetwork){
    if (!tmpnetwork){
        tmpnetwork = network;
    }
    var result = testNode(node, tmpnetwork);
    raise(result, node);
    if (result !== ""){
        return false;
    }
    tmpnetwork.nodes.push(node);
    return true;
}

function newLink(link, tmpnetwork){
    if (!tmpnetwork){
        tmpnetwork = network;
    }
    var result = testLink(link, tmpnetwork);
    if (result.match(/Link cannot connect the same nodes/)) {
        // "Link cannot connect the same nodes" はエラーメッセージを出さない
        // TerminalCircle内でクリックしたつもりでカーソルが少しでも動くとこのエラーになるため
        raise("", link);
    } else {
        raise(result, link);
    }
    if (result !== ""){
        return false;
    }
    tmpnetwork.links = tmpnetwork.links || [];
    tmpnetwork.links.push(link);
    return true;
}

function newTerminalRole(role, tmpnetwork){
    if (!tmpnetwork){
        tmpnetwork = network;
    }
    var result = testTerminalRole(role, tmpnetwork);
    raise(result, role);
    if (result !== ""){
        return false;
    }
    // console.log(tmpnetwork);
    tmpnetwork.terminalRoles = tmpnetwork.terminalRoles || [];
    tmpnetwork.terminalRoles.push(role);
    updateTerminalRoles(role, "add");
    return true;
}

function newSheet(sheet, tmpnetwork){
    if (!tmpnetwork){
        tmpnetwork = network;
    }
    var result = testSheet(sheet, tmpnetwork);
    raise(result, sheet);
    if (result !== ""){
        return false;
    }
    tmpnetwork.sheets[sheet.name] = sheet;
    return true;
}

function newParameter(param, node){

    // prune parameter types
    var mapper = {
        any: "int",
        String: "string"
    };
    if (mapper[param.type]) {param.type = mapper[param.type];}
    if (param.type.indexOf("Vector") !== -1) {param.type = "object";}

    var result = testParameter(param);
    raise(result, param);
    if (result !== ""){
        return false;
    }

    // new parameter name cannot be duplicated.
    if (_.indexOf(_.pickArray(node.parameters, "name"), param.name) !== -1){
        raise("Parameter name cannot be duplicated", param);
        return false;
    }
    node.parameters.push(param);
    return;
}

function testSelection(param, selection) {
    if (!(selection === param.originalType || selection === "subnet_param")) {
        var err = param.name + " should be " + param.originalType + " or subnet_param.";
        $(this).addClass("error");
        raise(err, param);
    }
}

function testParameter(param){
    // condition: if not empty, the format must be *=*.
    if (!_.isEmpty(param.condition) && param.condition.indexOf("=") === -1 ){
        return param.name + " Parameter coondition must include =";
    }

    // description: no empty
    // if (_.isEmpty(param.description)){
    //     return param.name + " Parameter description cannot be empty";
    // }

    // list       : must be list
    if (!_.isArray(param.list)){
        param.list = [];
        // return param.name + " Parameter list must be array: " + param.list;
    }

    // name       : not empty
    if (_.isEmpty(param.name)){
        return param.name + " Parameter name cannot be empty.";
    }

    // type       : [int, string, bool, object, float, any, String, subnet_param]
    var types = ["int", "string", "bool", "object", "float", "any", "String", "subnet_param", "file"];
    if (_.indexOf(types, param.type) === -1){
        return param.name + " Parameter type unknown (" + param.type + ")";
    }

    // value      : depends on the type
    if (!_.isEmpty(param.value)){
        switch (param.type){
            case "int":
            if (_.isString(param.value) && param.value.match(/^-?\d+$/) === null){
                return param.name + " Parameter format is not integer.";
            }
            param.value = parseInt(param.value, 10);
            if (!_.isNumber(param.value)){return param.name + " Parameter value is not number";}
            if (_.isNaN(param.value)){return param.name + " Parameter value is NaN (int)";}
            if (parseFloat(param.value) % 1 !== 0){return param.name + " Parameter value is not integer";}
            break;

            case "float":
            // Commented some validation since float input is now accepting expotential value
            // which can be string input
            //if (_.isString(param.value) && param.value.match(/^-?[\d.]+$/) === null){
            //    return param.name + " Parameter format is not float.";
            //}
            param.value = parseFloat(param.value);
            //if (!_.isNumber(param.value)){return param.name + " Parameter value is not number";}
            if (_.isNaN(param.value)){return param.name + " Parameter value is NaN (float)";}
            break;

            case "object":
            // < and > must match
            if (param.value.split("<").length != param.value.split(">").length){
                return param.name + " Parameter value parenthesis <> mismatched";
            }
            if (param.value.indexOf(">>") != -1){
                return param.name + " Parameter value >> must be > >.";
            }
            break;
        }
    }
    return "";
}

function testNode(node, tmpnetwork){
    if (!tmpnetwork){
        tmpnetwork = network;
    }
    var mandatory = ["name", "id", "sheet", "index", "pos", "parameters", "inputs", "outputs"];
    if (!hasMandatory(node, mandatory)){
        return node.id + " node does not have mandatory options";
    }

    // name: defined,
    if (_.findObject(nodedefs, "name", node.name) === undefined){
        return node.id + " unknown node name";
    }

    // id  : node_<name>_<index> format
    if (node.id !== "node_" + node.name + "_" + node.index){
        return node.id + " wrong node id format";
    }

    // sheet: in network.sheets, not(node.name === node.sheet)
    if (_.keys(tmpnetwork.sheets).indexOf(node.sheet) === -1){
        return node.id + " node sheet is undefined";
    }
    if (node.name === node.sheet){
        return node.id + " The dynamic node cannot be placed in the same sheet";
    }

    // index: integer, positive, unique
    if (!_.isNumber(node.index)){
        return node.id + " node index must be integer";
    }
    if (node.index <= 0){
        return node.id + " node index must be positive";
    }
    if (_.chain(tmpnetwork.nodes)
        .filter(function(n){return n.name === node.name;})
        .map(function(n){return n.index;}).value().indexOf(node.index) !== -1){
        return node.id + " Node index is duplicated";
    }

    // pos  : positive, two-length array
    // if (!(_.isArray(node.pos) && node.pos.length === 2)){
    //     return node.id + " Node pos must be an array with length = 2";
    // }
    // if (node.pos[0] < 0){return node.id + " node position[0] is not positive";}
    // if (node.pos[1] < 0){return node.id + " node position[1] is not positive";}

    // parameters: array
    if (!_.isArray(node.parameters)){
        return node.id + " Node parameters must be an array!";
    }

    return "";
}

function escapeHTML(text){
    return $("<div>").text(text).html();
}

function testLink(link, tmpnetwork){
    if (!tmpnetwork){
        tmpnetwork = network;
    }

    var mandatory = ["fromNodeId", "fromTermName", "toNodeId", "toTermName", "sheet"];
    if (!hasMandatory(link, mandatory)){
        return "Link does not have mandatory properties";
    }

    var fromNodeObj = _.findObject(tmpnetwork.nodes, "id", link.fromNodeId);
    var toNodeObj = _.findObject(tmpnetwork.nodes, "id", link.toNodeId);

    // fromNodeId: node_*_\d, in nodes
    if (link.fromNodeId.match(/node_.+_\d/) === null){
        return "Link fromNodeId (" + link.fromNodeId + ") is invalid format";
    }
    if (fromNodeObj === undefined){
        return "Link fromNodeId (" + link.fromNodeId + ") is not found in nodes";
    }

    // toNodeId: node_*_\d, in nodes
    if (link.toNodeId.match(/node_.+_\d/) === null){
        return "Link toNodeId (" + link.toNodeId + ")is invalid format";
    }
    if (toNodeObj === undefined){
        return "Link toNodeId (" + link.toNodeId + ") is not found in nodes";
    }

    var fromDef = _.findObject(nodedefs, "name", fromNodeObj.name) || { category: "Undefined" };
    var toDef = _.findObject(nodedefs, "name", toNodeObj.name) || { category: "Undefined" };
    var fromTerms = _.pickArray(fromNodeObj.outputs, "name");
    var toTerms = _.pickArray(toNodeObj.inputs, "name");

    // fromTermName: in nodedef.outputs
    // if (fromTerms.indexOf(link.fromTermName) === -1){
    //     return "Link fromTermName is not defined in outputs";
    // }

    // toTermName: in nodedef.inputs
    // if (toTerms.indexOf(link.toTermName) === -1 &&
    //    toDef.category !== "Dynamic"){
    //     return "Link toTermName is not defined in inputs";
    // }

    // matching: *nodeId cannot be the same.
    if (link.fromNodeId === link.toNodeId){
        return "Link cannot connect the same nodes (" + link.toNodeId + ")";
    }

    // sheetname: fromNodeObj.sheet === toNodeObj.sheet === sheets
    if (fromNodeObj.sheet !== toNodeObj.sheet){
        return "Link cannot connect different sheets";
    }
    if (fromNodeObj.sheet !== link.sheet){
        return "Link node.sheets and link.sheets are different";
    }

    // matching: object_type must be matched.
    var fromType = _.findObject(fromNodeObj.outputs, "name", link.fromTermName);
    var toType   = _.findObject(toNodeObj.inputs, "name", link.toTermName);
    if (!fromType || !toType){
        return "Link cannot connect INPUT-INPUT nor OUTPUT-OUTPUT";
    }

    if ( toDef.category !== "Dynamic" && fromDef.category !== "Dynamic" &&
        !isTypesMatch(fromType.type, toType.type)){
        return "Link input and output types are mismatched " +
                "from type is " + escapeHTML(fromType.type) + ",     but " +
                "to type is " + escapeHTML(toType.type);
    }

    // toTerms cannot have multiple inputs
    var wrongTo = _.find(tmpnetwork.links, function(l){
        return l.toNodeId === link.toNodeId && l.toTermName === link.toTermName;
    });
    if (wrongTo){
        return "Link input terminal (" + wrongTo.toTermName + " of " + wrongTo.toNodeId + ") cannot take multiple links";
    }

    return "";
}

function testTerminalRole(role, tmpnetwork){
    if (!tmpnetwork){
        tmpnetwork = network;
    }
    var mandatory = ["nodeId", "termName", "name", "role", "sheet"];
    if (!hasMandatory(role, mandatory)){
        return "role does not have mandatory properties";
    }
    if (role.role === "input" || role.role === "output"){
        if (!hasMandatory(role, ["object_type", "description"])){
            return "role does not have mandatory properties";
        }
    }

    // nodeId: format, in tmpnetwork.nodes
    if (role.nodeId.match(/node_.+_\d/) === null){
        return "Role nodeId has wrong format";
    }
    var thenode = _.findObject(tmpnetwork.nodes, "id", role.nodeId);
    if (thenode === undefined){
        return "Role nodeId does not exist in the network";
    }

    // role: input/output/condition
    if (["input", "output", "condition"].indexOf(role.role) === -1){
        return "Role role must be input / output / condition";
    }

    // termName: input -> in node.input, output/condition -> in node.output
    // var def = _.find(nodedefs, function(d){return d.name === thenode.name});
    var def = _.findObject(nodedefs, "name", thenode.name) || {
        category: "Undefined"
    };

    if (role.role === "input"){
        if (def.category !== "Dynamic" &&
           _.findObject(thenode.inputs, "name", role.termName) === undefined){
            return "Role termname does not found in inputs";
        }
    } else {
        if (def.category !== "Dynamic" &&
           _.findObject(thenode.outputs, "name", role.termName) === undefined){
            return "Role termname does not found in outputs";
        }
    }

    // sheet in network.sheets, nodeId.sheet === sheet
    if (_.keys(tmpnetwork.sheets).indexOf(role.sheet) === -1){
        return "Role sheet does not exist";
    }

    if (role.sheet !== thenode.sheet){
        return "Role sheet and nodeId.sheet mismatched";
    }

    // name: not empty, no conflict in the same sheet, not empty
    if (_.isEmpty(role.name)){
        return "Role name cannot be empty";
    }

    var existingNames = _.pickArray(_.groupBy(tmpnetwork.terminalRoles, "sheet")[role.sheet], "name");
    if (existingNames.indexOf(role.name) !== -1){
        return "Role name already exist in the sheet";
    }

    // object_type: if not condition --> not empty
    // if (role.role !== "condition" && _.isEmpty(role.object_type)){
    //     return "Role object_type cannot be empty";
    // }
    // description: no check

    return "";

// nodeId   : nodeId,
// termname : termname,
// name     : rolename,
// role     : "input",
// sheet    : sheetname,
// object_type: objtype,
// description: desc
}

function testSheet(sheet, tmpnetwork){
    if (!tmpnetwork){
        tmpnetwork = network;
    }
    var mandatory = ["name", "type"];
    if (!hasMandatory(sheet, mandatory)){
        return "Sheet does not have mandatory property";
    }

    // type cannot be empty
    if (_.isEmpty(sheet.type)){
        return "Sheet type cannot be empty";
    }

    // type must be the list
    var typelist = ["subnet", "iterator"];
    if (typelist.indexOf(sheet.type) === -1){
        return "Sheet type must be subnet or iterator";
    }

    // no duplicate
    if (_.keys(tmpnetwork.sheets).indexOf(sheet.name) !== -1){
        if (sheet.name !== "MAIN"){
            return "Sheet name is duplicated";
        }
    }

    // not empty
    if (_.isEmpty(sheet.name)){
        return "Sheet name cannot be empty";
    }

    return "";
}

function pruneNetwork(tmpnetwork){
    // var nodeIds = _.pickArray(network.nodes, "id");
    // network.links = _.chain(network.links)
    // .filter(function(l){
    //     return nodeIds.indexOf(l.fromNodeId) !== -1 &&
    //            nodeIds.indexOf(l.toNodeId);
    // })
    // .filter(function(l){
    //     var node = _.findObject(network.nodes, "id", l.fromNodeId);
    //     return _.findObject(node.outputs, "name", l.fromTermName);
    // })
    // .filter(function(l){
    //     var node = _.findObject(network.nodes, "id", l.toNodeId);
    //     return _.findObject(node.inputs, "name", l.toTermName);
    // }).value();

    _.each(tmpnetwork.nodes, function(n){
        n.pos[0] = truncate(n.pos[0]);
        n.pos[1] = truncate(n.pos[1]);
    });

    if (typeof(tmpnetwork.links) === "undefined"){
        tmpnetwork.links = [];
    }
    if (typeof(tmpnetwork.terminalRoles) === "undefined"){
        tmpnetwork.terminalRoles = [];
    }
}

function setupSavePerSheet(sheet) {
    var sheetname_ = sheet;
    var newNetwork = jQuery.extend(true, {}, network);

    $("#saveFilePerSheetDialog").submit(function(event){
        if (app){
            $($('#saveFilePerSheetDialog').siblings('.ui-dialog-buttonpane').find("button")[0]).click();
        }
        event.preventDefault();
        event.stopPropagation();
    });
    $("#saveFilePerSheetDialog").dialog({
        autoOpen: false,
        modal: true,
        buttons : {
            OK: function(){
                var that = this;
                var filename = $("#saveFilePerSheetDialog input[name=name]").val();
                if(filename.slice(-2) !== ".n") {
                    filename+= ".n";
                }
                if (_.isEmpty(filename)){
                    alert("You must specify the file name");
                    return;
                }

                $.ajax({
                    type: "GET",
                    url: rootPath + "filename/"
                }).done(function(data){

                    // replace all "file" type parameters to "string"
                    _.each(newNetwork.nodes, function(n){
                        _.each(n.parameters, function(p){
                            if (p.type === "file"){
                                p.type = "string";
                                // console.log(p);
                            }
                        });
                    });
                    // console.log(JSON.stringify({object: network}));
                    app.set("currentFileName", filename);

                    var activeNodes = _.filter(newNetwork.nodes, function(n) {return n.sheet == sheetname_;});
                    var activeLinks = _.filter(newNetwork.links, function(n) {return n.sheet == sheetname_ ;});
                    var activeTerminalRoles = _.filter(newNetwork.terminalRoles, function(n) {return n.sheet == sheetname_ ;});
                    var activeNotes = _.filter(newNetwork.notes, function(n) {return n.sheet == sheetname_ ;});

                    newNetwork.nodes = activeNodes;
                    newNetwork.links = activeLinks;
                    newNetwork.terminalRoles = activeTerminalRoles;
                    newNetwork.notes = activeNotes;

                    // check if data already contains a file the same with filename
                    var isContained = _.contains(data, filename);

                    _.each(data, function (item) {
                        var willwrite = false;
                        if (item === filename) {
                            if (confirm("Overwrite network?")) {
                                willwrite = true;
                            }
                        }
                        else {
                            if(!isContained) {
                                willwrite = true;
                            }
                        }

                        if (willwrite) {
                            $.ajax({
                                url: rootPath + "file/object/sheet/" + filename,
                                type: "PUT",
                                data: newNetwork,
                                // dataType: "json"
                            }).done(function(){
                                app.set("execstat", "Saved");
                                loadNodeListAjax('user-generated.def', true);
                                updateFileList();
                                $(that).dialog("close");
                            }).fail(function(data){
                                console.log("failed", data);
                                app.set("execstat", data.responseText);
                            });
                        }
                    });
                }).fail(function(err) {
                    console.log(err);
                });
            },
            Cancel: function(){$(this).dialog("close");}
        }
    });
}

function setupSaveFile(){
    $("#saveFileDialog").submit(function(event){
        if (app){
            $($('#saveFileDialog').siblings('.ui-dialog-buttonpane').find("button")[0]).click();
        }
        event.preventDefault();
        event.stopPropagation();
    });
    $("#saveFileDialog").dialog({
        autoOpen: false,
        modal: true,
        buttons : {
            OK: function(){
                var $saveDialog = this;
                fileCollection
                    .fetch({reset: true}) // Fetches data without pagination
                    .done(function () {
                        var that = $saveDialog;
                        var filename = $("#saveFileDialog input[name=name]").val();

                        if (_.isEmpty(filename)){
                            alert("You must specify the file name");
                            return;
                        }
                        if (_.pluck(fileCollection.toJSON(), "name").indexOf(filename) !== -1 &&
                            app.get("currentFileName") !== filename){
                            if (!confirm("Overwrite network?")) {return;}
                        }
                        app.set("currentFileName", filename);

                        // replace all "file" type parameters to "string"
                        _.each(network.nodes, function(n){
                            _.each(n.parameters, function(p){
                                if (p.type === "file"){
                                    p.type = "string";
                                }
                            });
                        });

                        revertSheetNames();

                        $.ajax({
                            url: rootPath + "file/object/" + encodeURIComponent(filename), // Encondes fileaname
                            type: "PUT",
                            data: network,
                            // dataType: "json"
                        }).done(function(){
                            app.set("execstat", "Saved");
                            updateFileList();
                            $(that).dialog("close");

                            correctSheetNames();
                        }).fail(function(data){
                            console.log("failed", data);
                            app.set("execstat", data.responseText);
                        });
                    });
            },
            Cancel: function(){$(this).dialog("close");}
        }
    });
    $("#save")
        .click(function(event){
            $("#saveFileDialog").dialog("open");
        });
}

/* update canvas size */
function updateSheetsSize(){
    _.each($("#sheets canvas"), function(canvas){
        canvas = $(canvas);
        canvas.attr("height", canvas.css("height"));
        canvas.attr("width", canvas.css("width"));
    });

    // one implementation: canvas size = sidebar size
    // var side = parseInt($("#sidebar").css("height"), 10);
    // _.each($("#sheets .sheetContainer"), function(div){
    //     $(div).css("height", (side - 50) + "px");
    // });
    // $("#sheets").css("height", side + "px");
    // return;

    // another implementation: canvas size = window size
    var body = $("body").height();
    var footer = $("#footer").height();
    var padding = 30;
    // console.log(body, footer, padding);
    _.each($("#sheets .sheetContainer"), function(div){
        $(div).css("height", (body - $(div).offset().top - footer - padding) + "px");
    });

    $("#nodelistContainer").css("height",
        (body - $("#nodelistContainer").offset().top - footer - padding) + "px");
}
/* setup save network to file list */
function updateFileList(){
    var that = $(document).find("#fileManagerDialog .dialogMain");
    var filter = that.parent().find(".backgrid-filter");
    var paginator = that.parent().find(".backgrid-paginator");

    // Inserts loading icon
    that.find("#grid").html(_.template($("#spinnerTemplate").html()));

    // Fetches data from backend for this collection
    pageableFileCollection.fetch({
        url: rootPath + "filelist",
        reset: true,
        add: true,
    }).done(function () {
        // Unhide the filter and pagination
        $.when(filter.css("visibility", ""), paginator.css("visibility", ""))
            .then(function () {
                // Inserts the file list
                that.find("#grid").html(fileListGrid.$el);
            });
    }).fail(function () {
        // Hides the filter and pagination
        $.when(filter.css("visibility", "hidden"), paginator.css("visibility", "hidden"))
            .then(function () {
                // Inserts the error message
                that.find("#grid").html("<p style='text-align:center'>Error: Directory not found.</p>");
            });

        // Display error message in console
        app.set("execstat", "");
        app.set("execstat", "Error: Directory not found.");
    });
}

function setupPreference(){
    $("#preferenceButton").click(function(event){
        $("#preferenceDialog").dialog("open");
    });

    $("#preferenceDialog").dialog({
        autoOpen: false,
        modal: true,
        minWidth: 700,
        open: function(){
            _.each($("#preferenceDialog .dialogMenu button"), function(li){
                $(li).css("width", "100%");
            });

            $("#preferenceDialog .dialogMain > div").css("display", "none");
        },
        close: function(){
            $("#preferenceDialog .dialogMenu button").removeClass("ui-state-highlight");
            $("#preferenceDialog .dialogMain > div").css("display", "none");
        }
    });

    $("#preferenceDialog .package").on("change", "form", function(event){
        var newTheme = event.target.value;
        var selectedDef = $(event.target).siblings("label").text();

        _.chain(nodedefs)
         .filter(function(d){return d.def === selectedDef;})
         .each(function(d){d.theme = newTheme;});

         _.chain(network.nodes)
          .filter(function(n){return n.def === selectedDef;})
          .each(function(n){
            $("#" + n.id).removeClass("nodeTheme" + n.theme);
            $("#" + n.id).addClass("nodeTheme" + newTheme);
            n.theme = newTheme;
          });
    });
    var prefEnvironment = function(data){
        function addVarForm(that, name, variable){
            var num = $("#preferenceDialog .dialogMain .setenv fieldset").length || 1;
            if (!_.isNumber(num)){
                num = _.max(_.map(num, function(n){
                    console.log($(n).attr("id").replace("variable", ""), 10);
                    return parseInt($(n).attr("id").replace("variable", ""), 10) + 1;}));
            }

            var html = _.template(
                '<fieldset id="variable<%= num %>" class="val">' +
                '    <input id="varname<%= num %>" <% if (name) {%> value="<%= name %>" <% } %>>' +
                '    <input id="varval<%= num %>"  <% if (variable) {%> value="<%= variable %>" <% } %>>' +
                '    <button id="delvar<%= num %>">' +
                '</fieldset>',
                {num: num, name: name, variable: variable});
            $(that).siblings(".setenv").append(html);
            $("#preferenceDialog .dialogMain #delvar" + num).button({
                text: false,
                icons: {primary: "ui-icon-minusthick"}
            }).click(function(event){
                $(this).parent().unbind().remove();
            });
        }

        $("#preferenceDialog .dialogMain .package").html(data);
        $("#preferenceDialog .dialogMain .package").css("display", "block");
        $("#preferenceDialog .dialogMain .addvar").button({
            label: "Add new variable",
            icons: {primary: "ui-icon-plusthick"}
        }).click(function(event){
            addVarForm(this);
        });

        _.each(_.keys(userPreference.get("env") || {}), function(e){
            if (e === "rosVersionInternal") {return;}
            addVarForm($("#preferenceDialog .dialogMain .addvar"),
                       e, userPreference.get("env")[e]);
        });

        $("#preferenceDialog .dialogMain .setvariables").button().click(
            function (event){
                var variables = {};
                variables.rosVersionInternal = $(this).siblings(".setros").find("select").val();

                _.each($("#preferenceDialog .dialogMain .setenv fieldset.val"), function(f){
                    var inputs = $(f).children("input");
                    variables[$(inputs[0]).val()] = $(inputs[1]).val();
                });
                console.log(variables);
                userPreference.set("env", variables);
                $.ajax({
                    url: "/preference/environment",
                    method: "PUT",
                    data: variables
                }).always(function(data){
                    $(this).siblings(".msg", data);
                });
            }
        );

        $("#preferenceDialog .dialogMain .mypkg .file").fileupload({
            // dataType: "json",
            prependFiles: true,
            done: function(e){
                console.log("mypkg done");
                updateMyPkgList();
            },
            add: function(e, data){
                data.url = "/package/" + data.files[0].name;
                data.submit();
            }
        });

        // load localStorage
        if (userPreference.get("env")){
            var variables = userPreference.get("env");
            if (variables.rosVersionInternal){
                $("#preferenceDialog .dialogMain .setros select").val(variables.rosVersionInternal);
            }
            // delete variables["rosVersionInternal"];
            // var varnames = _.keys(variables);
            // for (var i = 0; i < varnames.length; i++){
                // $("#preferenceDialog .dialogMain .addvar").trigger("click");
                // var ins = $("#preferenceDialog .dialogMain fieldset:last");
                // $($(ins).find("input")[0]).val(varnames[i]);
                // $($(ins).find("input")[1]).val(variables[varnames[i]]);
            // }
        }
        updateMyPkgList();
    };

    var updateMyPkgList = function(e, data){
        $.ajax({
            url: "/package",
            method: "GET"
        }).done(function(data){
            $("#preferenceDialog .dialogMain .mypkglist button").button("destroy");
            $("#preferenceDialog .dialogMain .mypkglist").html("");
            var html = "<tr><th>Package name</th><th>.tlb</th><th>.def</th><th></th></tr>";
            _.each(_.keys(data), function(pkg){
                html += "<tr><td>" + pkg + "</td>";
                if (_.indexOf(data[pkg], "tlb") !== -1){
                    html += "<td>OK</td>";
                } else {
                    html += "<td>Missing</td>";
                }

                if (_.indexOf(data[pkg], "def") !== -1) {
                    html += "<td>OK</td>";
                } else {
                    html += "<td>Missing</td>";
                }
                html += "<td><button></button></td>";
                html += "</tr>";
             });
            $("#preferenceDialog .dialogMain .mypkglist").append(html);
            $("#preferenceDialog .dialogMain .mypkglist button").button({
                text: false,
                icons: {primary: "ui-icon-closethick"}
            }).click(function(e){
                var pkg = $($($(this).parentsUntil("tr").parent()).children("td")[0]).text();
                userPreference.delPackage(pkg + ".def");
                console.log(pkg);
                $.ajax({
                    url: "/package/" + pkg,
                    method: "DELETE"
                }).always(function(data){
                    console.log("done");
                    updateMyPkgList();
                    updateNodeList();
                });
            });
        });
    };

    var prefPackage = function(data){
        $("#preferenceDialog .dialogMain .package").html(data);
        $("#preferenceDialog .dialogMain .package").css("display", "block");

        var maxw = _.chain($("#preferenceDialog .dialogMain .package button"))
                    .map(function(b){return $(b).next().width();})
                    .max().value();
        _.chain($("#preferenceDialog .dialogMain .package button"))
         .each(function(b){
            $(b).next().css("width", maxw * 1.1)
                       .css("margin-left", "10px");
        });

        $("#preferenceDialog").dialog("option", "width", maxw * 1.1);
        _.each($("#preferenceDialog .dialogMain .package button"),
            function(b){
                var def = $(b).next().text();
                var icon = "ui-icon-bullet";
                if (userPreference.get("packages").indexOf(def) === -1){
                    icon = "ui-icon-radio-off";
                    $(b).siblings("select").attr("disabled", "disabled");
                    $(b).siblings("label").css("color", "#AAA");
                }

                var tmpdef = _.find(nodedefs, function(d){return d.def === def;});
                if (tmpdef){
                    $(b).siblings("select").val(tmpdef.theme);
                }

                $(b).button({
                    text: false,
                    icons: {primary: icon}
                }).click(function(event){
                    var isExist = userPreference.get("packages").indexOf(def) !== -1;
                    if (isExist){
                        // nodedefs = _.reject(nodedefs, function(d){return d.def === def;});
                        // userPreference.set("packages",
                        //     _.without(userPreference.get("packages"), def));
                        userPreference.delPackage(def);
                        $(this).children("span").removeClass("ui-icon-bullet");
                        $(this).children("span").addClass("ui-icon-radio-off");
                        $(b).siblings("select").attr("disabled", "disabled");
                        $(b).siblings("label").css("color", "#AAA");
                    } else {
                        if (userPreference.get("packages").indexOf(def) === -1){
                            // console.log("loading");
                            loadNodeListAjax(def);
                            userPreference.addPackage(def);
                            $(this).children("span").addClass("ui-icon-bullet");
                            $(this).children("span").removeClass("ui-icon-radio-off");
                            $(b).siblings("select").removeAttr("disabled");
                            $(b).siblings("label").css("color", "#000");
                        }
                    }

                    if(localStorage.userPref && JSON.parse(localStorage.userPref).origPackages) {
                        //retrieve local storage for user preferences
                        var value = JSON.parse(localStorage.userPref);
                        //validate if all packages are enabled
                        toggleDisplayInfoIcon(value.packages.length === value.origPackages.length);
                    }

                    updateNodeList();
                    event.stopPropagation();
                    event.preventDefault();
                });
            }
        );
    };

    var prefVersion = function(data){
        console.log(data);
        console.log(data.replace(/\n/g, "<br>"));
        $("#preferenceDialog .dialogMain .version").html(data.replace(/\n/g, "<br>"));
        $("#preferenceDialog .dialogMain .version").css("display", "block");
    };

    $("#preferenceDialog .dialogMain .display select[name=line]").on("change", function(event){
        console.log(event.target.value);
        userPreference.set("link", event.target.value);
        clearCanvas();
        drawLinks();
    });
    $("#preferenceDialog .dialogMain .display select[name=lang]").on("change", function(event){
        console.log(event.target.value);
        userPreference.set("lang", event.target.value);
    });
    $("#preferenceDialog .dialogMain .display select[name=tooltip]").on("change", function(event){
        console.log(event.target.value);
        userPreference.set("showTooltip", event.target.value);
        if (event.target.value === "Yes"){
            $(document).tooltip("enable");
        } else {
            console.log("destroyed");
            $(document).tooltip("disable");
        }
    });
    $("#preferenceDialog .dialogMain .display select[name=candidate]").on("change", function(event){
        console.log(event.target.value);
        userPreference.set("showCandidate", event.target.value);
    });
    $("#preferenceDialog .dialogMain .display select[name=realign]").on("change", function(event){
        console.log(event.target.value);
        userPreference.set("reAlign", event.target.value);
    });
    $("#preferenceDialog .dialogMenu button").each(function(li){
        $(this).button().click(function(event){
            $("#preferenceDialog .dialogMenu button")
                .removeClass("ui-state-highlight");
            $(this).addClass("ui-state-highlight");

            $("#preferenceDialog .dialogMain > div").css("display", "none");
            var cmd = $(this).text().toLowerCase();
            if (cmd === "packages"){
                $.ajax({url: rootPath + "preference/" + cmd}).done(prefPackage);
            } else if (cmd === "version"){
                $.ajax({url: rootPath + "preference/" + cmd}).done(prefVersion);
            } else if (cmd === "display"){
                $("#preferenceDialog .dialogMain .display").css("display", "block");
            } else if (cmd === "all reset"){
                if (confirm("Do you really want to reset all preferences?")){
                    localStorage.clear();
                    window.location.reload();
                    window.onbeforeunload = undefined;
                }
            } else if (cmd === "environment"){
                console.log(cmd);
                $.ajax({url: rootPath + "preference/" + cmd}).done(prefEnvironment);
            }

        });
    });
    $("#tutorialButton").button().click(function(){
        $("#preferenceDialog").dialog("close");
        introJs().start();
    });
    $("#packagesButton").button({
        icons :{
            secondary: "ui-icon-info"
        }
    });

    //hide icon info of packagesButton as default display
    $("#packagesButton > span.ui-button-icon-secondary.ui-icon.ui-icon-info").hide();
    $("#packagesButton > span:first-child").css("padding-left", "27.28px").css("padding-right", "27.28px");

    if(localStorage.userPref && JSON.parse(localStorage.userPref).origPackages) {
        //retrieve local storage for user preferences
        var value = JSON.parse(localStorage.userPref);
        //validate if all packages are enabled
        toggleDisplayInfoIcon(value.packages.length === value.origPackages.length);
    }
}

function toggleDisplayInfoIcon(willNotDisplay) {
    if(willNotDisplay) {
        //remove info icon in Packages option
        $("#packagesButton > span.ui-button-icon-secondary.ui-icon.ui-icon-info").hide();
        $("#packagesButton > span:first-child").css("padding-left", "27.28px").css("padding-right", "27.28px");
    } else {
        //display info icon in Packages option
        $("#packagesButton > span.ui-button-icon-secondary.ui-icon.ui-icon-info").show();
        $("#packagesButton > span:first-child").css("padding-left", "17.6px").css("padding-right", "36.96px");
    }
}

function setupFileManager(){
    var datapath = $("#fileManagerDialog").find("#datapath");

    $("#fileManagerButton").click(function(event){
        updateFileList();
        datapath.val(base);
        $("#fileManagerDialog").dialog("open");
    });

    // Binds event for clicking home icon
    $("#fileManagerDialog").find("#homeDir").on("click", function () {
        // Gets the default directory
        $.ajax({
            url: rootPath + "datapath",
            method: "GET"
        }).done(function(res){
            datapath.val(res);
            datapath.trigger("apply");
        }).fail(function (data) {
            // Display error message in console
            app.set("execstat", "Error: " + (data.responseText) ? data.responseText : data.statusText);
        });
    });

    datapath.on("apply", function(event){
        // Sets root dynamically
        $.ajax({
            url: '/dirlist/set_dir',
            type: 'POST',
            data: { dir : $(this).val() },
        }).done(function (data) {
            datapath.val(data.root);
            base = data.root;
            app.set("execstat", "");
            updateFileList();
            localStorage.setItem('basePath', base);
        }).fail(function (data){
            datapath.val(base);
            // Display error message in console
            app.set("execstat", "Error: Directory not found.");
        });
    });

    // When pressing "return" triggers blur to set root dynamically
    datapath.on("keypress", function (e) {
        if (e.which === 13) {
            $(this).trigger("apply");
        }
    });
}

function setupFileTree() {
    // Setup open directory button
    var $el = $("#fileManagerDialog").find("#openDir");

    // Binds click event
    $el.on("click", function () {
        var dir = $("#fileManagerDialog").find("#datapath").val();
        var fileTree = $('#fileTree').jstree(true);

        // Feeds data to jstree
        fileTree.settings.core.data = {
            url: '/dirlist?operation=get_node',
            data: function (node) {
                return {dir: encodeURIComponent(dir),id: encodeURIComponent(node.id)};
            },
            error: function (resp) {
                $("#fileTreeDialog").dialog("close");
                app.set("execstat", "");
                app.set("execstat", "Error: " + resp.responseText);
            }
        };

        fileTree.refresh();

        $("#fileTreeDialog").dialog("open");
    });
}

function setupButtons(){
    $("#clear").click(function(event){
        if (confirm("Are you sure to clear current network?")){
            reset();
        }
    });

    $("#helpButton").click(function(event){
        window.open("http://www.hark.jp/document/hark-designer-" + userPreference.get("lang"));
    });
}

function setupStatus(){
    $("#clearExecuteStatus").button().click(function(e){
        $("#executeStatus").text("");
        app.set("execstat", "");
        $(this).css("visibility", "hidden");
    });
    $("#clearExecuteStatus .ui-button-text").css("padding", "2px");
    $("#clearExecuteStatus .ui-button-text").css("font-size", "15px");
    $("#clearExecuteStatus").css("visibility", "hidden");
    $("#clearExecuteStatus").css("float", "right");

    $("#executeStatus").parent().  on("mouseover", function(event){
        if ($("#executeStatus").text()){
            $("#clearExecuteStatus").css("visibility", "visible");
        }
    });
}

function findEdgeObj(obj, direction, index) {
    var id;
     var edgeObject = _.min(obj, function(n){
        if(_.has(n, 'noteId')) {
            id = 'noteId';
        } else {
            id = 'id';
        }
        if(direction === 'right') {
            return $('#canvas_' + getActiveSheet()).width() - ($('#' + n[id]).width() + n.pos[index]);
        } else if(direction === 'down') {
            return $('#canvas_' + getActiveSheet()).height() - ($('#' + n[id]).height() + n.pos[index]);
        }
    });
    return edgeObject;
}

function adjustPosition(selectedObject, posVal, posIndex) {
    var id;
    _.each(selectedObject, function(n){
        if(_.has(n, 'noteId')) {
            id = 'noteId';
        } else {
            id = 'id';
        }
        n.pos[posIndex] += posVal;
        $("#" + n[id]).css({left: n.pos[0], top: n.pos[1]});
    });
}

function move(obj, direction, index) {
    var edgeObj = {};
    var boundary;

    if(direction === 'left' || direction === 'up') {
        edgeObj = _.min(obj, function(n){
            return n.pos[index];
        });
    } else if(direction === 'right') {
        edgeObj = findEdgeObj(obj, 'right', 0);
    } else if(direction === 'down') {
        edgeObj = findEdgeObj(obj, 'down', 1);
    }
    boundary = getBoundary(edgeObj, direction);

    if(edgeObj.pos[index] !== boundary) {
        switch(direction) {
            case 'left':
                adjustPosition(obj, -10, 0);
                break;
            case 'up':
                adjustPosition(obj, -10, 1);
                break;
            case 'right':
                adjustPosition(obj, 10, 0);
                break;
            case 'down':
                adjustPosition(obj, 10, 1);
                break;
        }
    }
}

function getBoundary(obj, direction) {
    var boundary;
    var elementSize;
    var id;

    if(_.has(obj, 'noteId')) {
        id = 'noteId';
    } else {
        id = 'id';
    }

    if(direction === 'left' || direction === 'up') {
        boundary = 0;
    } else if(direction === 'right') {
        elementSize = $('#' + obj[id]).width();
        boundary = $('#canvas_' + getActiveSheet()).width() - elementSize;
    } else if(direction === 'down') {
        elementSize = $('#' + obj[id]).height();
        boundary = $('#canvas_' + getActiveSheet()).height() - elementSize;
    }

    boundary = parseInt(boundary / 10, 10) * 10;

    return boundary;
}

function setupNodeKeyBindings(){
    $(document).keydown(function(event){
        // skip if any dialog is opened
        var opened = _.filter($(".ui-dialog"), function(d){return $(d).css("display") !== "none";}).length;
        var $sheetTab = $("[aria-controls='#sheet_" + getActiveSheet() + "']");
        if (opened > 0){
            return true;
        }

        var selected = getSelectedObjs();
        var selectedArray = selected.nodes.concat(selected.notes);

        // $(".ui-state-default:focus").blur();
        var doNothing = false;

        var key = event.keyCode;

        if ([37, 9, 38, 39, 40, 27, 46].indexOf(key) !== -1){
            event.preventDefault();
            event.stopPropagation();
        }

        if (selected.nodes.length > 0 || selected.notes.length > 0){
            switch (key){
                case 67: //Ctrl + C
                if (event.ctrlKey){
                    copyObjsToClipboard(selected);
                }
                break;

                case 88: //Ctrl + X
                // alert("Currently, Cut is disabled. Use Ctrl + C instead");
                if (event.ctrlKey){
                    cutObjsToClipboard(selected);
                }
                break;

                case 9: //Tab
                _.each(selected.nodes, function(selectedNodeObj){
                    $(".nodeSelected").removeClass("nodeSelected");
                    var shownNodes = _.filter(network.nodes, function(n){return n.sheet === currentTab;});
                    var tmp = _.chain(shownNodes)
                               .pluck("id").indexOf(selectedNodeObj.id).value();
                    console.log(tmp + "--" + shownNodes.length);
                    if (tmp == shownNodes.length - 1){
                        selectedNodeObj = undefined;
                        $("#nodelistSearch input").focus();
                        console.log($("#nodelistSearch input"));
                    } else {
                        selectedNodeObj = shownNodes[tmp + 1];
                        $("#" + selectedNodeObj.id).addClass("nodeSelected");
                    }
                    adjustScroll(currentTab, function(n){return n.hasClass("nodeSelected") || n.hasClass("noteSelected");});
                });
                break;

                case 37: //left
                case 72: // h
                    if(!$sheetTab.is(":focus")) {
                        move(selectedArray, 'left', 0);
                    }
                break;

                case 38: //up
                case 75: //k
                    if(!$sheetTab.is(":focus")) {
                        move(selectedArray, 'up', 1);
                    }
                break;

                case 39: //right
                case 76: //l
                    if(!$sheetTab.is(":focus")) {
                        move(selectedArray, 'right', 0);
                    }
                break;

                case 40: //down
                case 74: //j
                    if(!$sheetTab.is(":focus")) {
                        move(selectedArray, 'down', 1);
                    }
                break;

                case 27: //escape
                _.each(selected.nodes, function(selectedNodeObj){
                    $("#" + selectedNodeObj.id).removeClass("nodeSelected");
                    selectedNodeObj = undefined;
                });

                _.each(selected.notes, function(n){
                    $("#" + n.noteId).removeClass("noteSelected");
                });
                break;

                case 46: //delete
                    deleteObjs(selected);
                break;

                case 13: //enter
                var poses = [];
                _.each(selected.nodes, function(selectedNodeObj){
                    var propertyDialog = $("#" + selectedNodeObj.id + delim + "parameters");
	                var isSamePosition = function(n){
	                        return (n.left == propertyDialog.offset().left && n.top == propertyDialog.offset().top);
	                    };
                    propertyDialog.dialog("open");
                    propertyDialog = propertyDialog.closest('.ui-dialog');
                    for (var i = 0; i < 1000; i++) {
                        if (_.find(poses, isSamePosition)) {
                            propertyDialog.offset({
                                left: propertyDialog.offset().left + 10,
                                top: propertyDialog.offset().top + 10
                            });
                        } else {
                            break;
                        }
                    }
                    poses.push(propertyDialog.offset());
                });

                _.each(selected.notes, function(selectedNoteObj){
                    var propertyDialog = initNote(selectedNoteObj.noteId, false);
                    var isSamePosition = function(n){
                        return (n.left == propertyDialog.offset().left && n.top == propertyDialog.offset().top);
                    };

                    propertyDialog.dialog("open");
                    propertyDialog = propertyDialog.closest('.ui-dialog');
                    for (var i = 0; i < 1000; i++) {
                        if (_.find(poses, isSamePosition)) {
                            propertyDialog.offset({
                                left: propertyDialog.offset().left + 10,
                                top: propertyDialog.offset().top + 10
                            });
                        } else {
                            break;
                        }
                    }
                    poses.push(propertyDialog.offset());
                });

                event.preventDefault();
                event.stopPropagation();
                break;

                default:
                doNothing = true;
                break;
            }
            if (!doNothing){
                clearCanvas();
                drawLinks();
            }
        } else if ($("#nodelistSearch input:focus")){
            $(".nodeSelected").removeClass("nodeSelected");

            var selectedList = $("#flatNodelist li.active:visible");
            var visibleList = $("#flatNodelist li:visible");
            var selectedIndex;
            for (var i = 0; i < visibleList.length; i++){
                if ($(visibleList[i]).text() === selectedList.text()){
                    selectedIndex = i;
                }
            }
            switch (key){
                case 9: // tab
                if (network.nodes.length > 0 || network.notes.length > 0){
                    $("#nodelistSearch input").blur();

                    var shownNodes = _.filter(network.nodes, function(n){return n.sheet === currentTab;});

                    if(shownNodes[0]) {
                        $("#" + shownNodes[0].id).addClass("nodeSelected");
                    }
                    adjustScroll(currentTab, function(n){return n.hasClass("nodeSelected") || n.hasClass("noteSelected");});
                }
                break;

                case 38: //up
                if (selectedIndex === undefined){
                    $("#flatNodelist li:visible").last().addClass("active");
                } else {
                    $("#flatNodelist li").removeClass("active");
                    if (selectedIndex === 0){
                        $(visibleList[visibleList.length - 1]).addClass("active");
                    } else {
                        $(visibleList[selectedIndex - 1]).addClass("active");
                    }
                }
                break;

                case 40: //down
                if (selectedIndex === undefined){
                    $("#flatNodelist li:visible").first().addClass("active");
                } else {
                    $("#flatNodelist li").removeClass("active");
                    if (selectedIndex === visibleList.length - 1){
                        $(visibleList[0]).addClass("active");
                    } else {
                        $(visibleList[selectedIndex + 1]).addClass("active");
                    }
                }
                break;
            }
        }

        if (event.ctrlKey && event.keyCode === 83){ //Ctrl + S
            event.preventDefault();
            if (app.get("currentFileName") === ""){
                $("#save").click();
            } else {
                $.ajax({
                    url: rootPath + "file/object/" + app.get("currentFileName"),
                    type: "PUT",
                    data: network,
                }).done(function(){
                    app.set("execstat", "Saved");
                    updateFileList();
                }).fail(function(data){
                    console.log("failed", data);
                    app.set("execstat", data.responseText);
                });
            }
        }
        if (event.ctrlKey && event.keyCode === 86){ //Ctrl + V
            pasteClipboard();
        }

        if (event.ctrlKey && event.keyCode === 65 && $("#nodelistSearch input:focus").length === 0){ //Ctrl + A
            $(".nodeSelected").removeClass("nodeSelected");
            $("#nodelistSearch input").blur();
            _.chain(network.nodes)
             .filter(function(n){return n.sheet === currentTab;})
             .each(function(n){
                $("#" + n.id).addClass("nodeSelected");
             });
            event.preventDefault();
            event.stopPropagation();
        }
    });
}

var adjustScroll = _.throttle(_adjustScroll, 500);
function _adjustScroll(sheetname, filter){
    sheetname = (sheetname)?sheetname.replace(/ /g, ""):null || currentTab;
    filter = filter || function(n){return true;};
    var margin = 10;
    var poses = _.chain($("#canvas2_" + sheetname).find(".node:visible,.note:visible"))
                 .map(function(n){return $(n);})
                 .filter(filter)
                 .map(function(n){return _.extend(n.position(),
                                                  {obj: n});}).value();
    var lefts = _.pluck(_.sortBy(poses, function(p){return p.left;}), "left");
    var tops  = _.pluck(_.sortBy(poses, function(p){return p.top ;}), "top");
    var scroll = {scrollLeft: $("#sheet_" + sheetname).scrollLeft(),
                  scrollTop : $("#sheet_" + sheetname).scrollTop()};

    var obj = $("#sheet_" + sheetname);
    var max = {
        top:  obj.height() + obj.offset().top,
        left: obj.width()  + obj.offset().left
    };

    // top
    // if (_.chain(lefts).map(function(p){return p - margin < 0;}).all().value()){
    //     scroll.scrollLeft += _.min(lefts) - margin;
    // }
    // else if (_.chain(lefts).map(function(p){return p + margin > max.left;}).all().value()){
    //     scroll.scrollLeft += _.max(lefts) - margin;
    // }
    // if (_.chain(tops).map(function(p){return p - margin < 0;}).all().value()){
    //     scroll.scrollTop += _.min(tops) - margin;
    // }
    // else if (_.chain(tops ).map(function(p){return p + margin > max.top ;}).all().value()){
    //     scroll.scrollTop += _.max(tops) - margin;
    // }
    if (!_.sortBy(poses, function(p){return p.left;})[0]) {return;}
    var leftest = _.sortBy(poses, function(p){return p.left;})[0].obj;
    var toppest = _.sortBy(poses, function(p){return p.top; })[0].obj;
    // console.log(leftest, toppest);
    scroll.scrollLeft = - ($("#sheets").width()  - leftest.width() ) / 2 +
                           leftest.position().left;
    scroll.scrollTop  = - ($("#sheets").height() - toppest.height()) / 2 +
                           toppest.position().top;

    scroll.scrollLeft = (scroll.scrollLeft > 0) ? scroll.scrollLeft : 0;
    scroll.scrollTop = (scroll.scrollTop > 0) ? scroll.scrollTop : 0;

    // console.log(scroll);
    $("#sheet_" + sheetname).animate(scroll, {easing: "easeOutCubic"});
}

function setupIntroJS(){
    var counter = 1;
    function helper(selector, description){
        $(selector).attr("data-intro", description);
        $(selector).attr("data-step", counter);
        counter += 1;
    }
    $("*").removeAttr("data-intro");
    $("*").removeAttr("data-step");
    if (userPreference.get("lang") === "en"){
        helper("#sheets", "You can build your network here.");
        helper("#addSheetButton", "You can add a new sheet from here");
        helper("#reshapeButton", "You can reshape the network (Graphiz required)");
        helper("#sidebar", "Here is the list of available nodes");
        helper("#nodelistSearch input", "You can search a node from here.");
        helper("#saveButton", "You can save current network here.");
        helper("#executeButton", "You can also run your network");
        helper("#console", "If you have some error, the message will be displayed here");
        helper("#fileManagerButton", "You can upload/download all of your files from here");
        helper("#preferenceButton", "You can configure, e.g., look&feel, language");
        helper("#zoomSlider", "You can zoom up or zoom out with this slider");
    } else if (userPreference.get("lang") === "ja"){
        helper("#sheets", "この画面でネットワークを作成できます");
        helper("#addSheetButton", "ここから新たなシートを追加できます");
        helper("#reshapeButton", "このボタンでノードの整列ができます (Graphvizが必要)");
        helper("#sidebar", "ここに使用できるノードのリストがあります");
        helper("#nodelistSearch input", "ここでノードの検索ができます");
        helper("#saveButton", "現在のネットワークを保存できます。");
        helper("#executeButton", "現在のネットワークを実行できます");
        helper("#console", "エラーがあればここに表示されます。このコンソールは移動できます");
        helper("#fileManagerButton", "ここからすべてのファイルのアップロードやダウンロードができます");
        helper("#preferenceButton", "見た目や言語などの設定ができます");
        helper("#zoomSlider", "このスライダでノードの拡大・縮小ができます");
    } else {
        alert("unknwon language" + userPreference.get("lang"));
    }
}

function forScreenshot(){
    $("#topmenu,#sidebar,#console,#nodelistSearch,#footer").css("display", "none");
    $("#container").css("padding", "0px").css("margin", "0px");
    $("#sheets").css("padding", "0px").css("margin", "0px");
    $("body").css("padding", "0px").css("margin", "0px");
    $("html,body,#container,#sheets,.sheetContainer").css("height", "90%");
    $("#sheets").tabs("refresh");
}

function subnetParamMaintain(nodeid){
    var nodeobj = _.findObject(network.nodes, "id", nodeid);
    // console.log(nodeobj.parameters);
    if (nodeobj.sheet === "MAIN"){
        return;
    }
    _.each(nodeobj.parameters, function(p){
        if (p.type !== "subnet_param"){return;}
        if (p.value === ""){return;}

        var addparam = function(n){
            var origDef = _.findObject(nodedefs, "name", nodeobj.name);
            var origType = (origDef) ? _.findObject(
                origDef.parameters,
                "name", p.name) : undefined;
            origType = (origType) ? origType.type : "subnet_param";
            if (n.name === nodeobj.sheet &&
               !_.findObject(n.parameters, "name", p.value)){
                n.parameters.push({
                    condition: "",
                    description: "subnet_param",
                    list: [],
                    name: p.value,
                    type: origType,
                    value: ""            // same as ...
                });
            }
        };
        _.each(network.nodes, addparam);    // add to nodes
        _.each(nodedefs, addparam);         // add to nodedef
    });
    removeNonExistingSubnetParam();
}

function removeNonExistingSubnetParam(){
    // remove all non-existing subnet param from dynamic
    var filterNonExistingSubnetParam = function(nodeobj){
        return _.filter(nodeobj.parameters, function(paramobj){
            return _.chain(network.nodes)
                    .filter(function(n){return n.sheet === nodeobj.name;})
                    .pluck("parameters")
                    .flatten()
                    .filter(function(p){return p;})
                    .filter(function(p){return p.type === "subnet_param";})
                    .find(function(p){return paramobj.name === p.value;}).value() !== undefined;
            });
    };

    _.each(network.nodes, function(n){
        if (n.category === "Dynamic"){
            n.parameters = filterNonExistingSubnetParam(
                JSON.parse(JSON.stringify(n)));
        }
    });
    _.each(nodedefs, function(n){
        if (n.category === "Dynamic"){
            n.parameters = filterNonExistingSubnetParam(
                JSON.parse(JSON.stringify(n)));
        }
    });
}

function updateCurrentTab() {
    //set current tab
    var sheetTabs = $(".sheetTab");
    currentTab = $(sheetTabs[sheetTabs.length - 1]).attr("aria-controls").replace("sheet_", "");
}
